// This file is part of the disk cataloger Wssindex, but was written by a
// 3rd party.  See comments below for author and licensing information.

#include <stdio.h>
#include "OtherWssi/wssindex.h"
#ifdef FILENAME
	#undef FILENAME
#endif
#include "unzip.h"

//---------------------------------------------------------------------------
//
//  unshrink.c                     version 0.94                     26 Apr 94
//
//  Shrinking is basically a dynamic LZW algorithm with allowed code sizes of
//  up to 13 bits; in addition, there is provision for partial clearing of
//  leaf nodes.  PKWARE uses the special code 256 (decimal) to indicate a
//  change in code size or a partial clear of the code tree:  256,1 for the
//  former and 256,2 for the latter.  See the notes in the code below about
//  orphaned nodes after partial clearing.
//
//  This replacement version of unshrink.c was written from scratch.  It is
//  based only on the algorithms described in Mark Nelson's _The Data Compres-
//  sion Book_ and in Terry Welch's original paper in the June 1984 issue of
//  IEEE _Computer_; no existing source code, including any in Nelson's book,
//  was used.
//
//  Memory requirements are fairly large.  While the NODE struct could be mod-
//  ified to fit in a single 64KB segment (as a "far" data structure), for now
//  it is assumed that a flat, 32-bit address space is available.  outbuf2 is
//  always malloc'd, and flush() is always called with unshrink == false.
//
//  Copyright (C) 1994 Greg Roelofs.  See the accompanying file "COPYING" in
//  the UnZip 5.11 (or later) source distribution.
//
//  ---------------------------------------------------------------------------

#define farmalloc(x)	malloc(x)
#define farfree(x)	free(x)

#ifdef NEXTBYTE
	#undef NEXTBYTE
#endif
#define NEXTBYTE (compLength-- < 0 ? EOF : fgetc(fp))

extern unsigned short mask_bits[];

typedef struct leaf
	{
	struct leaf *parent;
	struct leaf *next_sibling;
	struct leaf *first_child;
	uch value;
	uch packing[3];
	} NODE;

static void partial_clear(NODE *cursib);

static NODE *node, *bogusnode, *lastfreenode;

int wssUnshrink(FILE *fp, long compLength, char *outptr, unsigned int outbufsiz)
	{
	static uch *stacktop;
	uch *newstr;
	int codesize=9, code, oldcode=0, len, KwKwK;
	NODE *freenode, *curnode, *lastnode=node, *oldnode;
		ulg bitbuf, outcnt;
	int bits_left;
	bool zipeof;
	unsigned char *stack;

//---------------------------------------------------------------------------
//	Initialize various variables.
//---------------------------------------------------------------------------
// Array node was too big for DOS version of Wssindex
	if((node = (NODE *)farmalloc(8192UL*sizeof(NODE))) == (NODE *)NULL)
		return PK_MEM;
	if(NULL == (stack = (unsigned char *)malloc(8193*sizeof(short))))
		{
		farfree((void *)node);
		return PK_MEM;
		}
	stacktop = stack + 8192 - 1;
	bogusnode = node + 256;
	lastfreenode = node + 256;

	for (code = 0;	code < 256; ++code)
		{
		node[code].value = code;
		node[code].parent = bogusnode;
		node[code].next_sibling = &node[code+1];
		node[code].first_child = (NODE *)NULL;
		}
	node[255].next_sibling = (NODE *)NULL;
	for (code = 257;  code < 8192;	++code)
		node[code].parent = node[code].next_sibling = (NODE *)NULL;

	bitbuf = 0;
	bits_left = 0;
	outcnt = 0L;
	zipeof = false;

//---------------------------------------------------------------------------
//	Get and output first code, then loop over remaining ones.
//---------------------------------------------------------------------------

	READBITS(codesize, oldcode)
	if (!zipeof)
		{
		*outptr++ = (uch)oldcode;
		if (++outcnt == outbufsiz)
			zipeof = true;
		}

	do {
		READBITS(codesize, code)
		if (zipeof)
			break;
		if (code == 256)
			{		// GRR:  possible to have consecutive escapes?
			READBITS(codesize, code)
			if (code == 1)
				{
				++codesize;
				}
			else if (code == 2)
				{
				partial_clear(node);			// recursive clear of leafs
				lastfreenode = bogusnode;  // reset start of free-node search
				}
			continue;
			}

	//-----------------------------------------------------------------------
	//		Translate code:	 traverse tree from leaf back to root.
	//-----------------------------------------------------------------------

		curnode = &node[code];
		newstr = stacktop;

		if (curnode->parent)
			KwKwK = false;
		else
			{
			KwKwK = true;
			--newstr;	// last character will be same as first character
			curnode = &node[oldcode];
			}

		do {
			*newstr-- = curnode->value;
			curnode = curnode->parent;
			} while (curnode != bogusnode);

		len = (int)(stacktop - newstr++);
		if (KwKwK)
			*stacktop = *newstr;

	//-----------------------------------------------------------------------
	//	Write expanded string in reverse order to output buffer.
	//-----------------------------------------------------------------------

		{
		uch *p;

		for (p = newstr;  p < newstr+len;  ++p)
			{
			*outptr++ = *p;
			if (++outcnt == outbufsiz)
				{
				zipeof = true;
				break;
				}
			}
		}
		if(zipeof)
			break;
	//-----------------------------------------------------------------------
	//	Add new leaf (first character of newstr) to tree as child of oldcode.
	//-----------------------------------------------------------------------

		// search for freenode
		freenode = lastfreenode + 1;
		while (freenode->parent)		// add if-test before loop for speed?
			++freenode;
		lastfreenode = freenode;

		oldnode = &node[oldcode];
		if (!oldnode->first_child)
			{   // no children yet:	add first one
			if (!oldnode->parent)
				{
				// oldnode is itself a free node:  the only way this can happen
				// is if a partial clear occurred immediately after oldcode was
				// received and therefore immediately before this step (adding
				// freenode).  This is subtle:	even though the parent no longer
				// exists, it is treated as if it does, and pointers are set as
				// usual.  Thus freenode is an orphan, *but only until the tree
				// fills up to the point where oldnode is reused*.	At that
				// point the reborn oldnode "adopts" the orphaned node.	 Such
				// wacky guys at PKWARE...
				//
				// To mark this, we set oldnode->next_sibling to point at the
				// bogus node (256) and then check for this in the freenode sec-
				// tion just below.
				oldnode->next_sibling = bogusnode;
				}
			oldnode->first_child = freenode;
			}
		else
			{
			curnode = oldnode->first_child;
			while (curnode)
				{				// find last child in sibling chain
				lastnode = curnode;
				curnode = curnode->next_sibling;
				}
			lastnode->next_sibling = freenode;
			}
		freenode->value = *newstr;
		freenode->parent = oldnode;
		if (freenode->next_sibling != bogusnode)  // no adoptions today...
			freenode->first_child = (NODE *)NULL;
		freenode->next_sibling = (NODE *)NULL;

		oldcode = code;
		} while (!zipeof);

//---------------------------------------------------------------------------
//	Flush any remaining data, free malloc'd space and return to sender...
//---------------------------------------------------------------------------

	free(stack);
	farfree((void *)node);
	return PK_OK;

	} // end function unshrink()



static void partial_clear(NODE *cursib)	// like, totally recursive, eh?
	{
	NODE *lastsib=(NODE *)NULL;

	// Loop over siblings, removing any without children; recurse on those
	// which do have children.	This hits even the orphans because they're
	// always adopted (parent node is reused) before tree becomes full and
	// needs clearing.
	do {
		if (cursib->first_child)
			{
			partial_clear(cursib->first_child);
			lastsib = cursib;
			}
		else if ((cursib - node) > 256)
			{	// no children (leaf):	clear it
			if (!lastsib)
				cursib->parent->first_child = cursib->next_sibling;
			else
				lastsib->next_sibling = cursib->next_sibling;
			cursib->parent = (NODE *)NULL;
			}
		cursib = cursib->next_sibling;
		} while (cursib);
	return;
	}
