// This file is part of the disk cataloger Wssindex.  Wssindex is copyright
// 1985-2021 by Robert W. Babcock and WSS Division of DDC.  Effective 2010,
// Wssindex is distributed under the terms of the GNU General Public
// License.  Contact the author at wssddc@wssddc.com or see the project web
// page at <https://www.wssddc.com/wssindex/>.
//
// Wssindex uses the open source Qt toolkit from the Qt Project (qt-project.org)
// and The Qt Company (www.qt.io).
//
// Wssindex is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the
// Free Software Foundation, either version 3 of the License, or (at your
// option) any later version.
//
// Wssindex is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with Wssindex.  If not, see <https://www.gnu.org/licenses/>.

#include "OtherWssi/wssindex.h"
#include "Dialogs/archivesdialog.h"
#include "ui_archivesdialog.h"
#include "Dialogs/mainwindow.h"
#include "OtherWssi/initialpos.h"
#include "OtherWssi/wssiproto.h"

ArchivesQueryDialog::ArchivesQueryDialog(QWidget *parent)
	: QMainWindow(parent), aq(new Ui::ArchivesQueryDialog)
	{
	aq->setupUi(this);
	this->setWindowFlags(MyWindowFlags);
	restoreSettings();

	aq->pushButton_query->setDefault(true);
	aq->radioButton_new_query->setChecked(true);
	aq->radioButton_arcfiles->setChecked(true);
	connect(this, SIGNAL(signal_updateDisplay(bool)),
		mainWindowPtr, SLOT(slot_updateDisplay(bool)));
	connect(aq->pushButton_close, SIGNAL(clicked()),
		this, SLOT(slot_pushButtonCloseClicked()));
	connect(aq->pushButton_query, SIGNAL(clicked()),
		this, SLOT(slot_pushButtonQueryClicked()));
	connect(this, SIGNAL(signal_keyToTable(int,int)),
		mainWindowPtr, SLOT(slot_keyToTable(int,int)));
	connect(this, SIGNAL(signal_wheelToTable(QPoint,int,Qt::MouseButtons,Qt::KeyboardModifiers,Qt::Orientation)),
		mainWindowPtr, SLOT(slot_wheelToTable(QPoint,int,Qt::MouseButtons,Qt::KeyboardModifiers,Qt::Orientation)));
	connect(aq->radioButton_arc_by_name, SIGNAL(toggled(bool)),
			  this, SLOT(slot_radioButtonArcByNameToggled(bool)));
	connect(aq->lineEdit_arcname, SIGNAL(textChanged(QString)),
			  this, SLOT(slot_lineEditArcnameChanged()));
	connect(mainWindowPtr, SIGNAL(signal_close_queries()),
			  this, SLOT(close()));
	}

ArchivesQueryDialog::~ArchivesQueryDialog()
	{
	delete aq;
	}

void ArchivesQueryDialog::restoreSettings()
	{
	QSettings settings("WSS-DDC", "Wssindex\\Geometry");
	if(settings.contains("ArchivesQuerySize"))
		{
		QSize size= settings.value("ArchivesQuerySize").toSize();
		resize(size);
		}
	if(settings.contains("ArchivesQueryPos"))
		{
		QPoint pos = settings.value("ArchivesQueryPos").toPoint();
		move(pos);
		forceOnscreen(this);
		}
	else
		{
// frameGeometry is the same as geometry until the dialog is displayed
		this->show();
		move(mainWindowPtr->offCenter(this->frameGeometry(), ARCHIVE_OFFSET));
		}
	}

// ------------------------------ Archive options -----------------------------
void ArchivesQueryDialog::slot_radioButtonArcByNameToggled(bool checked)
	{
// If archives and contents by name radio button is turned on,
// set the focus to the input field for the archive name.
	if(checked)
		aq->lineEdit_arcname->setFocus();
	}

void ArchivesQueryDialog::slot_lineEditArcnameChanged()
	{
// If an archive name is entered or changed, turn on the
// radio button to use this value.
	if(aq->lineEdit_arcname->text().size() > 0)
		aq->radioButton_arc_by_name->setChecked(true);
	}

// ------------------------------ Action buttons ------------------------------
void ArchivesQueryDialog::slot_pushButtonCloseClicked()
	{
	QSettings settings("WSS-DDC", "Wssindex\\Geometry");
	settings.setValue("ArchivesQueryPos", pos());
	settings.setValue("ArchivesQuerySize", size());
	this->close();
	}

void ArchivesQueryDialog::slot_pushButtonQueryClicked()
	{
	if(0 == Diskno)
		return;
	for(int i=0; i<SEL_DIM; i++)
		Usesel[i] = false;
	Nselect = 0;
// If not a new query, mark which files were previously selected
	if(!aq->radioButton_new_query->isChecked())
		updatePreviouslySelected();

	if(aq->radioButton_add_to->isChecked())
		{
		Sel[Nselect++] = SEL_ADD_TO_PREVIOUS;
		Usesel[SEL_ADD_TO_PREVIOUS] = true;
		}

	if(aq->radioButton_pick_from->isChecked())
		{
		Sel[Nselect++] = SEL_FROM_PREVIOUS;
		Usesel[SEL_FROM_PREVIOUS] = true;
		}

	if(aq->radioButton_exclude_from->isChecked())
		{
		Sel[Nselect++] = SEL_REM_FROM_PREVIOUS;
		Usesel[SEL_REM_FROM_PREVIOUS] = true;
		}

	if(aq->radioButton_arcfiles->isChecked())
		{
		Sel[Nselect++]	= SEL_ARC_FILES;
		Usesel[SEL_ARC_FILES] = true;
		}

	if(aq->radioButton_arc_and_contents->isChecked())
		{
		Sel[Nselect++] = SEL_ARC_AND_MEMB;
		Usesel[SEL_ARC_AND_MEMB] = true;
		}

	if(aq->radioButton_arc_by_name->isChecked())
		{
		Sel[Nselect++] = SEL_ARCMEM_BY_ARCNAME;
		Usesel[SEL_ARCMEM_BY_ARCNAME] = true;
		Comparc = aq->lineEdit_arcname->text();
		}

	mainWindowPtr->doSelection();
	Q_EMIT signal_updateDisplay(true);
	}

// ------------------------------ Event handlers ------------------------------
void ArchivesQueryDialog::closeEvent(QCloseEvent *event)
	{
	event->accept();
	}

void ArchivesQueryDialog::keyPressEvent(QKeyEvent *event)
	{
	switch(event->key())
		{
		case Qt::Key_Home:
		case Qt::Key_End:
			if(!(Qt::ControlModifier & event->modifiers()))
				break;
			Q_FALLTHROUGH();
		case Qt::Key_PageDown:
		case Qt::Key_PageUp:
			Q_EMIT signal_keyToTable(event->key(), static_cast<int>(event->modifiers()));
			break;
		case Qt::Key_Escape:
			this->close();
// Intercepting keyPressEvents seems to break default button activation, so
// do it here.  Also take care setting focus after other buttons are clicked.
			Q_FALLTHROUGH();
		case Qt::Key_Enter:
		case Qt::Key_Return:
			slot_pushButtonQueryClicked();
			break;
		default:
			return;
		}
	event->ignore();
	return;
	}

void ArchivesQueryDialog::wheelEvent(QWheelEvent *event)
	{
	Q_EMIT signal_wheelToTable(event->pos(), event->delta(), event->buttons(),
										event->modifiers(), event->orientation());
	event->ignore();
	return;
	}
