// This file is part of the disk cataloger Wssindex.  Wssindex is copyright
// 1985-2021 by Robert W. Babcock and WSS Division of DDC.  Effective 2010,
// Wssindex is distributed under the terms of the GNU General Public
// License.  Contact the author at wssddc@wssddc.com or see the project web
// page at <https://www.wssddc.com/wssindex/>.
//
// Wssindex uses the open source Qt toolkit from the Qt Project (qt-project.org)
// and The Qt Company (www.qt.io).
//
// Wssindex is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the
// Free Software Foundation, either version 3 of the License, or (at your
// option) any later version.
//
// Wssindex is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with Wssindex.  If not, see <https://www.gnu.org/licenses/>.

#include "OtherWssi/wssindex.h"
#include "Dialogs/mainwindow.h"
#include "ui_mainwindow.h"
#include "Dialogs/fullsortdialog.h"
#include "ui_fullsortdialog.h"
#include "Dialogs/mainwindow.h"
#include "OtherWssi/initialpos.h"
#include "OtherWssi/wssiproto.h"
#include "Workarounds/submenuproxy.h"

#include <QtDebug>

// Order of names here must match numerical order of SortKey_XXX defines
static QString keynames[] =
	{QObject::tr("File name"),
	 QObject::tr("Extension"),
	 QObject::tr("Volume label"),
	 QObject::tr("Numeric label"),
	 QObject::tr("Subdirectory"),
	 QObject::tr("Size"),
	 QObject::tr("Timestamp"),
	 QObject::tr("Category"),
	 QObject::tr("Comment"),
	 QObject::tr("Extra info"),
	 QObject::tr("Archive name"),
	 QObject::tr("Archive extension")};

FullSortDialog::FullSortDialog(QWidget *parent)
	: QMainWindow(parent), fsd(new Ui::FullSortDialog)
	{
	fsd->setupUi(this);
	this->setWindowFlags(MyWindowFlags);
	restoreSettings();
	connect(fsd->pushButton_close, SIGNAL(clicked()),
		this, SLOT(slot_pushButtonCloseClicked()));
	connect(fsd->pushButton_sort, SIGNAL(clicked()),
			  this, SLOT(slot_pushButtonSortClicked()));

	connect(this, SIGNAL(signal_sortNeeded()),
		mainWindowPtr, SLOT(slot_sortdatabase()));

	connect(fsd->tableWidget_available_keys, SIGNAL(cellDoubleClicked(int,int)),
			  this, SLOT(slot_pushButtonAddKeyClicked()));
	connect(fsd->tableWidget_use_keys, SIGNAL(cellDoubleClicked(int,int)),
		this, SLOT(slot_useTableDoubleClicked(int,int)));

	connect(fsd->pushButton_add_key, SIGNAL(clicked(bool)),
			  this, SLOT(slot_pushButtonAddKeyClicked()));
	connect(fsd->pushButton_remove_key, SIGNAL(clicked()),
			  this, SLOT(slot_pushButtonRemoveKeyClicked()));
	connect(fsd->pushButton_move_key_up, SIGNAL(clicked()),
			  this, SLOT(slot_pushButtonMoveKeyUpClicked()));
	connect(fsd->pushButton_move_key_down, SIGNAL(clicked()),
			  this, SLOT(slot_pushButtonMoveKeyDownClicked()));

	connect(this, SIGNAL(signal_keyToTable(int,int)),
		parent, SLOT(slot_keyToTable(int,int)));
	connect(this, SIGNAL(signal_wheelToTable(QPoint,int,Qt::MouseButtons,Qt::KeyboardModifiers,Qt::Orientation)),
		mainWindowPtr, SLOT(slot_wheelToTable(QPoint,int,Qt::MouseButtons,Qt::KeyboardModifiers,Qt::Orientation)));

	connect(mainWindowPtr, SIGNAL(signal_close_sorts()),
			  this, SLOT(close()));

// Populate the list of possible sort keys
	ascending = new QTableWidgetItem;
	descending = new QTableWidgetItem;
	ascending->setText(tr("Ascending"));
	descending->setText(tr("Descending"));
	int keycount = sizeof(keynames)/sizeof(QString);
	fsd->tableWidget_available_keys->setRowCount(keycount);
	fsd->tableWidget_available_keys->setColumnCount(1);
	fsd->tableWidget_use_keys->setRowCount(0);
	fsd->tableWidget_use_keys->setColumnCount(2);
	fsd->tableWidget_available_keys->setShowGrid(false);
	fsd->tableWidget_use_keys->setShowGrid(false);
	this->setStyleSheet(
		tr("QTableWidget::item:selected:focus {background-color: %1; border:0px;}")
				.arg(QApplication::palette(mainWindowPtr->ui->menuBar).highlight().color().name()));
	for(int i=0; i<keycount; i++)
		{
		QTableWidgetItem *item = new QTableWidgetItem;
		item->setText(keynames[i]);
		item->setData(Qt::UserRole, i+1);
		fsd->tableWidget_available_keys->setItem(i, 0, item);
		}
	fsd->tableWidget_available_keys->resizeColumnsToContents();
	fsd->tableWidget_available_keys->setCurrentCell(0, 0);
	fsd->label_sort_time->setText(QFA(""));
	}

FullSortDialog::~FullSortDialog()
	{
	delete fsd;
	}

void FullSortDialog::restoreSettings()
	{
	QSettings settings("WSS-DDC", "Wssindex\\Geometry");
	if(settings.contains("FullSortSize"))
		{
		QSize size = settings.value("FullSortSize").toSize();
		resize(size);
		}
	if(settings.contains("FullSortPos"))
		{
		QPoint pos = settings.value("FullSortPos").toPoint();
		move(pos);
		forceOnscreen(this);
		}
	else
		{
// frameGeometry is the same as geometry until the dialog is displayed
		this->show();
		move(mainWindowPtr->offCenter(this->frameGeometry(), FULL_SORT_OFFSET));
		}
	}

// ----------------------------- Move key buttons -----------------------------
void FullSortDialog::slot_pushButtonAddKeyClicked()
	{
	if(0 == fsd->tableWidget_available_keys->rowCount())
		return;
// Following test requires that the buttons for moving keys have focus
// policy set to NoFocus.  Otherwise, the button gets the focus when
// clicked and this test is done too late.
	if(!fsd->tableWidget_available_keys->hasFocus())
		return;

	int row = fsd->tableWidget_available_keys->currentRow();
	int newrow = fsd->tableWidget_use_keys->rowCount();
	fsd->tableWidget_use_keys->setRowCount(newrow+1);
	fsd->tableWidget_use_keys->setItem(newrow, 0,
			fsd->tableWidget_available_keys->takeItem(row, 0));
	fsd->tableWidget_use_keys->setItem(newrow, 1, ascending->clone());
	fsd->tableWidget_use_keys->resizeColumnsToContents();

	fsd->tableWidget_available_keys->removeRow(row);
	if(row >= fsd->tableWidget_available_keys->rowCount())
		row--;
	if(row < 0)
		return;
	fsd->tableWidget_available_keys->setCurrentCell(row, 0);
	fsd->tableWidget_available_keys->setFocus();
	}

void FullSortDialog::slot_pushButtonRemoveKeyClicked()
	{
	if(0 == fsd->tableWidget_use_keys->rowCount())
		return;
	if(!fsd->tableWidget_use_keys->hasFocus())
		return;

	int row = fsd->tableWidget_use_keys->currentRow();
	int newrow = fsd->tableWidget_available_keys->rowCount();
	fsd->tableWidget_available_keys->setRowCount(newrow+1);
	fsd->tableWidget_available_keys->setItem(newrow, 0,
			fsd->tableWidget_use_keys->takeItem(row, 0));
	fsd->tableWidget_available_keys->setItem(newrow, 1,
			fsd->tableWidget_use_keys->takeItem(row, 1));
	fsd->tableWidget_available_keys->resizeColumnsToContents();

	fsd->tableWidget_use_keys->removeRow(row);
	if(row >= fsd->tableWidget_use_keys->rowCount())
		row--;
	if(row < 0)
		return;
	fsd->tableWidget_use_keys->setCurrentCell(row, 0);
	fsd->tableWidget_use_keys->setFocus();
	}

void FullSortDialog::slot_pushButtonMoveKeyUpClicked()
	{
	if(0 == fsd->tableWidget_use_keys->rowCount())
		return;
	if(!fsd->tableWidget_use_keys->hasFocus())
		return;

	int row_to_move = fsd->tableWidget_use_keys->currentRow();
	if(0 == row_to_move)
		return;
	fsd->tableWidget_use_keys->insertRow(row_to_move-1);
	fsd->tableWidget_use_keys->setItem(row_to_move-1,0,
				fsd->tableWidget_use_keys->takeItem(row_to_move+1,0));
	fsd->tableWidget_use_keys->setItem(row_to_move-1,1,
				fsd->tableWidget_use_keys->takeItem(row_to_move+1,1));
	fsd->tableWidget_use_keys->removeRow(row_to_move+1);
	fsd->tableWidget_use_keys->setCurrentCell(row_to_move-1, 0);
	fsd->tableWidget_use_keys->setFocus();
	}

void FullSortDialog::slot_pushButtonMoveKeyDownClicked()
	{
	int  rows = fsd->tableWidget_use_keys->rowCount();
	if(0 == rows)
		return;
	if(!fsd->tableWidget_use_keys->hasFocus())
		return;

	int row_to_move = fsd->tableWidget_use_keys->currentRow();
	if(rows-1 == row_to_move)
		return;
	fsd->tableWidget_use_keys->insertRow(row_to_move);
	fsd->tableWidget_use_keys->setItem(row_to_move,0,
				fsd->tableWidget_use_keys->takeItem(row_to_move+2,0));
	fsd->tableWidget_use_keys->setItem(row_to_move,1,
				fsd->tableWidget_use_keys->takeItem(row_to_move+2,1));
	fsd->tableWidget_use_keys->removeRow(row_to_move+2);
	fsd->tableWidget_use_keys->setCurrentCell(row_to_move+1, 0);
	fsd->tableWidget_use_keys->setFocus();
	}

// --------------------------- Change sort direction --------------------------
void FullSortDialog::slot_useTableDoubleClicked(int row, int column)
	{
	if(1 != column)
		{
		slot_pushButtonRemoveKeyClicked();
		return;
		}
	int sortkey = -fsd->tableWidget_use_keys->item(row,0)->data(Qt::UserRole).toInt();
	fsd->tableWidget_use_keys->item(row,0)->setData(Qt::UserRole, sortkey);
	fsd->tableWidget_use_keys->removeCellWidget(row,1);
	if(sortkey > 0)
		fsd->tableWidget_use_keys->setItem(row, 1, ascending->clone());
	else
		fsd->tableWidget_use_keys->setItem(row, 1, descending->clone());
	fsd->tableWidget_use_keys->resizeColumnsToContents();
	}

// ------------------------------ Action buttons ------------------------------
void FullSortDialog::slot_pushButtonCloseClicked()
	{
	QSettings settings("WSS-DDC", "Wssindex\\Geometry");
	settings.setValue("FullSortPos", pos());
	settings.setValue("FullSortSize", size());
	this->close();
	}

void FullSortDialog::slot_pushButtonSortClicked()
	{
	if(0 == Noselect)
		return;
	int nkey = fsd->tableWidget_use_keys->rowCount();
	if(0 == nkey)
		return;
	Nkey = nkey;
	for(int i=0; i<Nkey; i++)
		{
		SortKey[i] = fsd->tableWidget_use_keys->item(i,0)->data(Qt::UserRole).toInt();
		if(SortKey[i] < 0)
			{
			SortKey[i] = -SortKey[i];
			Sortdir[i] = -1;
			}
		else
			Sortdir[i] = 1;
		}

	QElapsedTimer timer;
	timer.start();
	Q_EMIT signal_sortNeeded();
	fsd->label_sort_time->setText(tr("Sort time: %1 seconds").arg(static_cast<double>(timer.elapsed())/1000.));
	}

// ------------------------------ Event handlers ------------------------------
void FullSortDialog::closeEvent(QCloseEvent *event)
	{
	event->accept();
	}

void FullSortDialog::keyPressEvent(QKeyEvent *event)
	{
	switch(event->key())
		{
		case Qt::Key_Home:
		case Qt::Key_End:
			if(!(Qt::ControlModifier & event->modifiers()))
				break;
			Q_FALLTHROUGH();
		case Qt::Key_PageDown:
		case Qt::Key_PageUp:
			Q_EMIT signal_keyToTable(event->key(), static_cast<int>(event->modifiers()));
			break;
		case Qt::Key_Escape:
		case Qt::Key_Return:
			this->close();
			break;
		default:
			return;
		}
	event->ignore();
	return;
	}

void FullSortDialog::wheelEvent(QWheelEvent *event)
	{
	Q_EMIT signal_wheelToTable(event->pos(), event->delta(), event->buttons(),
										event->modifiers(), event->orientation());
	event->ignore();
	return;
	}
