// This file is part of the disk cataloger Wssindex.  Wssindex is copyright
// 1985-2021 by Robert W. Babcock and WSS Division of DDC.  Effective 2010,
// Wssindex is distributed under the terms of the GNU General Public
// License.  Contact the author at wssddc@wssddc.com or see the project web
// page at <https://www.wssddc.com/wssindex/>.
//
// Wssindex uses the open source Qt toolkit from the Qt Project (qt-project.org)
// and The Qt Company (www.qt.io).
//
// Wssindex is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the
// Free Software Foundation, either version 3 of the License, or (at your
// option) any later version.
//
// Wssindex is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with Wssindex.  If not, see <https://www.gnu.org/licenses/>.

#include "OtherWssi/wssindex.h"
#include "Dialogs/indexwindow.h"
#include "ui_indexwindow.h"
#include "OtherWssi/globalsettings.h"
#include "OtherWssi/wssiproto.h"
#include "Dialogs/mainwindow.h"
#include "Dialogs/messagelogwindow.h"

#include <QButtonGroup>

IndexWindow::IndexWindow(QWidget *parent)
	 : QMainWindow(parent), ui(new Ui::IndexWindow)
	{
	MessageLogWindow *ml = new MessageLogWindow(this);
	ml->setWindowModality(Qt::NonModal);
	ml->setAttribute(Qt::WA_DeleteOnClose);
	ml->show();

	mlptr = ml;
	if(0 == Maxfiles)
		mainWindowPtr->simpleconf(static_cast<unsigned long>(Maxfiles_initial),
			static_cast<unsigned long>(Maxsubs_initial), static_cast<unsigned long>(Maxdisks_initial));

	ui->setupUi(this);
	this->setWindowFlags(MyWindowFlags);
	this->setWindowModality(Qt::WindowModal);
//	restoreSettings();
	drive_letter[0] = ui->comboBox_drive1;
	drive_letter[1] = ui->comboBox_drive2;
	index_button[0] = ui->pushButton_index1;
	index_button[1] = ui->pushButton_index2;
	currentIndexingButton = 0;

// Populate the drive letter pulldowns
	for(int i=0; i<26; i++)
		{
		drive_letter[0]->insertItem(i+1, QChar('A'+i), i);
		drive_letter[1]->insertItem(i+1, QChar('A'+i), i);
		}
	drive_letter[0]->setCurrentIndex(Indexing_drive_letter[0] - 'A');
	drive_letter[1]->setCurrentIndex(Indexing_drive_letter[1] - 'A');
	widerCombo(drive_letter[0]);
	widerCombo(drive_letter[1]);

	ui->lineEdit_startdir->setText(IndexStartingDir);
	ui->checkBox_useExclusions->setChecked(IndexTimeFiltering);

	slot_setIndexButtonLabels();
	connect(drive_letter[0], SIGNAL(currentIndexChanged(int)),
			  this, SLOT(slot_setIndexButtonLabels()));
	connect(drive_letter[1], SIGNAL(currentIndexChanged(int)),
			  this, SLOT(slot_setIndexButtonLabels()));

// Make a button group to make it easier to tell which index button is clicked
	ibuttons = new QButtonGroup();
	ibuttons->addButton(ui->pushButton_index1, 0);
	ibuttons->addButton(ui->pushButton_index2, 1);
	connect(ibuttons, SIGNAL(buttonClicked(int)), this, SLOT(slot_indexButtonClicked(int)));

// And another button group for file information extraction options
	ebuttons = new QButtonGroup();
	ebuttons->addButton(ui->radioButton_extract_none);
	ebuttons->setId(ui->radioButton_extract_none, EXTRACT_NONE);
	ebuttons->addButton(ui->radioButton_extract_configured);
	ebuttons->setId(ui->radioButton_extract_configured, EXTRACT_CONFIGURED);
	ebuttons->addButton(ui->radioButton_extract_all);
	ebuttons->setId(ui->radioButton_extract_all, EXTRACT_ALL);
	connect(ebuttons, SIGNAL(buttonClicked(int)), this, SLOT(slot_extractionOptionChanged(int)));

	connect(ui->pushButton_done, SIGNAL(clicked()),
			  this, SLOT(slot_buttonDoneClicked()));
	connect(ui->pushButton_saveAndDone, SIGNAL(clicked()),
			  this, SLOT(slot_buttonsaveAndDoneClicked()));
	connect(ui->pushButton_browseDir, SIGNAL(clicked()),
			  this, SLOT(slot_buttonBrowseClicked()));
	connect(ui->pushButton_configExclusions, SIGNAL(clicked()),
			  this, SLOT(slot_configExcludeClicked()));
	connect(ui->pushButton_extractConfigure, SIGNAL(clicked()),
			  this, SLOT(slot_indexSettingsClicked()));
	connect(this, SIGNAL(signal_settingsClicked()),
			  mainWindowPtr, SLOT(slot_settingsClicked()));
	connect(ui->pushButton_saveTheseSettings, SIGNAL(clicked()),
			  this, SLOT(slot_buttonSaveTheseSettings()));
	connect(ui->lineEdit_startdir, SIGNAL(editingFinished()),
			  this, SLOT(slot_fixupStartingDir()));

	NofileBeforeIndexing = NoPlainFiles;
	SubnoBeforeIndexing = Subno;
	Stop_indexing = true;
	countUpdateTimer = new QTimer();
	countUpdateTimer->setSingleShot(true);
	connect(countUpdateTimer, SIGNAL(timeout()), this, SLOT(slot_updateDbTotals()));
	slot_updateDbTotals();

	switch(VolnameOpt)
		{
		case VOLNAME_TRUE:
			ui->radioButton_trueLabel->setChecked(true);
			break;
		case VOLNAME_PROMPT:
			ui->radioButton_promptLabel->setChecked(true);
			break;
		case VOLNAME_GENERATE:
			ui->radioButton_serialLabel->setChecked(true);
			break;
		}

	switch(Extract_opt)
		{
		case EXTRACT_NONE:
			ui->radioButton_extract_none->setChecked(true);
			break;
		case EXTRACT_CONFIGURED:
			ui->radioButton_extract_configured->setChecked(true);
			break;
		case EXTRACT_ALL:
			ui->radioButton_extract_all->setChecked(true);
			break;
		}

	ui->lineEdit_serString->setText(Serialization_format);
	ui->spinBox_serNum->setValue(static_cast<int>(Next_serial_number));

	connect(this, SIGNAL(signal_logMessage(QString)),
			  ml, SLOT(slot_logMessage(QString)));
	connect(this, SIGNAL(signal_closeLog()), ml, SLOT(slot_closeLog()));
	connect(this, SIGNAL(signal_diskListChanged()),
			  mainWindowPtr, SLOT(slot_diskListChanged()));
	connect(this, SIGNAL(signal_updateDisplay(bool)),
		mainWindowPtr, SLOT(slot_updateDisplay(bool)));
	restoreSettings();
	}

IndexWindow::~IndexWindow()
	{
	Q_EMIT signal_closeLog();
	Indexing_drive_letter[0] = drive_letter[0]->currentText().toLatin1()[0];
	Indexing_drive_letter[1] = drive_letter[1]->currentText().toLatin1()[0];
	IndexStartingDir = ui->lineEdit_startdir->text();
	IndexTimeFiltering = ui->checkBox_useExclusions->isChecked();
	if(ui->radioButton_extract_none->isChecked())
		Extract_opt = EXTRACT_NONE;
	else if(ui->radioButton_extract_configured->isChecked())
		Extract_opt = EXTRACT_CONFIGURED;
	else
		Extract_opt = EXTRACT_ALL;

	if(ui->radioButton_trueLabel->isChecked())
		VolnameOpt = VOLNAME_TRUE;
	else if(ui->radioButton_promptLabel->isChecked())
		VolnameOpt = VOLNAME_PROMPT;
	else
		VolnameOpt = VOLNAME_GENERATE;
	Next_serial_number = static_cast<unsigned int>(ui->spinBox_serNum->value());
	Serialization_format = ui->lineEdit_serString->text();
	if(NoId)
		selectAllExceptIdUnsorted();
	else
		selectAllUnsorted();
	delete countUpdateTimer;
	delete ebuttons;
	delete ibuttons;
	delete ui;
	Q_EMIT signal_updateDisplay(true);
	}

void IndexWindow::keyPressEvent(QKeyEvent *event)
	{
	switch(event->key())
		{
		case Qt::Key_Enter:
		case Qt::Key_Return:
				QMetaObject::invokeMethod(this, "slot_indexButtonClicked",
					Qt::QueuedConnection, Q_ARG(int, currentIndexingButton));
			event->ignore();
			return;
		default:
			return;
		}
	}

// Don't let the user close this window if indexing is in progress
void IndexWindow::closeEvent(QCloseEvent *event)
	{
	if(InhibitQuit)
		{
		event->ignore();
		return;
		}
	event->accept();
	}
