// This file is part of the disk cataloger Wssindex.  Wssindex is copyright
// 1985-2021 by Robert W. Babcock and WSS Division of DDC.  Effective 2010,
// Wssindex is distributed under the terms of the GNU General Public
// License.  Contact the author at wssddc@wssddc.com or see the project web
// page at <https://www.wssddc.com/wssindex/>.
//
// Wssindex uses the open source Qt toolkit from the Qt Project (qt-project.org)
// and The Qt Company (www.qt.io).
//
// Wssindex is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the
// Free Software Foundation, either version 3 of the License, or (at your
// option) any later version.
//
// Wssindex is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with Wssindex.  If not, see <https://www.gnu.org/licenses/>.

#ifndef INDEXWINDOW_H
#define INDEXWINDOW_H

#define VOLUME		0x08
#define SUBDIR		0x10

#define NOT_DUPLICATE		0
#define REPLACE				1
#define REPLACE_AND_COPY	2
#define DONT_REPLACE			3

#include <QMainWindow>
#include <QtGui>
#include <QPushButton>
#include <QComboBox>

#ifndef EXTERN
	#define EXTERN extern
#endif
#include "ExtractSubs/extractinfo.h"

enum DirSkipVal {SkipAll, SkipFiles, SkipNone};
enum disktypes {WSS_HARD, WSS_USB, WSS_NETWORK, WSS_CDROM, WSS_RAMDISK, WSS_MOUNTED, WSS_FLOPPY};

namespace Ui
	{
	class IndexWindow;
	}

class MainWindow;
class MessageLogWindow;

class IndexWindow : public QMainWindow
	{
	Q_OBJECT

public:
	IndexWindow(QWidget *parent);
	~IndexWindow() override;
	Ui::IndexWindow *ui;
	void saveIndexTimeFilter();
	disktypes diskAnimation;

public:
	bool Stop_indexing;

private:
	void keyPressEvent(QKeyEvent *event) override;
	void closeEvent(QCloseEvent *event) override;
	void restoreSettings();
	void diskadd(int *action);
	bool skip_extract(QString file_name);
	void makeid(int id_type);
	void index_disk(QString path, bool doFiles);
	void open_indexing_message(struct diskrec *currentVrec);
	int check_duplicate_volname(QString newvolname);
	void delete_disk(unsigned int oldDisk);
	void copy_and_delete(unsigned int oldDisk);
	void setIndexButtonColors(bool indexing);
	DirSkipVal indexTimeDirectoryFilteringSkip(QString dirname);
	bool indexTimeFileFilteringSkip(QString filename);
	void indexFinish();
	void indexAbort();
	enum disktypes getDiskType(QString rootPath);

// Info extraction routines
	int zip7add(QString path, QString filename, int mode);
	int arcadd(QString path, QString filename, int mode);
	int arjadd(QString path, QString filename, int mode);
	int auadd(QString path, QString filename, int dummy);
	int aviadd(QString path, QString filename, int dummy);
	int bmpadd(QString path, QString filename, int dummy);
	int cabadd(QString path, QString filename, int dummy);
	int fliadd(QString path, QString filename, int dummy);
	int gifadd(QString path, QString filename, int dummy);
	int hapadd(QString path, QString filename, int dummy);
	int hypadd(QString path, QString filename, int dummy);
	int jpgadd(QString path, QString filename, int dummy);
	int lbradd(QString path, QString filename, int dummy);
	int lhaadd(QString path, QString filename, int dummy);
	int mediaInfoSoundAdd(QString path, QString filename, int mode);
	int mediaInfoVideoAdd(QString path, QString filename, int mode);
	int modadd(QString path, QString filename, int dummy);
	int mp3add(QString path, QString filename, int dummy);
	int pcxadd(QString path, QString filename, int dummy);
	int pngadd(QString path, QString filename, int dummy);
	int psfadd(QString path, QString filename, int dummy);
	int raradd(QString path, QString filename, int dummy);
	int tgaadd(QString path, QString filename, int dummy);
	int tifadd(QString path, QString filename, int dummy);
	int ttfadd(QString path, QString filename, int dummy);
	int vocadd(QString path, QString filename, int dummy);
	int wavadd(QString path, QString filename, int dummy);
	int zipadd(QString path, QString filename, int mode);
	int zooadd(QString path, QString filename, int mode);

// Used by info extraction routines
	FILE *arc_open(QString path, QString filename);
	bool com_arch(char *buffer, struct filerec *dr_nofile, int length);
	bool extra_arch(char *buffer, struct filerec *dr_nofile, int length);
	bool extra_arch(QString buffer, struct filerec *dr_nofile);
	bool ExtractDiz(unsigned long filenum, short diz_disk, unsigned long diz_loc_head,
				FILE **fp, QString path, QString filename);
	void extract_info(QString path, struct filerec *dr_info);
	unsigned long fget_7_long(FILE *fp);
	unsigned long fget_guess_long(FILE *fp, unsigned long id3_max);
	unsigned long fget_unsync_long(FILE *fp, bool unsync);
	unsigned long find_sync(FILE *fp);
	void make_arj_crctable(void);
	void charToQstring(char *ch, int size, QString &str);
	char *no_path(char *str, int length);
	void no_path(QString *fn);
	QString num_color_str(int nbits);
	QString id3text(FILE *fp, unsigned long frame_size, bool hasShortDesc, bool &corruptMp3);
	bool read_avi_chunk(FILE *fp, unsigned long *chunkname1, unsigned long *chunkname2, unsigned long *length);
	void swap2(void *swapme);
	void swap2_tif(char *byte_val);
	void swap4(void *swapme);
	void swap4_tif(char *byte_val);
	long tif_decode(void);
	size_t zread(void *buffer, size_t size, size_t n, FILE **fp, bool *continued,
				QString path,	QString filename);

private:
	MessageLogWindow *mlptr;
	int currentIndexingButton;
	bool Ctlchit;
	QString Ddrive[2];
	QString Ddall[2];
	unsigned long Subuse;
	unsigned long NofileBeforeIndexing;
	unsigned long SubnoBeforeIndexing;
	QComboBox *drive_letter[2];
	QPushButton *index_button[2];
	unsigned int oldVolno;
	int duplicate_volname;
	unsigned long Nfold;
	bool QueryOverride;
	bool SkipOverrideAnswer;
	int offset_to_saved_dir;
	unsigned long skippedDirCount;
	unsigned long almostSkippedDirCount;
	unsigned long skippedFileCount;
	unsigned long almostSkippedFileCount;
	QTimer *countUpdateTimer;
	QButtonGroup *ibuttons;
	QButtonGroup *ebuttons;
	quint64 diskSize;

private Q_SLOTS:
	void slot_indexSettingsClicked();
	void slot_setIndexButtonLabels();
	void slot_indexButtonClicked(int button);
	void slot_extractionOptionChanged(int radiobutton);
	void slot_buttonDoneClicked();
	void slot_buttonsaveAndDoneClicked();
	void slot_buttonStopClicked();
	void slot_buttonBrowseClicked();
	void slot_configExcludeClicked();
	void slot_buttonSaveTheseSettings();
	void slot_updateDbTotals();
	void slot_fixupStartingDir();

Q_SIGNALS:
	void signal_diskListChanged();
	void signal_settingsClicked();
	void signal_indexIsDone();
	void signal_showVolname(QString volname);
	void signal_logMessage(QString message);
	void signal_closeLog();
	void signal_updateDisplay(bool);
	};

#endif // INDEXWINDOW_H
