// This file is part of the disk cataloger Wssindex.  Wssindex is copyright
// 1985-2021 by Robert W. Babcock and WSS Division of DDC.  Effective 2010,
// Wssindex is distributed under the terms of the GNU General Public
// License.  Contact the author at wssddc@wssddc.com or see the project web
// page at <https://www.wssddc.com/wssindex/>.
//
// Wssindex uses the open source Qt toolkit from the Qt Project (qt-project.org)
// and The Qt Company (www.qt.io).
//
// Wssindex is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the
// Free Software Foundation, either version 3 of the License, or (at your
// option) any later version.
//
// Wssindex is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with Wssindex.  If not, see <https://www.gnu.org/licenses/>.

#include "OtherWssi/wssindex.h"
#include "OtherWssi/wssiproto.h"
#include "Dialogs/indexwindow.h"
#include "ui_indexwindow.h"
#include "Dialogs/indexingdialog.h"
#include "OtherWssi/globalsettings.h"
#include "Dialogs/mainwindow.h"
#include <QDebug>

#include <windows.h>
BOOL ejectNT(TCHAR cDriveLetter);

#include <QInputDialog>

// This recursive routine does the actual indexing
// that was started by slot_indexButtonClicked()

void IndexWindow::index_disk(QString path, bool doFiles)
	{
	if(Stop_indexing)
		return;
	if(!QFileInfo(path).isDir())
		{
		QMessageBox::critical(this, tr("Error"),
			tr("index_disk was called with non-directory argument\n%1\n\n"
			"This probably means there was a disk read error of some sort.  Indexing will stop.").arg(path));
		Stop_indexing = true;
		return;
		}
//	There's a known problem in the QDirIterator flags: if QDir::Hidden is specified,
// you get both hidden files and hidden subdirectories even if you only want
// one or the other.  It's not so much a bug as a missing capability in the API.
// See https://bugreports.qt.io/browse/QTBUG-12965
// and https://bugreports.qt.io/browse/QTBUG-12584
	QFlags<QDir::Filter> fileindexmode = QDir::Files;
	QFlags<QDir::Filter> dirindexmode = QDir::Dirs | QDir::NoDotAndDotDot | QDir::AllDirs;
	if(Index_hidden)
		{
		fileindexmode |= QDir::Hidden | QDir::System;
		dirindexmode |= QDir::Hidden | QDir::System;
		}
	QDirIterator files(path, fileindexmode);

	struct diskrec *vrec_i = Vrec + Diskno;
	while (doFiles && files.hasNext())
		{
		if(Stop_indexing)
			return;
		files.next();
		QFileInfo finfo = files.fileInfo();
		if(!finfo.isFile())		// skip directories
			continue;
		if(indexTimeFileFilteringSkip(finfo.fileName()))
			continue;
		struct filerec *dr_nofile = Dr + Nofile;
		dr_nofile->Ddate = finfo.lastModified().date();
		dr_nofile->Dtime = finfo.lastModified().time();
		dr_nofile->Dsize = static_cast<unsigned long long>(finfo.size());
		dr_nofile->Subdirno = Subuse;
		dr_nofile->Volno = Diskno;
		dr_nofile->Bitflags = 0;
		dr_nofile->Diskbits = HAS_LFN;
		dr_nofile->Lfn = finfo.fileName();
		dr_nofile->ExtOffset = static_cast<unsigned short>(lastdot(dr_nofile->Lfn));
		if(mainWindowPtr->incfile_and_check_for_error(false))
			return;
		dr_nofile = Dr + Nofile - 1;	// If Dr had to be expanded, dr_nofile is invalid
		vrec_i->Vfiles++;
		if(!ui->radioButton_extract_none->isChecked())
			extract_info(path, dr_nofile);
		}

// This loop processes directories
	QDirIterator dir(path, dirindexmode);
	while (dir.hasNext())
		{
		dir.next();
		qApp->processEvents();
		if(Stop_indexing)
			return;
		QFileInfo finfo = dir.fileInfo();
		if(!finfo.isDir() || finfo.isSymLink())		// skip files
			continue;
		DWORD attributes = GetFileAttributesW(LPCWSTR(dir.filePath().utf16()));
		if(Skip_NTFS_mounted && (attributes & FILE_ATTRIBUTE_REPARSE_POINT) != 0)
			continue;

		DirSkipVal skip = indexTimeDirectoryFilteringSkip(path.mid(2)+finfo.fileName()+"/");
		if(SkipAll == skip)
			{
			skippedDirCount++;
			continue;
			}
		bool doFilesBelow = (SkipNone == skip);
		Subuse = Subno;
		makeid(SUBDIR);
		if(mainWindowPtr->incfile_and_check_for_error(true))
			return;
// The 2+ here skips over drive letter and will be wrong if UNC indexing is implemented
		SubDir[Subno] = path.mid(2+offset_to_saved_dir) + finfo.fileName();
		SubDisk[Subno] = Diskno;
		if(mainWindowPtr->incsub_and_check_for_error())
			return;
		index_disk(path + finfo.fileName() + "/", doFilesBelow);
		vrec_i->Vsubdir++;
		}
	}

// This may be called early on to cancel indexing or as the last
// thing after finishing indexing of a disk.
void IndexWindow::indexAbort()
	{
// Re-enable buttons that were disabled during indexing
	ui->pushButton_saveTheseSettings->setEnabled(true);
	ui->pushButton_browseDir->setEnabled(true);
	ui->pushButton_configExclusions->setEnabled(true);
	ui->pushButton_extractConfigure->setEnabled(true);
	ui->pushButton_index1->setEnabled(true);
	ui->pushButton_index2->setEnabled(true);
	ui->pushButton_done->setEnabled(true);
	InhibitQuit = false;

	index_button[currentIndexingButton]->setFocus();
	slot_setIndexButtonLabels();
	Q_EMIT signal_indexIsDone();
	}

void IndexWindow::indexFinish()
	{
	Stop_indexing = true;
	countUpdateTimer->stop();
	slot_updateDbTotals();
	if(IndexTimeFiltering)
		{
		Q_EMIT signal_logMessage(tr("%1 files were skipped by index-time editing").arg(skippedFileCount));
		Q_EMIT signal_logMessage(tr("%1 files would have been skipped if not explicitly included").arg(almostSkippedFileCount));
		Q_EMIT signal_logMessage(tr("%1 directories (and contents) were skipped by index-time editing").arg(skippedDirCount));
		Q_EMIT signal_logMessage(tr("%1 directories would have been skipped if not explicitly included").arg(almostSkippedDirCount));
		}
	if(EjectAfterIndexing)
		{
		ejectNT(static_cast<TCHAR>(drive_letter[currentIndexingButton]->currentText().at(0).toLatin1()));
		}
	currentIndexingButton = 1 - currentIndexingButton;
	indexAbort();
	}

void IndexWindow::open_indexing_message(struct diskrec *currentVrec)
	{
	IndexingDialog *id = new IndexingDialog(this, mlptr, currentVrec, diskAnimation);
	id->setAttribute(Qt::WA_DeleteOnClose);
	id->show();
	}

void IndexWindow::setIndexButtonColors(bool indexing)
	{
// Note: the Done button gets a similar style in the form definition
	if(indexing)
		{
		index_button[currentIndexingButton]->setStyleSheet(
			"margin: 0px;"
				"border: 1px solid rgb(112,112,112);"
				"border-radius: 2px;"
				"padding: 3px;"
				"background: qlineargradient(x1:0, y1:0, x2:0, y2:1,"
					"stop: 0.0 rgb(255,165,0),"
					"stop: 0.5 rgb(246,160,0),"
					"stop: 0.51 rgb(240,156,0),"
					"stop: 1.0 rgb(229,148,0));");
		}
	else
		{
		index_button[currentIndexingButton]->setStyleSheet(
			"margin: 0px;"
				"border: 1px solid rgb(112,112,112);"
				"border-radius: 2px;"
				"padding: 3px;"
				"background: qlineargradient(x1:0, y1:0, x2:0, y2:1,"
					"stop: 0.0 rgb(0,255,0),"
					"stop: 0.5 rgb(0,244,0),"
					"stop: 0.51 rgb(0,238,0),"
					"stop: 1.0 rgb(0,231,0));");
		}
	index_button[1-currentIndexingButton]->setStyleSheet(
		"margin: 0px;"
			"border: 1px solid rgb(112,112,112);"
			"border-radius: 2px;"
			"padding: 3px;"
			"background: qlineargradient(x1:0, y1:0, x2:0, y2:1,"
				"stop: 0.0 rgb(222,220,214),"
				"stop: 0.5 rgb(215,213,207),"
				"stop: 0.51 rgb(209,207,201),"
				"stop: 1.0 rgb(198,196,190));");
	}

void IndexWindow::slot_setIndexButtonLabels()
	{
	index_button[0]->setText(tr("  &Index next disk from drive ")
							+ drive_letter[0]->currentText() + "  ");
	index_button[1]->setText(tr("  Index &next disk from drive ")
							+ drive_letter[1]->currentText() + "  ");
	setIndexButtonColors(false);
	}

void IndexWindow::slot_buttonDoneClicked()
	{
	QSettings settings("WSS-DDC", "Wssindex\\Geometry");
	settings.setValue("IndexGeometryPos", pos());
	settings.setValue("IndexGeometrySize", size());
	close();
	}

void IndexWindow::slot_buttonsaveAndDoneClicked()
	{
	slot_buttonSaveTheseSettings();
	slot_buttonDoneClicked();
	}

void IndexWindow::slot_buttonStopClicked()
	{
	Stop_indexing = true;
	}

// If Index_starting-dir starts with Letter: (as might happen if a path
// is pasted in), remove it
void IndexWindow::slot_fixupStartingDir()
	{
	if(ui->lineEdit_startdir->text().length() < 2)
		return;
	if(QChar::fromLatin1(':') == ui->lineEdit_startdir->text().at(1))
		ui->lineEdit_startdir->setText(ui->lineEdit_startdir->text().remove(0, 2));
	}

void IndexWindow::slot_indexButtonClicked(int button)
	{
	currentIndexingButton = button;
//	open_indexing_message();
	bool is_mounted_volume = false;
	bool StoreRelativePath_use = Store_relative_path;
	if(ui->checkBox_useExclusions->isChecked() &&
		(!IndexTimeFileExclude || 0 == IndexFileExcludeList.size()) &&
			(!IndexTimeDirExclude || 0 == IndexDirExcludeList.size()))
		{
		QMessageBox msgBox;
		msgBox.setText(tr("<html><head/><body><p>Index-time editing is checked, but the current settings "
								"will not exclude any files.<p>"
				"If you continue, the result will be the same as if this checkbox "
				"was not checked.</p></body></html>"));
		msgBox.addButton(tr("Go ahead and index"), QMessageBox::ActionRole);
		QPushButton *cancelButton = msgBox.addButton(tr("Cancel"), QMessageBox::ActionRole);
		msgBox.exec();
		if(msgBox.clickedButton() == cancelButton)
			{
			indexAbort();
			return;
			}
		}
	IndexTimeFiltering = ui->checkBox_useExclusions->isChecked();

// Don't want user pushing buttons while indexing
	ui->pushButton_saveTheseSettings->setDisabled(true);
	ui->pushButton_browseDir->setDisabled(true);
	ui->pushButton_configExclusions->setDisabled(true);
	ui->pushButton_extractConfigure->setDisabled(true);
	ui->pushButton_index1->setDisabled(true);
	ui->pushButton_index2->setDisabled(true);
	ui->pushButton_done->setDisabled(true);
	InhibitQuit = true;

	skippedDirCount = almostSkippedDirCount = skippedFileCount = almostSkippedFileCount = 0;
	QString disk;
	if(mainWindowPtr->incdisk_error_check())
		{
		indexAbort();
		return;
		}
	disk = drive_letter[currentIndexingButton]->currentText();
	WCHAR mounted_volname[50];
	if(!ui->lineEdit_startdir->text().isEmpty())
// If a starting directory is specified, it may be an NTFS mount point.  To detect this:
// 1. Call GetVolumeNameForVolumeMountPointW, making sure path includes trailing slash.
// Non-zero return code means it is a mount point.  The volume id returned lookes like
// \\?\Volume{11e5481f-849d-11df-abe7-485b39cada34}\ (funny text is GUID)
// 2. Call GetVolumeInformationW with this volume id to get the real volume name
// 3. Get used and free space by calling GetDiskFreeSpaceExW with starting path
// GetVolumeInformationW fails if the mount point is on a mapped drive.
		{
		if('\\' != ui->lineEdit_startdir->text().at(0) &&
			'/' != ui->lineEdit_startdir->text().at(0))
			{
			ui->lineEdit_startdir->setText(QDir::separator() + ui->lineEdit_startdir->text());
			}
		if("\\" != ui->lineEdit_startdir->text().right(1) &&
			"/" != ui->lineEdit_startdir->text().right(1))
			{
			ui->lineEdit_startdir->setText(ui->lineEdit_startdir->text() + QDir::separator());
			}
		ui->lineEdit_startdir->setText(QDir::toNativeSeparators(ui->lineEdit_startdir->text()));
		if(!QFileInfo(disk + ":" + ui->lineEdit_startdir->text()).isDir())
			{
			QMessageBox::critical(this, tr("Error"), tr("Index starting directory doesn't exist on %1:").arg(disk));
			indexAbort();
			return;
			}
		QString tv = disk + ":" + ui->lineEdit_startdir->text();
		WCHAR tp[MAX_PATH];
		int len = tv.toWCharArray(tp);
		tp[len] = 0;
		is_mounted_volume = (0 != GetVolumeNameForVolumeMountPointW(tp, mounted_volname, 50));
		if(is_mounted_volume)
			{
			StoreRelativePath_use = true;
			int drivetype = static_cast<int>(GetDriveType(tp));
			if(DRIVE_REMOTE == drivetype)
				{
				Q_EMIT signal_logMessage(tr("Starting directory points to a mounted disk"
					" on a remote volume.  Label and space information are not accessible."));
				is_mounted_volume = false;
				}
			else
				{
				Q_EMIT signal_logMessage(tr("Starting directory points to a mounted disk."
					"  This disk will be indexed as itself, not as a subdirectory of %1.").arg(disk));
				}
			}
		}
	if(!is_mounted_volume)
		{
		QString tv = disk + ":\\";
		int len = tv.toWCharArray(mounted_volname);
		mounted_volname[len] = 0;
		}
	index_button[currentIndexingButton]->setText(tr("  Indexing drive ")
							+ disk + "  ");
	setIndexButtonColors(true);
	NofileBeforeIndexing = NoPlainFiles;
	SubnoBeforeIndexing = Subno;

	TCHAR volbuffer[MAX_PATH+1];
	struct diskrec *vrec_i = Vrec + Diskno;
	if(!GetVolumeInformationW(mounted_volname, volbuffer, MAX_PATH+1, nullptr, nullptr, nullptr, nullptr, 0)
		|| !GetDiskFreeSpaceExW(mounted_volname, nullptr, reinterpret_cast<_ULARGE_INTEGER *>(&vrec_i->Vbytes),
										reinterpret_cast<_ULARGE_INTEGER *>(&vrec_i->Vfree)))
		{
		QMessageBox::critical(this, tr("Error"), tr("Unable to read drive %1").arg(disk));
		indexFinish();
		return;
		}
	if(ui->radioButton_trueLabel->isChecked())
		{
		vrec_i->VolName = QString::fromUtf16(reinterpret_cast<ushort*>(volbuffer));
		if(vrec_i->VolName.isEmpty())
			{
			bool ok;
			QString text = QInputDialog::getText(this, tr("Unlabelled Volume"),
				  tr("Volume is unlabelled, enter label to use:"), QLineEdit::Normal,
																  tr(""), &ok, MyWindowFlags | Qt::MSWindowsFixedSizeDialogHint);
			if (ok && !text.isEmpty())
				vrec_i->VolName = text;
			else
				{
				indexFinish();
				return;
				}
			}
		}
	else if(ui->radioButton_promptLabel->isChecked())
		{
		bool ok;
		QString text = QInputDialog::getText(this, tr("Prompt for volume requested"),
			tr("Enter label to use (default is the true label):"), QLineEdit::Normal,
			QString::fromUtf16(reinterpret_cast<ushort*>(volbuffer)), &ok, MyWindowFlags | Qt::MSWindowsFixedSizeDialogHint);
		if (ok && !text.isEmpty())
			vrec_i->VolName = text;
		else
			{
			indexFinish();
			return;
			}
		}
	else
		{
		vrec_i->VolName = ui->lineEdit_serString->text().arg(ui->spinBox_serNum->value());
		ui->spinBox_serNum->stepUp();
		}
	Nfold = Nofile;
	duplicate_volname = check_duplicate_volname(vrec_i->VolName);
	if(DONT_REPLACE == duplicate_volname)
		{
		indexFinish();
		return;
		}
	Q_EMIT signal_logMessage(tr("Indexing %1").arg(vrec_i->VolName));
	vrec_i->Vdate = QDate::currentDate();
	vrec_i->Vfiles = 0;
	vrec_i->Vsubdir = 0;
	Subuse = 0;
	makeid(VOLUME);
	mainWindowPtr->statusChanged(true);
	if(mainWindowPtr->incfile_and_check_for_error(true))
		return;
	Stop_indexing = false;
	slot_updateDbTotals();
	QueryOverride = false;
	offset_to_saved_dir = 0;
	bool doFiles = (SkipNone == indexTimeDirectoryFilteringSkip("/"));
	if(ui->lineEdit_startdir->text().isEmpty())
		{
		diskAnimation = getDiskType(disk + ":/");
		open_indexing_message(vrec_i);
		index_disk(disk + ":/", doFiles);
		}
	else
		{
		QString abspath = QDir::cleanPath(QDir::fromNativeSeparators(ui->lineEdit_startdir->text()));
// Relative path is stored if configured or if a mounted volume
		if(StoreRelativePath_use)
			{
			offset_to_saved_dir = abspath.size();
			Q_EMIT signal_logMessage(tr("Indexing starts from %1 "
				"but only path relative to starting directory will be stored.").arg(abspath));
			diskAnimation = getDiskType(disk + ":"	+ abspath + "/");
			open_indexing_message(vrec_i);
			index_disk(disk + ":"	+ abspath + "/", doFiles);
			}
		else
			{
			Subuse = Subno;
			makeid(SUBDIR);
			if(mainWindowPtr->incfile_and_check_for_error(true))
				return;
			SubDir[Subno] = abspath;
			Q_EMIT signal_logMessage(tr("Indexing starts from %1").arg(abspath));
			SubDisk[Subno] = Diskno;
			if(mainWindowPtr->incsub_and_check_for_error())
				return;
			diskAnimation = WSS_MOUNTED;
			open_indexing_message(vrec_i);
			index_disk(disk + ":" + abspath + "/", doFiles);
			}
	}
// ***** If incfile or incsub above return true, Diskno doesn't get incremented
// ***** Nofile-- if incsub fails???
	Diskno++;
	Q_EMIT signal_diskListChanged();
	switch(duplicate_volname)
		{
		case NOT_DUPLICATE:
			break;
		case REPLACE:
			delete_disk(oldVolno);
			break;
		case REPLACE_AND_COPY:
			copy_and_delete(oldVolno);
			break;
		}

	indexFinish();
	}

void IndexWindow::slot_updateDbTotals()
	{
// Subtract 1 from Subno so root directory isn't counted
// This subtraction has not been done for SubnoBeforeIndexing, so
// also isn't needed when computing subs on current disk
	ui->files_db->setText(tr("%L1").arg(NoPlainFiles));
	ui->subdirs_db->setText(tr("%L1").arg(Subno-1));
	ui->disks_db->setText(tr("%L1").arg(Diskno));
	ui->files_current->setText(tr("%L1").arg(NoPlainFiles-NofileBeforeIndexing));
	ui->subdirs_current->setText(tr("%L1").arg(Subno-SubnoBeforeIndexing));
	if(!Stop_indexing)
		{
		if(!countUpdateTimer->isActive())
			countUpdateTimer->start(200);
		}
	qApp->processEvents();
	}

enum disktypes IndexWindow::getDiskType(QString rootPath)
	{
	WCHAR rp[20];
	rootPath.toWCharArray(rp);
	unsigned int dtype = GetDriveType(rp);
//	qDebug() << "Drive type is" << dtype;
	switch(dtype)
		{
		default:
		case DRIVE_UNKNOWN:		// 0
		case DRIVE_NO_ROOT_DIR:	// 1
		case DRIVE_FIXED:			// 3 USB key may get this code
			return WSS_HARD;
		case DRIVE_REMOVABLE:	// 2 Could be floppy, usbkey, flash drive
			if(diskSize < 2000000)
				return WSS_FLOPPY;
			else
				return WSS_USB;
		case DRIVE_REMOTE:		// 4
			return WSS_NETWORK;
		case DRIVE_CDROM:			// 5
			return WSS_CDROM;
		case DRIVE_RAMDISK:		// 6
			return WSS_RAMDISK;
		}
	}
