// This file is part of the disk cataloger Wssindex.  Wssindex is copyright
// 1985-2021 by Robert W. Babcock and WSS Division of DDC.  Effective 2010,
// Wssindex is distributed under the terms of the GNU General Public
// License.  Contact the author at wssddc@wssddc.com or see the project web
// page at <https://www.wssddc.com/wssindex/>.
//
// Wssindex uses the open source Qt toolkit from the Qt Project (qt-project.org)
// and The Qt Company (www.qt.io).
//
// Wssindex is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the
// Free Software Foundation, either version 3 of the License, or (at your
// option) any later version.
//
// Wssindex is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with Wssindex.  If not, see <https://www.gnu.org/licenses/>.

#include "OtherWssi/wssindex.h"
#include "Dialogs/mainwindow.h"
#include "ui_mainwindow.h"
#include "Dialogs/textviewer.h"
#include "ui_textviewer.h"
#include "OtherWssi/globalsettings.h"
#include "OtherWssi/tablelayout.h"

#include <QApplication>
#include <QtDebug>

// ------------------------------ Event handlers ------------------------------
void MainWindow::closeEvent(QCloseEvent *event)
	{
// InhibitQuit is set if indexing is in progress.
	if(InhibitQuit)
		{
		event->ignore();
		return;
		}
	if(Changed)
		{
		QMessageBox msgBox;
		msgBox.setText(tr("<html><head/><body><p>The database has been modified.<p>"
			"Do you want to save your changes before exiting?</p></body></html>"));
		msgBox.setStandardButtons(QMessageBox::Save | QMessageBox::Discard | QMessageBox::Cancel);
		msgBox.setDefaultButton(QMessageBox::Cancel);
		int ret = msgBox.exec();

		switch (ret)
			{
			case QMessageBox::Save:
				if(slot_saveClicked())
					break;
				Q_FALLTHROUGH();
			case QMessageBox::Cancel:
				event->ignore();
				return;
			case QMessageBox::Discard:
				break;
			default:
// should never be reached
				break;
			}
		event->accept();
		}
	event->accept();
	}

void MainWindow::keyPressEvent(QKeyEvent *event)
	{
	switch(event->key())
		{
// Control-F or control-S trigger simple query dialog if a database is
// loaded (Nofile >0).  Simply calling slot_pushButtonQueryClicked()
// works, but it seems safer to queue a connection instead.
		case Qt::Key_F:
		case Qt::Key_S:
			if(Nofile > 0 && (Qt::ControlModifier & event->modifiers()))
				{
				QMetaObject::invokeMethod(this, "slot_pushButtonQueryClicked",
												Qt::QueuedConnection);
				event->ignore();
				}
			return;
// Control-D means copy previous non-empty cell, but only if selected cell
// is one of the fields where this makes sense.  If displaying as multiple
// bunches, do not look outside the current bunch.
		case Qt::Key_D:
			if(0 == Nofile || !(Qt::ControlModifier & event->modifiers()))
				return;
			{
			int row = ui->tableView->currentIndex().row();
			if(row <= 0)
				return;
			int column = ui->tableView->currentIndex().column();
			int itemrow = row % rowsPerItem;
			unsigned long thisSortindx = static_cast<unsigned long>(row/rowsPerItem) + file_start;
			unsigned long filenum = Sortindx[thisSortindx];
			struct filerec *dr_k = Dr + filenum;
			if(itemrow == categoryRow && column == categoryCol)
				{
				for(unsigned long i=1; i<=thisSortindx-file_start; i++)
					{
					unsigned long j = Sortindx[thisSortindx-i];
					if(!(HAS_CATEGORY & Dr[j].Diskbits))
						continue;
					dr_k->Category = Dr[j].Category;
					setCategoryEdited(dr_k);
					setHasCategory(dr_k);
					statusChanged(true);
					Ftm->refreshData();
					event->ignore();
					return;
					}
				return;
				}
			if(itemrow == commentRow && column == commentCol)
				{
				for(unsigned long i=1; i<=thisSortindx; i++)
					{
					unsigned long j = Sortindx[thisSortindx-i];
					if(!(HAS_COMMENT & Dr[j].Diskbits))
						continue;
					dr_k->Comment = Dr[j].Comment;
					setCommentEdited(dr_k);
					setHasComment(dr_k);
					statusChanged(true);
					Ftm->refreshData();
					event->ignore();
					return;
					}
				return;
				}
			if(itemrow == extraRow && column == extraCol)
				{
				for(unsigned long i=1; i<=thisSortindx; i++)
					{
					unsigned long j = Sortindx[thisSortindx-i];
					if(!(HAS_EXTRA_INFO & Dr[j].Diskbits))
						continue;
					dr_k->ExtraInfo = Dr[j].ExtraInfo;
					setExtraEdited(dr_k);
					setHasExtra(dr_k);
					statusChanged(true);
					Ftm->refreshData();
					event->ignore();
					return;
					}
				return;
				}
			}
			return;
		default:
			return;
		}
	}

void MainWindow::resizeEvent(QResizeEvent * /* event */)
	{
	resizeTimer->stop();
	resizeTimer->start(RESIZING_DELAY_MS);
	}

void MainWindow::slot_resizeTimeout()
	{
	resizeTimer->stop();
	fileBalance->colWidthsValid = false;
	fileBalance->doBalance();
	}

void MainWindow::slot_keyToTable(int key, int modifier)
	{
	QApplication::sendEvent(ui->tableView,
		new QKeyEvent(QEvent::KeyPress, key, static_cast<Qt::KeyboardModifier>(modifier)));
	QApplication::sendEvent(ui->tableView,
		new QKeyEvent(QEvent::KeyRelease, key, static_cast<Qt::KeyboardModifier>(modifier)));
	}

void MainWindow::slot_wheelToTable(QPoint pos, int delta, Qt::MouseButtons buttons,
									Qt::KeyboardModifiers modifiers, Qt::Orientation orientation)
	{
	QApplication::sendEvent(ui->tableView->viewport(),
		new QWheelEvent(pos, delta, buttons, modifiers, orientation));
	}

// This allows clicking on some text in the about screen
// to trigger special actions.
void MainWindow::slot_slotCaller(QUrl url)
	{
	if(0 == url.toString().compare("slotcaller:://QtAbout"))
		{
		QApplication::aboutQt();
		return;
		}
	if(0 == url.toString().compare("slotcaller:://Gpl3"))
		{
		gpl3License();
		return;
		}
	if(0 == url.toString().compare("slotcaller:://OtherLicenses"))
		{
		otherLicense();
		return;
		}
	}

void MainWindow::slot_recentFileMenuOpened()
	{
	for(int i=1; i<NrecentFiles; i++)
		recentFileOpenActs[i]->slot_unhovered();
	recentFileOpenActs[0]->pinWidget->setFocus();
	}

void MainWindow::slot_recentDirMenuOpened()
	{
	for(int i=1; i<NrecentDirs; i++)
		recentDirOpenActs[i]->slot_unhovered();
	recentDirOpenActs[0]->pinWidget->setFocus();
	}

void MainWindow::slot_recentFileTBOpened()
	{
	for(int i=1; i<NrecentFiles; i++)
		recentFileOpenTBActs[i]->slot_unhovered();
	recentFileOpenTBActs[0]->pinWidget->setFocus();
	}

void MainWindow::slot_recentDirTBOpened()
	{
	for(int i=1; i<NrecentDirs; i++)
		recentDirOpenTBActs[i]->slot_unhovered();
	recentDirOpenTBActs[0]->pinWidget->setFocus();
	}

void MainWindow::slot_saveRecentDirMenuOpened()
	{
	for(int i=1; i<NrecentDirs; i++)
		recentDirSaveActs[i]->slot_unhovered();
	recentDirSaveActs[0]->pinWidget->setFocus();
	}

void MainWindow::slot_saveRecentDirTBOpened()
	{
	for(int i=1; i<NrecentDirs; i++)
		recentDirSaveTBActs[i]->slot_unhovered();
	recentDirSaveTBActs[0]->pinWidget->setFocus();
	}
