// This file is part of the disk cataloger Wssindex.  Wssindex is copyright
// 1985-2021 by Robert W. Babcock and WSS Division of DDC.  Effective 2010,
// Wssindex is distributed under the terms of the GNU General Public
// License.  Contact the author at wssddc@wssddc.com or see the project web
// page at <https://www.wssddc.com/wssindex/>.
//
// Wssindex uses the open source Qt toolkit from the Qt Project (qt-project.org)
// and The Qt Company (www.qt.io).
//
// Wssindex is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the
// Free Software Foundation, either version 3 of the License, or (at your
// option) any later version.
//
// Wssindex is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with Wssindex.  If not, see <https://www.gnu.org/licenses/>.

#include "OtherWssi/wssindex.h"
#include "Dialogs/mainwindow.h"
#include "ui_mainwindow.h"
#include "OtherWssi/globalsettings.h"
#include "Dialogs/tabledesigndialog.h"
#include "OtherWssi/tablelayout.h"

#include <QScrollBar>
#include <QtDebug>

void MainWindow::layout_default()
	{
	rowsPerItem = 1;
	colsPerItem = 5;

	for(int i=0; i<NFIELDS; i++)
		{
		fieldlayout[i].enabled = false;
		fieldlayout[i].row = 0;
		fieldlayout[i].height = 1;
		fieldlayout[i].width = 1;
		fieldlayout[i].alignment = justEnumFromInt(Qt::AlignLeft | Qt::AlignVCenter);
		}

// long file name
	fieldlayout[0].column = 0;
	fieldlayout[0].enabled = true;

// size
	fieldlayout[5].column = 1;
	fieldlayout[5].enabled = true;

// timestamp
	fieldlayout[6].column = 2;
	fieldlayout[6].enabled = true;

// volume
	fieldlayout[2].column = 3;
	fieldlayout[2].enabled = true;
	volumeRow = 0;
	volumeCol = 3;

// subdirectory
	fieldlayout[3].column = 4;
	fieldlayout[3].enabled = true;

	categoryRow = -1;
	commentRow = -1;
	extraRow = -1;
	}

void MainWindow::slot_pushButtonDesignClicked()
	{
	tDesignAction->setDisabled(true);
	tDesignActionTB->setDisabled(true);
	TableDesignDialog *t = new TableDesignDialog(this);
	t->setAttribute(Qt::WA_DeleteOnClose);
	t->show();
	tDesignAction->setEnabled(true);
	tDesignActionTB->setEnabled(true);
	}

void MainWindow::slot_setTableHeaders(bool update_display)
	{
	int celluse[MAXROWS][MAXCOLS];
	for(int i=0; i<MAXROWS; i++)
		for(int j=0; j<MAXCOLS; j++)
			celluse[i][j] = 0;
	for(int i=0; i<NFIELDS; i++)
		{
		if(!fieldlayout[i].enabled)
			continue;
		for(int k=fieldlayout[i].column;
			k<fieldlayout[i].column+fieldlayout[i].width; k++)
			{
			if(k>=MAXCOLS)
				continue;
			for(int j=fieldlayout[i].row;
				j<fieldlayout[i].row+fieldlayout[i].height; j++)
				{
				if(j>=MAXROWS)
					continue;
				celluse[j][k]++;
				}
			}
		}

	unused_cells.clear();
	for(int i=0; i<rowsPerItem; i++)
		for(int j=0; j<colsPerItem; j++)
			{
			if(0 == celluse[i][j])
				{
				struct wssicell tcell;
				tcell.row = i;
				tcell.column = j;
				unused_cells.push_back(tcell);
				}
			}

//	QString theader[colsPerItem];		// gcc allows this but MS does not
	QString *theader = new QString[static_cast<qulonglong>(colsPerItem)];
	for(int i=0; i<colsPerItem; i++)
		theader[i] = "";

	for(int i=0; i<rowsPerItem; i++)
		{
		for(int j=0; j<NFIELDS; j++)
			{
			if(!fieldlayout[j].enabled)
				continue;
			if(i == fieldlayout[j].row)
				{
				if(theader[fieldlayout[j].column].size() > 0)
					theader[fieldlayout[j].column] += "\n";
				theader[fieldlayout[j].column] += fieldlayout[j].fieldname;
				}
			}
		}

// The table width isn't defined (or at least isn't right) until the table
// has been displayed, so force display.  Otherwise, column widths at startup
// are absurd.  ensurePolished() seems like the right way to do this, but it
// doesn't seem to work.
	Ftm->setFields();
	this->show();
	for(int i=0; i<colsPerItem; i++)
		Ftm->setHeaderData(i, Qt::Horizontal, theader[i]);
	delete [] theader;
	ui->tableView->horizontalHeader()->setStretchLastSection(true);
	ui->tableView->verticalHeader()->setStretchLastSection(false);
	if(update_display)
		slot_updateDisplay(true);
	}

void MainWindow::set_field_names()
	{
	fieldlayout[0].fieldname = tr("Filename");
	fieldlayout[1].fieldname = tr("Extension");
	fieldlayout[2].fieldname = tr("Volume");
	fieldlayout[3].fieldname = tr("Subdirectory");
	fieldlayout[4].fieldname = tr("Archive");
	fieldlayout[5].fieldname = tr("Size");
	fieldlayout[6].fieldname = tr("Timestamp");
	fieldlayout[7].fieldname = tr("Category");
	fieldlayout[8].fieldname = tr("Comment");
	fieldlayout[9].fieldname = tr("Extra info");
	}

void MainWindow::set_editable_cells()
	{
	volumeRow = volumeCol = categoryRow = categoryCol =
		commentRow = commentCol = extraRow = extraCol = -1;
	if(fieldlayout[2].enabled)
		{
		volumeRow = fieldlayout[2].row;
		volumeCol = fieldlayout[2].column;
		}
	if(fieldlayout[7].enabled)
		{
		categoryRow = fieldlayout[7].row;
		categoryCol = fieldlayout[7].column;
		}
	if(fieldlayout[8].enabled)
		{
		commentRow = fieldlayout[8].row;
		commentCol = fieldlayout[8].column;
		}
	if(fieldlayout[9].enabled)
		{
		extraRow = fieldlayout[9].row;
		extraCol = fieldlayout[9].column;
		}
	}

void MainWindow::slot_rowHeightSpinboxChanged(int newvalue)
	{
	File_row_height = newvalue;
	ui->tableView->verticalHeader()->setDefaultSectionSize(File_row_height);
	ui->tableView->scrollTo(ui->tableView->model()->
			index(topVisibleRow,0), QAbstractItemView::PositionAtTop);
//	Toggling the width of the verticalHeader forces a redraw
	ui->tableView->verticalHeader()->setFixedWidth(4);
	ui->tableView->verticalHeader()->setFixedWidth(5);
// Don't want the spinbox to have focus in case user hits home/end
	ui->tableView->setFocus();
	}

void MainWindow::slot_rowTracker(int row)
	{
// Record what row is at the top when scroll value changes
	topVisibleRow = row;
	}
