// This file is part of the disk cataloger Wssindex.  Wssindex is copyright
// 1985-2021 by Robert W. Babcock and WSS Division of DDC.  Effective 2010,
// Wssindex is distributed under the terms of the GNU General Public
// License.  Contact the author at wssddc@wssddc.com or see the project web
// page at <https://www.wssddc.com/wssindex/>.
//
// Wssindex uses the open source Qt toolkit from the Qt Project (qt-project.org)
// and The Qt Company (www.qt.io).
//
// Wssindex is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the
// Free Software Foundation, either version 3 of the License, or (at your
// option) any later version.
//
// Wssindex is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with Wssindex.  If not, see <https://www.gnu.org/licenses/>.

#include "OtherWssi/wssindex.h"
#include "Dialogs/mainwindow.h"
#include "ui_mainwindow.h"
#include "OtherWssi/globalsettings.h"
#include "OtherWssi/tablelayout.h"
#include "OtherWssi/wssiproto.h"
#include "ModelView/filetablemodel.h"
#include "Dialogs/balancecolumns.h"
#include "3rdparty/helpmenustyle.h"
#include "3rdparty/windowsmodernstyle.h"
#include "Workarounds/submenuproxy.h"
#include "ModelView/autotooltipdelegate.h"

#include <QDesktopWidget>
#include <QRect>
#include <QScrollBar>
#include <QStyleFactory>
#include <QtDebug>
#include <QStyleFactory>

MainWindow::MainWindow(QWidget *parent)
	 : QMainWindow(parent), ui(new Ui::MainWindow)
	{
	mainWindowPtr = this;
	ui->setupUi(this);
	this->setAcceptDrops(true);
	resizeTimer = new QTimer(this);
// In appStyleTemplate
//   %1 is the groupbox border color
//   %2 is the tooltip font size, 15 is a good default
//   %3 is the status bar font size, 10 is a good default
	appStyleTemplate =
				"QGroupBox"
					"{"
					"border: 1px solid %1;"
					"margin-top: 2ex; padding-top: 2ex;"
					"}"
				"QGroupBox::title"
					"{"
					"margin: 0 5 0 5 px;"
					"subcontrol-origin: margin;"
					"subcontrol-position: top left;"
					"position: relative;"
					"left: 5px;"
					"}"
				"QTableView::item:focus"
					"{"
					"background-color: transparent; border: 1px solid #000000;"
					"}"
				"QColorDialog"
					"{"
					BLUE_WASH
					"}"
				"QMainWindow"
					"{"
					BLUE_WASH
					"}"
				"QInputDialog"
					"{"
					BLUE_WASH
					"}"
				"QMessageBox"
					"{"
					BLUE_WASH
					"}"
				"QMenuBar"
					"{"
					BLUE_WASH
					"}"
				"QStatusBar::item"
					"{"
					"border: 0px;"
					"}"
				"QStatusBar QLabel"
					"{"
					"background: white;"
					"margin-left: 2px; margin-right: 2px;"
					"font: %3px;"
					"}"
				"QToolTip"
					"{"
					"font-size: %2px;"
					"}"
				;
	toolbarStyle =
				"QToolBar"
					"{"
					BLUE_WASH
					"margin-top: 2px; margin-bottom: 0px; padding: 0px;"
					"border-color: transparent;"
					"}"
				"QToolBar:separator"
					"{"
					BLUE_WASH
					"width: 20px;"
					"}"
// These settings interact with those in raise2::eventFilter() when hovering over buttons
//				"QToolButton"
					"{"
					BLUE_WASH
					"border-color: black;"
					"border-left-width: 1px; border-right-width: 1px; border-bottom-width: 0px; border-top-width: 0px;"
					"padding: 1px; margin: 0px;"
					"}"
				;
	update_display_on_return_to_main = false;
	settings_dialog = nullptr;
	Max_threads = 1;
	threadsafeInit();
	recentFiles = new pinList(MAX_RECENT_FILES);
	recentDirs = new pinList(MAX_RECENT_DIRS);
	restoreSettings();
	VistaStyle = QStyleFactory::create("WindowsVista");
	XPStyle = QStyleFactory::create("Windows");
	menuBarStyle = new helpMenuStyle();
	setAppStyle();
	QLocale currentLocale = QLocale();
	ShortDateFormat = currentLocale.QLocale::dateFormat(QLocale::ShortFormat);
	collator = new QCollator();
	simpleCollator = new QCollator();
	collator->setCaseSensitivity(Ignore_case_sort ? Qt::CaseInsensitive : Qt::CaseSensitive);
	collator->setNumericMode(NaturalSort);
	simpleCollator->setCaseSensitivity(Qt::CaseInsensitive);
	simpleCollator->setNumericMode(false);
	set_field_names();	// Has to be after restoreSettings which clobbers field names

	Changed = InhibitQuit = false;
	topVisibleRow = 0;
	MaxfilesIncMultiplier = MaxsubsIncMultiplier = MaxdisksIncMultiplier = 1;

	statusMessage_loaded = new QLabel();
	statusMessage_db = new QLabel();
	statusMessage_files = new QLabel();
	statusMessage_selected = new QLabel();
	ui->statusBar->insertWidget(0, statusMessage_loaded);
	ui->statusBar->insertPermanentWidget(1, statusMessage_db);
	ui->statusBar->insertPermanentWidget(2, statusMessage_files);
	ui->statusBar->insertPermanentWidget(3, statusMessage_selected);
	statusNoDbLoaded();

	QDesktopServices::setUrlHandler("slotCaller", this, "slot_slotCaller");
	createMenus();
	createToolbars();
	updateRecentLists();
	enableMenusIfDbLoaded();
	fileBalance = new balanceColumns(ui->tableView);

	connect(ui->tableView->verticalScrollBar(), SIGNAL(valueChanged(int)),
			  this, SLOT(slot_rowTracker(int)));

//	ui->statusBar->showMessage(tr("No database loaded"));

	for(int i=0; i<NFIELDS; i++)
		headerSortDir[i] = 0;

	Ftm = new fileTableModel(this);
	slot_setTableHeaders(false);
	ui->tableView->setModel(Ftm);
	ui->tableView->setItemDelegate(new AutoToolTipDelegate(ui->tableView));
	fileSelection = ui->tableView->selectionModel();
	Vcm = new volComboModel();
	Vtm = new volTableModel(this);

// Narrow verticalHeader allows row resizing without showing numbers
	ui->tableView->verticalHeader()->setFixedWidth(5);
	ui->tableView->verticalHeader()->setDefaultSectionSize(fontMetrics().lineSpacing() + 10);
	ui->tableView->horizontalHeader()->setSectionsClickable(true);
	connect(ui->tableView->horizontalHeader(), SIGNAL(sectionClicked(int)),
			  this, SLOT(slot_tableClicked(int)));
	connect(resizeTimer, SIGNAL(timeout()), this, SLOT(slot_resizeTimeout()));

	searchModel = new QStringListModel();
	pastSearchesCompleter = new QCompleter();
	pastSearchesCompleter->setModel(searchModel);
	pastSearchesCompleter->setCaseSensitivity(Qt::CaseInsensitive);
	pastSearchesCompleter->setCompletionMode(QCompleter::PopupCompletion);
	pastSearchesCompleter->setFilterMode(Qt::MatchContains);
	pastSearchesCompleter->setMaxVisibleItems(25);

	int nargs = QCoreApplication::arguments().size();
	if(nargs > 1)
		openDbFromFile(QCoreApplication::arguments().at(nargs-1));

// Show the welcome window if it has never been shown before
	QSettings options("WSS-DDC", "Wssindex");
	bool welcomeShown = options.value("welcomeShown", false).toBool();
	if(!welcomeShown)
		slot_welcome();
	}

MainWindow::~MainWindow()
	{
	delete ui;
	}

// Anything that changes the list of disks in the database should send a
// signal to this slot.  Anything that depends on the disk list, such as
// pulldowns in some of the query menus, connects to signal_diskListChanged
// to find out when they need to update.
void MainWindow::slot_diskListChanged()
	{
	Q_EMIT signal_diskListChanged();
	}

void MainWindow::slot_updateDisplay(bool goToStart)
	{
	update_display_on_return_to_main = false;
	enableMenusIfDbLoaded();
	if(0 == Nofile)
		{
		no_files_in_db();
		return;
		}

	if(goToStart)
		{
		file_start = 0;
		ui->tableView->horizontalHeader()->setSortIndicatorShown(false);
		}
	ui->tableView->verticalHeader()->setDefaultSectionSize(File_row_height);
	unsigned long file_end = qMin(Noselect, file_start+static_cast<unsigned long>(Max_bunch));
	Ftm->setSizeAndValidity(static_cast<int>(file_end-file_start)*rowsPerItem, colsPerItem, file_start);
	if(!fileBalance->colWidthsValid)
		fileBalance->doBalance();

	enableMenusIfDbLoaded();
	firstActionTB->setEnabled(file_start > 0);
	previousActionTB->setEnabled(file_start > 0);
	nextActionTB->setEnabled(file_end < Noselect);
	lastActionTB->setEnabled(file_end < Noselect);
	fitWidthActionTB->setEnabled(Noselect > 0);
	ui->tableView->clearSpans();

	ui->tableView->scrollToTop();
	ui->tableView->setFocus();

	statusMessage_db->setText(tr("  %L1  ").arg(QDir::toNativeSeparators(database_name)));
	statusMessage_files->setText(tr("  %L1 files, %L2 directories, %L3 disks  ")
											.arg(NoPlainFiles).arg(Subno-1).arg(Diskno));
	if(Noselect > 0)
		{
		statusMessage_selected->setText(tr("  %L1 items selected, %L2-%L3 displayed  ")
												.arg(Noselect).arg(file_start+1).arg(file_end));
		}
	else
		{
		statusMessage_selected->setText(tr("  No items selected  "));
		}
	statusChanged(Changed);
//	ui->statusBar->clearMessage();
//	ui->statusBar->show();
	connect(ui->tableView->model(), SIGNAL(signal_fileEditCompleted(QModelIndex,QString)),
			  this, SLOT(slot_cellEdited(QModelIndex,QString)));
	}

// Display no files, either because no database is loaded or because
// all disks in the database have been deleted.
void MainWindow::no_files_in_db()
	{
	Ftm->setSizeAndValidity(0, 0, 0);
	Ftm->setFileModelValidity(false);
	Vtm->setVolModelValidity(false);

	ui->tableView->clearSpans();
	statusNoDbLoaded();
//	ui->statusBar->showMessage(tr("No database loaded"));
//	statusMessage_db->clear();
//	statusMessage_files->clear();
//	statusMessage_selected->clear();
	firstActionTB->setDisabled(true);
	previousActionTB->setDisabled(true);
	nextActionTB->setDisabled(true);
	lastActionTB->setDisabled(true);
	}

// Compute position of upper left corner of qrect so that center of rectangle is
// offset from center of screen by xOffset, yOffset.  Then, constrain so rectangle
// stays on screen.  Note that with multiple monitors, upper left of screen may
// not be (0,0).
QPoint MainWindow::offCenter(QRect qrect, int xOffset, int yOffset)
	{
	QRect mainScreen = QApplication::desktop()->screenGeometry(this->centralWidget());
	int newx = mainScreen.x() + (mainScreen.width() - qrect.width())/2 + xOffset;
	if(newx+qrect.width() > mainScreen.x() + mainScreen.width())
		newx = mainScreen.x() + mainScreen.width() - qrect.width();
	if(newx < mainScreen.x())
		newx = mainScreen.x();
	int newy = mainScreen.y() + (mainScreen.height() - qrect.height())/2 + yOffset;
	if(newy+qrect.height() > mainScreen.y() + mainScreen.height())
		newy = mainScreen.y() + mainScreen.height() - qrect.height();
	if(newy < mainScreen.y())
		newy = mainScreen.y();
	return QPoint(newx, newy);
	}

void MainWindow::setAppStyle()
	{
	if(UseModernStyle)
		{
		qApp->setStyle(new WindowsModernStyle());
		qApp->setStyleSheet(appStyleTemplate.arg(DARK_BORDER).arg(ToolTipFontSize).arg(StatusBarFontSize));
//		qDebug() << "Dark:" << appStyleTemplate.arg(DARK_BORDER).arg(ToolTipFontSize);
		}
	else
		{
//		qApp->setStyle(new SubMenuProxy());
		qApp->setStyle(QStyleFactory::create("WindowsVista"));
		qApp->setStyleSheet(appStyleTemplate.arg(LIGHT_BORDER).arg(ToolTipFontSize).arg(StatusBarFontSize));
//		qDebug() << "Light:" << appStyleTemplate.arg(LIGHT_BORDER).arg(ToolTipFontSize);
		}
	}

void MainWindow::statusNoDbLoaded()
	{
	statusMessage_loaded->setText(tr("No database loaded"));
	statusMessage_db->setText(tr("Database name"));
	statusMessage_files ->setText(tr("Database item count"));
	statusMessage_selected->setText(tr("Selected item count"));
	}

// Update changed/saved status and status line message
void MainWindow::statusChanged(bool isChanged)
	{
	Changed = isChanged;
	if(Changed)
		statusMessage_loaded->setText(tr("Unsaved changes"));
	else
		statusMessage_loaded->setText(tr("Ready"));
	}
