// This file is part of the disk cataloger Wssindex.  Wssindex is copyright
// 1985-2021 by Robert W. Babcock and WSS Division of DDC.  Effective 2010,
// Wssindex is distributed under the terms of the GNU General Public
// License.  Contact the author at wssddc@wssddc.com or see the project web
// page at <https://www.wssddc.com/wssindex/>.
//
// Wssindex uses the open source Qt toolkit from the Qt Project (qt-project.org)
// and The Qt Company (www.qt.io).
//
// Wssindex is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the
// Free Software Foundation, either version 3 of the License, or (at your
// option) any later version.
//
// Wssindex is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with Wssindex.  If not, see <https://www.gnu.org/licenses/>.

#include "OtherWssi/wssindex.h"
#include "Dialogs/rangequerydialog.h"
#include "ui_rangequerydialog.h"
#include "ui_mainwindow.h"
#include "Dialogs/mainwindow.h"
#include "OtherWssi/initialpos.h"
#include "OtherWssi/wssiproto.h"

RangeQueryDialog::RangeQueryDialog(QWidget *parent)
	: QMainWindow(parent), rq(new Ui::RangeQueryDialog)
	{
// Could use KiB, MiB, GiB here for binary size units, but who would understand?
	QString units[] =
		{tr("bytes"),
		 tr("KB (dec)"),
		 tr("KB (bin)"),
		 tr("MB (dec)"),
		 tr("MB (bin)"),
		 tr("GB (dec)"),
		 tr("GB (bin)")};

	rq->setupUi(this);
	this->setWindowFlags(MyWindowFlags);
	restoreSettings();
	rq->pushButton_search->setDefault(true);
	rq->radioButton_new_query->setChecked(true);
	rq->dateTimeEdit_latest->setDateTime(QDateTime::currentDateTime());

// Populate the size units controls
	for(int i=0; i<7; i++)
		{
		rq->comboBox_min_size_unit->insertItem(i+1, units[i], i);
		rq->comboBox_max_size_unit->insertItem(i+1, units[i], i);
		}

	connect(this, SIGNAL(signal_updateDisplay(bool)),
		mainWindowPtr, SLOT(slot_updateDisplay(bool)));

	connect(rq->pushButton_close, SIGNAL(clicked()),
		this, SLOT(slot_pushButtonCloseClicked()));
	connect(rq->pushButton_search, SIGNAL(clicked()),
		this, SLOT(slot_pushButtonSearchClicked()));

	connect(this, SIGNAL(signal_keyToTable(int,int)),
		mainWindowPtr, SLOT(slot_keyToTable(int,int)));
	connect(this, SIGNAL(signal_wheelToTable(QPoint,int,Qt::MouseButtons,Qt::KeyboardModifiers,Qt::Orientation)),
		mainWindowPtr, SLOT(slot_wheelToTable(QPoint,int,Qt::MouseButtons,Qt::KeyboardModifiers,Qt::Orientation)));
	connect(mainWindowPtr, SIGNAL(signal_close_queries()),
			  this, SLOT(close()));

// When any checkbox changes state, Query button may need to be enabled or disabled
	QList<QCheckBox*> checkboxes = this->findChildren<QCheckBox*>(QRegularExpression(QFA("checkBox*")));
	foreach (QCheckBox* checkbox, checkboxes)
		connect(checkbox, SIGNAL(stateChanged(int)), this, SLOT(slot_checkboxChanged(int)));
	this->setAttribute(Qt::WA_AlwaysShowToolTips);
	slot_checkboxChanged(0);
	}

RangeQueryDialog::~RangeQueryDialog()
	{
	delete rq;
	}

void RangeQueryDialog::restoreSettings()
	{
	QSettings settings("WSS-DDC", "Wssindex\\Geometry");
	if(settings.contains("RangeQuerySize"))
		{
		QSize size= settings.value("RangeQuerySize").toSize();
		resize(size);
		}
	if(settings.contains("RangeQueryPos"))
		{
		QPoint pos = settings.value("RangeQueryPos").toPoint();
		move(pos);
		forceOnscreen(this);
		}
	else
		{
// frameGeometry is the same as geometry until the dialog is displayed
		this->show();
		move(mainWindowPtr->offCenter(this->frameGeometry(), RANGE_OFFSET));
		}
	}

// ------------------------------ Action buttons ------------------------------
void RangeQueryDialog::slot_pushButtonCloseClicked()
	{
	QSettings settings("WSS-DDC", "Wssindex\\Geometry");
	settings.setValue("RangeQueryPos", pos());
	settings.setValue("RangeQuerySize", size());
	this->close();
	}

void RangeQueryDialog::slot_pushButtonSearchClicked()
	{
	if(0 == Diskno)
		return;
// Check for impossible max<min conditions being requested
	if(rq->checkBox_disk_num_from->isChecked() && rq->checkBox_disk_num_to->isChecked()
		&& rq->spinBox_disk_num_from->value() > rq->spinBox_disk_num_to->value())
		{
		QMessageBox::warning(this, tr("Impossible request"),
			tr("You've requested a range of disk numbers with the lower limit "
				"larger than the upper.  This can't match anything.  Please fix and retry."));
		return;
		}
	if(rq->checkBox_earliest->isChecked() && rq->checkBox_latest->isChecked()
		&& rq->dateTimeEdit_earliest->dateTime() > rq->dateTimeEdit_latest->dateTime())
			{
			QMessageBox::warning(this, tr("Impossible request"),
				tr("You've requested a time range with the lower limit "
					"after the upper.  This can't match anything.  Please fix and retry."));
			return;
			}
	if(rq->checkBox_min_size->isChecked() &&	rq->checkBox_max_size->isChecked()
		&& scaledsize(rq->spinBox_min_size->value(), rq->comboBox_min_size_unit->currentIndex()) >
			scaledsize(rq->spinBox_max_size->value(),	rq->comboBox_max_size_unit->currentIndex()))
			{
			QMessageBox::warning(this, tr("Impossible request"),
				tr("You've requested a minimum file size greater than the maximum file "
					"size.  This can't match anything.  Please fix and retry."));
			return;
			}

	for(int i=0; i<SEL_DIM; i++)
		Usesel[i] = false;
	Nselect = 0;
// If not a new query, mark which files were previously selected
	if(!rq->radioButton_new_query->isChecked())
		updatePreviouslySelected();

	if(rq->radioButton_add_to->isChecked())
		{
		Sel[Nselect++] = SEL_ADD_TO_PREVIOUS;
		Usesel[SEL_ADD_TO_PREVIOUS] = true;
		}

	if(rq->radioButton_pick_from->isChecked())
		{
		Sel[Nselect++] = SEL_FROM_PREVIOUS;
		Usesel[SEL_FROM_PREVIOUS] = true;
		}

	if(rq->radioButton_exclude_from->isChecked())
		{
		Sel[Nselect++] = SEL_REM_FROM_PREVIOUS;
		Usesel[SEL_REM_FROM_PREVIOUS] = true;
		}

	if(rq->checkBox_disk_num_from->isChecked())
		{
		Sel[Nselect++] = SEL_VOL_NUM_LOW_LIMIT;
		Usesel[SEL_VOL_NUM_LOW_LIMIT] = true;
		StartDiskNum = static_cast<unsigned long>(rq->spinBox_disk_num_from->value());
		}
	if(rq->checkBox_disk_num_to->isChecked())
		{
		Sel[Nselect++] = SEL_VOL_NUM_HI_LIMIT;
		Usesel[SEL_VOL_NUM_HI_LIMIT] = true;
		StopDiskNum = static_cast<unsigned long>(rq->spinBox_disk_num_to->value());
		}

	if(rq->checkBox_earliest->isChecked())
		{
		Sel[Nselect++] = SEL_DATE_LOW_LIMIT;
		Usesel[SEL_DATE_LOW_LIMIT] = true;
		Compd1 = rq->dateTimeEdit_earliest->dateTime();
		}
	if(rq->checkBox_latest->isChecked())
		{
		Sel[Nselect++] = SEL_DATE_HI_LIMIT;
		Usesel[SEL_DATE_HI_LIMIT] = true;
		Compd2 = rq->dateTimeEdit_latest->dateTime();
		}

	if(rq->checkBox_min_size->isChecked())
		{
		Sel[Nselect++] = SEL_SIZE_LOW_LIMIT;
		Usesel[SEL_SIZE_LOW_LIMIT] = true;
		Compsiz1 = scaledsize(rq->spinBox_min_size->value(),
			rq->comboBox_min_size_unit->currentIndex());
		}
	if(rq->checkBox_max_size->isChecked())
		{
		Sel[Nselect++] = SEL_SIZE_HI_LIMIT;
		Usesel[SEL_SIZE_HI_LIMIT] = true;
		Compsiz2 = scaledsize(rq->spinBox_max_size->value(),
			rq->comboBox_max_size_unit->currentIndex());
		}

	mainWindowPtr->doSelection();
	Q_EMIT signal_updateDisplay(true);
	}

quint64 RangeQueryDialog::scaledsize(int unscaled, int scale)
	{
	quint64 result = static_cast<quint64>(unscaled);
	switch(scale)
		{
		case 0:
			break;
		case 1:
			result *= Q_INT64_C(1000);
			break;
		case 2:
			result *= Q_INT64_C(1024);
			break;
		case 3:
			result *= Q_INT64_C(1000000);
			break;
		case 4:
			result *= Q_INT64_C(1048576);
			break;
		case 5:
			result *= Q_INT64_C(1000000000);
			break;
		case 6:
			result *= Q_INT64_C(1073741824);
			break;
		}
	return result;
	}

// ------------------------------ Event handlers ------------------------------
// Some checkbox has changed state.  If any are enabled, query button should be enabled
void RangeQueryDialog::slot_checkboxChanged(int /* newstate */)
	{
	QList<QCheckBox*> checkboxes = this->findChildren<QCheckBox*>(QRegularExpression(QFA("checkBox*")));
	foreach (QCheckBox* checkbox, checkboxes)
		{
		if(checkbox->isChecked())
			{
			rq->pushButton_search->setEnabled(true);
			rq->pushButton_search->setToolTip(QFA(""));
			return;
			}
		}
	rq->pushButton_search->setEnabled(false);
	rq->pushButton_search->setToolTip(QFA("<html><head/><body><p>Check at least one range option to enable this query button.</p></body></html>"));
	}

void RangeQueryDialog::closeEvent(QCloseEvent *event)
	{
// If this dialog has opened other dialogs, closing it first sends focus
// to unknown places when the other dialogs are closed.  Setting the focus to
// the main window seems to fix this.  (If there's a debug console, it may
// end up with the focus.)
	mainWindowPtr->ui->tableView->setFocus();
	event->accept();
	}

void RangeQueryDialog::keyPressEvent(QKeyEvent *event)
	{
	switch(event->key())
		{
		case Qt::Key_Home:
		case Qt::Key_End:
			if(!(Qt::ControlModifier & event->modifiers()))
				break;
			Q_FALLTHROUGH();
		case Qt::Key_PageDown:
		case Qt::Key_PageUp:
			Q_EMIT signal_keyToTable(event->key(), static_cast<int>(event->modifiers()));
			break;
		case Qt::Key_Escape:
			this->close();
			break;
// Intercepting keyPressEvents seems to break default button activation, so
// do it here.  Also take care setting focus after other buttons are clicked.
		case Qt::Key_Enter:
		case Qt::Key_Return:
			slot_pushButtonSearchClicked();
			break;
		default:
			return;
		}
	event->ignore();
	return;
	}

void RangeQueryDialog::wheelEvent(QWheelEvent *event)
	{
	Q_EMIT signal_wheelToTable(event->pos(), event->delta(), event->buttons(),
										event->modifiers(), event->orientation());
	event->ignore();
	return;
	}

