// This file is part of the disk cataloger Wssindex.  Wssindex is copyright
// 1985-2021 by Robert W. Babcock and WSS Division of DDC.  Effective 2010,
// Wssindex is distributed under the terms of the GNU General Public
// License.  Contact the author at wssddc@wssddc.com or see the project web
// page at <https://www.wssddc.com/wssindex/>.
//
// Wssindex uses the open source Qt toolkit from the Qt Project (qt-project.org)
// and The Qt Company (www.qt.io).
//
// Wssindex is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the
// Free Software Foundation, either version 3 of the License, or (at your
// option) any later version.
//
// Wssindex is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with Wssindex.  If not, see <https://www.gnu.org/licenses/>.

#include "OtherWssi/wssindex.h"
#include "Dialogs/simplequerydialog.h"
#include "ui_simplequerydialog.h"
#include "ui_mainwindow.h"
#include "Dialogs/mainwindow.h"
#include "OtherWssi/initialpos.h"
#include "OtherWssi/searchcompleter.h"
#include "OtherWssi/wssiproto.h"

SimpleQueryDialog::SimpleQueryDialog(QWidget *parent)
	: QMainWindow(parent), sq(new Ui::SimpleQueryDialog)
	{
	sq->setupUi(this);
	this->setWindowFlags(MyWindowFlags);
	restoreSettings();
	sq->lineEdit_searchstring->setFocusPolicy(Qt::StrongFocus);
	sq->pushButton_search->setDefault(true);
	sq->radioButton_new_query->setChecked(true);
	sq->checkBox_no_dot_id->setChecked(NoId);
	sq->lineEdit_searchstring->setFocus();
	slot_enableQueryButton(sq->lineEdit_searchstring->text());

	connect(this, SIGNAL(signal_updateDisplay(bool)),
		parent, SLOT(slot_updateDisplay(bool)));

	connect(sq->pushButton_archives, SIGNAL(clicked()),
		mainWindowPtr, SLOT(slot_pushButtonArchivesClicked()));
	connect(sq->pushButton_filetype, SIGNAL(clicked()),
		mainWindowPtr, SLOT(slot_pushButtonFiletypeClicked()));
	connect(sq->pushButton_names, SIGNAL(clicked()),
		mainWindowPtr, SLOT(slot_pushButtonNamesClicked()));
	connect(sq->pushButton_range, SIGNAL(clicked()),
		mainWindowPtr, SLOT(slot_pushButtonRangeClicked()));
	connect(sq->pushButton_special, SIGNAL(clicked()),
		mainWindowPtr, SLOT(slot_pushButtonSpecialClicked()));

	sq->lineEdit_searchstring->setCompleter(mainWindowPtr->pastSearchesCompleter);

//	connect(sq->lineEdit_searchstring, SIGNAL(textEdited(QString)),
	connect(sq->lineEdit_searchstring, SIGNAL(textChanged(QString)),
		this, SLOT(slot_enableQueryButton(QString)));
	connect(sq->pushButton_close, SIGNAL(clicked()),
		this, SLOT(slot_pushButtonCloseClicked()));
	connect(sq->pushButton_search, SIGNAL(clicked()),
		this, SLOT(slot_pushButtonSearchClicked()));
	connect(sq->pushButton_showall, SIGNAL(clicked()),
		this, SLOT(slot_pushButtonShowallClicked()));
	connect(sq->checkBox_no_dot_id, SIGNAL(toggled(bool)),
		this, SLOT(slot_updateNoId(bool)));
	connect(mainWindowPtr, SIGNAL(signal_hideSettingChanged(bool)),
		this, SLOT(slot_updateNoId(bool)));

	connect(this, SIGNAL(signal_keyToTable(int,int)),
		parent, SLOT(slot_keyToTable(int,int)));
	connect(this, SIGNAL(signal_wheelToTable(QPoint,int,Qt::MouseButtons,Qt::KeyboardModifiers,Qt::Orientation)),
		mainWindowPtr, SLOT(slot_wheelToTable(QPoint,int,Qt::MouseButtons,Qt::KeyboardModifiers,Qt::Orientation)));
	connect(mainWindowPtr, SIGNAL(signal_close_queries()),
			  this, SLOT(close()));
	}

SimpleQueryDialog::~SimpleQueryDialog()
	{
	delete sq;
	}

void SimpleQueryDialog::restoreSettings()
	{
	QSettings settings("WSS-DDC", "Wssindex\\Geometry");
	if(settings.contains("SimpleQuerySize"))
		{
		QSize size= settings.value("SimpleQuerySize").toSize();
		resize(size);
		}
	if(settings.contains("SimpleQueryPos"))
		{
		QPoint pos = settings.value("SimpleQueryPos").toPoint();
		move(pos);
		forceOnscreen(this);
		}
	else
		{
// frameGeometry is the same as geometry until the dialog is displayed
		this->show();
		move(mainWindowPtr->offCenter(this->frameGeometry(), SIMPLEQUERY_OFFSET));
		}
	}

// ------------------------------ Action buttons ------------------------------
void SimpleQueryDialog::slot_enableQueryButton(QString newtext)
	{
	sq->pushButton_search->setEnabled(newtext.length() > 0);
	}

void SimpleQueryDialog::slot_pushButtonCloseClicked()
	{
	QSettings settings("WSS-DDC", "Wssindex\\Geometry");
	settings.setValue("SimpleQueryPos", pos());
	settings.setValue("SimpleQuerySize", size());
	this->close();
	}

void SimpleQueryDialog::slot_pushButtonSearchClicked()
	{
	if(0 == Diskno)
		return;
	QChar qstar = QLatin1Char('*');
	QChar qmark = QLatin1Char('?');
	NoId = sq->checkBox_no_dot_id->isChecked();
	for(int i=0; i<SEL_DIM; i++)
		Usesel[i] = false;
	Nselect = 1;
	QString ts = sq->lineEdit_searchstring->text();
	if(!ts.contains(qstar) && !ts.contains(qmark))
		ts = qstar + ts + qstar;
	if(sq->checkBox_filenames_only->isChecked())
		{
		MatchName = ts;
		if(sq->checkBox_ignore_diacritical->isChecked())
			{
			Sel[0] = SEL_FILENAME_NODIACRIT;
			Usesel[SEL_FILENAME_NODIACRIT] = true;
			}
		else
			{
			Sel[0] = SEL_FILENAME;
			Usesel[SEL_FILENAME] = true;
			}
		}
	else
		{
		ZipName = ts;
		if(sq->checkBox_ignore_diacritical->isChecked())
			{
			Sel[0] = SEL_ZIPPY_NODIACRIT;
			Usesel[SEL_ZIPPY_NODIACRIT] = true;
			}
		else
			{
			Sel[0] = SEL_ZIPPY;
			Usesel[SEL_ZIPPY] = true;
			}
		}
// If not a new query, mark which files were previously selected
	if(!sq->radioButton_new_query->isChecked())
		updatePreviouslySelected();

	if(sq->radioButton_add_to->isChecked())
		{
		Sel[Nselect++] = SEL_ADD_TO_PREVIOUS;
		Usesel[SEL_ADD_TO_PREVIOUS] = true;
		}

	if(sq->radioButton_pick_from->isChecked())
		{
		Sel[Nselect++] = SEL_FROM_PREVIOUS;
		Usesel[SEL_FROM_PREVIOUS] = true;
		}

	if(sq->radioButton_exclude_from->isChecked())
		{
		Sel[Nselect++] = SEL_REM_FROM_PREVIOUS;
		Usesel[SEL_REM_FROM_PREVIOUS] = true;
		}

//#if !defined(QT_NO_DEBUG)
	sq->label_debug_time->clear();
	this->repaint();
	QElapsedTimer timer;
	timer.start();
//#endif
	mainWindowPtr->doSelection();
//#if !defined(QT_NO_DEBUG)
	sq->label_debug_time->setText(tr("Query time: %1 seconds").arg(static_cast<double>(timer.elapsed())/1000.));
//#endif

// Update the search history here so it gets called whether the search
// was triggered by clicking the Query button or hitting return
	updateSearchHistory(sq->lineEdit_searchstring->text());
	Q_EMIT signal_updateDisplay(true);
	}

void SimpleQueryDialog::slot_pushButtonShowallClicked()
	{
	NoId = sq->checkBox_no_dot_id->isChecked();
	if(0 == Diskno)
		return;
	if(NoId)
		selectAllExceptIdUnsorted();
	else
		selectAllUnsorted();
	Q_EMIT signal_updateDisplay(true);
	sq->lineEdit_searchstring->setFocus();
	}

// This is called when the Exclude Subdir & Vol records checkbox changes state either
// in this dialog or in the main menu or toolbar.  Make sure all locations match.
void SimpleQueryDialog::slot_updateNoId(bool checked)
	{
	NoId = checked;
	sq->checkBox_no_dot_id->setChecked(NoId);
	mainWindowPtr->menuHideCheckbox->setChecked(NoId);
	mainWindowPtr->ToolbarHideCheckbox->setChecked(NoId);
	}

// ------------------------------ Event handlers ------------------------------
void SimpleQueryDialog::closeEvent(QCloseEvent *event)
	{
// If this dialog has opened other dialogs, closing it first sends focus
// to unknown palces when the other dialogs are closed.  Setting the focus to
// the main window seems to fix this.  (If there's a debug console, it may
// end up with the focus.)
	mainWindowPtr->ui->tableView->setFocus();
	event->accept();
	}

void SimpleQueryDialog::keyPressEvent(QKeyEvent *event)
	{
	switch(event->key())
		{
		case Qt::Key_Home:
		case Qt::Key_End:
			if(!(Qt::ControlModifier & event->modifiers()))
				break;
			Q_FALLTHROUGH();
		case Qt::Key_PageDown:
		case Qt::Key_PageUp:
			Q_EMIT signal_keyToTable(event->key(), static_cast<int>(event->modifiers()));
			break;
		case Qt::Key_Escape:
			this->close();
			break;
// Intercepting keyPressEvents seems to break default button activation, so
// do it here.  Also take care setting focus after other buttons are clicked.
		case Qt::Key_Enter:
		case Qt::Key_Return:
			slot_pushButtonSearchClicked();
			break;
		default:
			return;
		}
	event->ignore();
	return;
	}

void SimpleQueryDialog::wheelEvent(QWheelEvent *event)
	{
	Q_EMIT signal_wheelToTable(event->pos(), event->delta(), event->buttons(),
										event->modifiers(), event->orientation());
	event->ignore();
	return;
	}
