// This file is part of the disk cataloger Wssindex.  Wssindex is copyright
// 1985-2021 by Robert W. Babcock and WSS Division of DDC.  Effective 2010,
// Wssindex is distributed under the terms of the GNU General Public
// License.  Contact the author at wssddc@wssddc.com or see the project web
// page at <https://www.wssddc.com/wssindex/>.
//
// Wssindex uses the open source Qt toolkit from the Qt Project (qt-project.org)
// and The Qt Company (www.qt.io).
//
// Wssindex is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the
// Free Software Foundation, either version 3 of the License, or (at your
// option) any later version.
//
// Wssindex is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with Wssindex.  If not, see <https://www.gnu.org/licenses/>.

#include "OtherWssi/wssindex.h"
#include "OtherWssi/globalsettings.h"
#include "OtherWssi/wssiproto.h"
#include "OtherWssi/wsxxxadd.h"
#include "Dialogs/indexwindow.h"
#include "Dialogs/mainwindow.h"

#include <fcntl.h>
#include <io.h>

#define PAK_SFX_START	6600UL
#define PAK_SFX_STOP		8700UL
#define ARCMARK	26

// arcadd - read a .ARC file and add information to wssindex database.

//	Comments may be extracted from ARC files created by either PKARC
//	or PAK, which unfortunately use different comment formats.  File may
//	also be a self-extracting PAK file.  Marker for this is ? null at an
//	offset of 6865 for PAK 2.0 or 8599 for PAK 2.51, followed by a valid
//	header.

#define TEMPBUF_SIZE 4096
#pragma pack(1)

int IndexWindow::arcadd(QString path, QString filename, int mode)
	{
	char tempbuf[TEMPBUF_SIZE];
	struct head
		{
		char arc_name[13];	// filename.ext
		long compressed_size;		// compressed file size (bytes)
		unsigned short arc_date_stamp;	// original file date, usual MS-DOS format
		unsigned short arc_time_stamp;	// original file time, usual MS-DOS format
		short crc;								// cyclic redundancy check, not used here
		long uncompressed_size;			// uncompressed size
		} ;
	struct en
		{
		unsigned long sig;				// 0x55aa4b50
		unsigned long comm_pointer;				// offset to comment descriptor
		} ;
	struct pak
		{
		unsigned char marker;			// always 254
		char type;				// 0=EOF, 1=remark, 2=path,
												//	3=security envelope, 4=error cor. code
		short file_number;				// number of file for remark, 0->archive
		long length;						// length of comment
		} ;
	union
		{
		struct head header;
		struct en end;
		struct pak extra;
		char comment[33];
		} arc;

	int return_code;
	FILE *fp;
	unsigned long arcpos, com_length, exe_length, last_pos, save_position;

	if(skip_extract(filename))
		return(SKIPPED);

	unsigned long basefile = Nofile - 1;
	return_code = EXTRACTED_OK;

	if(nullptr == (fp = arc_open(path, filename)))
		return(CANT_OPEN);

	struct filerec *dr_nofile = Dr + basefile;
	if(0 == QFA("e").compare(dr_nofile->Lfn.midRef(dr_nofile->ExtOffset,1), Qt::CaseInsensitive))
		{
		exe_length = static_cast<unsigned long>(_filelength(_fileno(fp)));
		if(exe_length < PAK_SFX_START)
			{
			return_code = NOT_RECOGNIZED;
			goto norm_ret;
			}
		arcpos = PAK_SFX_START;
		last_pos = qMin(PAK_SFX_STOP, exe_length);
		for( ; arcpos < last_pos; arcpos++)
			{
			fseek(fp, static_cast<long>(arcpos), SEEK_SET);
			do {
				while (arcpos < last_pos)
					{
					arcpos++;
					if('?' == fgetc(fp))
						break;
					}
				if ('\0' != fgetc(fp))
					break;
				if (ARCMARK != fgetc(fp))
					break;
				goto found_header;
				} while (false);
			}
		return_code = NOT_RECOGNIZED;
		goto norm_ret;
		}

	if(fgetc(fp) != ARCMARK)
		goto corrupt;			// this isn't an error, just a warning

found_header:
	for(;;)
		{
		char compression_method = static_cast<char>(fgetc(fp));
		if(compression_method < 0)
			goto corrupt;		// this is an error, but not a severe one
		if(compression_method == 0)
			{
// either end of ARC file, or start of comment area if written by
//	PKARC or PKPAK

			if(extractArcComment & mode)
				{
				save_position = static_cast<unsigned long>(ftell(fp));
// is it PAK format?
				unsigned short read_length = static_cast<unsigned short>(fread(&arc.extra, 8, 1, fp));
				if(0 == read_length)
					goto norm_ret;
				fseek(fp, static_cast<long>(save_position), SEEK_SET);
				if(254 == arc.extra.marker)
					{
// yes, it's PAK format
					for(;;)
						{
						if(1 != fread(&arc.extra, 8, 1, fp)
							|| 254 != arc.extra.marker
							|| 0 == arc.extra.type)
							goto norm_ret;
						if(1 == arc.extra.type)
							{
							com_length = static_cast<unsigned long>(arc.extra.length);
							if(com_length > TEMPBUF_SIZE-1)
								com_length = TEMPBUF_SIZE-1;
							if(1 != fread(tempbuf, static_cast<size_t>(com_length), 1, fp))
								goto corrupt;
							dr_nofile = Dr + basefile + arc.extra.file_number;
							if(com_arch(tempbuf, dr_nofile, static_cast<int>(com_length)))
								goto no_core;
							}
						else
							com_length = 0;
						fseek(fp, arc.extra.length-static_cast<long>(com_length), SEEK_CUR);
						}
					}
				else
// no, not PAK format, try PKARC
					{
					do {
						read_length = static_cast<unsigned short>(fread(&arc.end, 32, 1, fp));
						if(arc.end.sig != 0x55aa4b50UL)
							continue;
// last fread probably went past EOF unless file is padded with junk
						clearerr(fp);
						if(0 != fseek(fp, static_cast<long>(save_position), SEEK_SET))
							break;
						dr_nofile = Dr + basefile;
						if(arc.end.comm_pointer == save_position+32)
							{
							if(1 != fread(arc.comment, 32, 1, fp))
								break;
							if(com_arch(arc.comment, dr_nofile, 32))
								goto no_core;
							}
						if(0 != fseek(fp, 32L, SEEK_CUR))
							break;
						while(++basefile < Nofile)
							{
							if(1 != fread(arc.comment, 32, 1, fp))
								break;
							if(arc.end.sig == 0x55aa4b50UL)
								break;
							dr_nofile = Dr + basefile;
							if(com_arch(arc.comment, dr_nofile, 32))
								goto no_core;
							}
						break;
						} while (read_length == 1);
					}
				}

			goto norm_ret;
			}

		if(compression_method == 1)
			{
			if(1 != fread(&arc.header, sizeof(struct head)-sizeof(long int), 1, fp))
				goto corrupt;
			arc.header.uncompressed_size = arc.header.compressed_size;
			}
		else
			if(1 != fread(&arc.header, sizeof(struct head), 1, fp))
				goto corrupt;

		if(extractMembers & mode)
			{
			dr_nofile = Dr + Nofile;
			switch(Case_option_indexing)
				{
				case CASE_PRESERVE:
					break;
				case CASE_FORCE_UP:
					_strupr(arc.header.arc_name);
					break;
				case CASE_FORCE_LOW:
					_strlwr(arc.header.arc_name);
					break;
				}
			dr_nofile->Ddate = datextract(arc.header.arc_date_stamp);
			dr_nofile->Dtime = timextract(arc.header.arc_time_stamp);
			dr_nofile->Dsize = static_cast<unsigned long long>(arc.header.uncompressed_size);
			dr_nofile->Volno = Diskno;
			dr_nofile->Subdirno = basefile;
			dr_nofile->Comment = QString();
			dr_nofile->Category = QString();
			dr_nofile->Bitflags = 0;
			dr_nofile->Diskbits = 0;
			setInArc(dr_nofile);
			unbad(arc.header.arc_name, 12);
			dr_nofile->Lfn = arc.header.arc_name;
			dr_nofile->ExtOffset = static_cast<unsigned short>(lastdot(dr_nofile->Lfn));
			setHasLFN(dr_nofile);

			if(mainWindowPtr->incfile_and_check_for_error(false))
				goto no_core;
			}
// space forward to next header
		if(0 != fseek(fp, arc.header.compressed_size, SEEK_CUR)
			|| fgetc(fp) != ARCMARK)
			goto corrupt;		// this is an error, but not a severe one
		}

// never fall out of above loop
corrupt:
		return_code = CORRUPT;
		Q_EMIT signal_logMessage(tr("%1%2 is corrupted or not an ARC archive.")
				.arg(path, filename));
		goto norm_ret;

no_core:
		return_code = DATABASE_FULL;
		QMessageBox::critical(this, tr("Error"), tr("Out of memory in arcadd"));
		Stop_indexing = true;

// fall through to norm_ret
norm_ret:
		fclose(fp);
//		update_db_totals();
		qApp->processEvents();
		return(return_code);
	}
