// This file is part of the disk cataloger Wssindex.  Wssindex is copyright
// 1985-2021 by Robert W. Babcock and WSS Division of DDC.  Effective 2010,
// Wssindex is distributed under the terms of the GNU General Public
// License.  Contact the author at wssddc@wssddc.com or see the project web
// page at <https://www.wssddc.com/wssindex/>.
//
// Wssindex uses the open source Qt toolkit from the Qt Project (qt-project.org)
// and The Qt Company (www.qt.io).
//
// Wssindex is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the
// Free Software Foundation, either version 3 of the License, or (at your
// option) any later version.
//
// Wssindex is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with Wssindex.  If not, see <https://www.gnu.org/licenses/>.

#include "OtherWssi/wssindex.h"
#include "OtherWssi/wssiproto.h"
#include "OtherWssi/wsxxxadd.h"
#include "Dialogs/indexwindow.h"
#include "Non_member_functions/customio.h"

#define BM_2chars ('B' + ('M'<<8))
#define BA_2chars ('B' + ('A'<<8))

// limit number of images extracted (guarantees against loop in bad file)
#define MAX_IMAGES	5

// bmpadd - read a Win 3 or OS/2 .BMP file and add information to wssindex
//	database.  The image size and number of colors are converted into extra
// info of the form WIDTHxHEIGHTxNCOLORS.
//
//	BMP file format is described in PC Magazine, Vol 10, No. 10, May 28, 1991,
//	page 365.
//
//	OS/2 seems to have some BMP files with multiple images in a single file.
//	These begin with 'BA' rather than 'BM'.  The structure of these appears to
//	be:
//
//	2 bytes	'BA'
//	4 bytes	0x28
//	4 bytes	offset from beginning of file to next BA record, 0 if last
//	4 bytes	0
//	OS/2 BMP header (with extra stuff appended?)

#pragma pack(1)

int IndexWindow::bmpadd(QString path, QString filename, int /* dummy */)
	{
	FILE *fp;
	unsigned long basefile;

	short	type;					// 'BA' or 'BM'
	struct bahead
		{
		long	unknown;				// always 0x28?
		long	offsetnext;			// absolute offset of next BA record or zero
		long	reserved;			// unknown
		} ba_header;

	struct bmphead
		{
		long	file_size;
		short	res[2];		// reserved, set to 0
		long	offset;		// offset to image from start of file
		long	headsize;	// header size from this point on, 12 for OS/2,
								//		16 or more for win3
		union
			{
			struct
				{
				long	width;
				long	height;
				short	planes;				// set to 1
				short	bits_per_pixel;
// Win 3 header may stop anywhere after here
				long	comp_method;
				long	image_size;			// total bytes in image
				long	resolution[2];
				long	colors_used;
				long	important_colors;
				} win3;
			struct
				{
				short width;
				short height;
				short	planes;				// set to 1
				short	bits_per_pixel;
				} os2;
			} bmp_var;
		} bmp_header;

	long next_ba;

	unsigned int ncolor_bits;

	short height,
			width;

	if(skip_extract(filename))
		return(SKIPPED);

	basefile = Nofile - 1;
	int return_code = EXTRACTED_OK;

	if(nullptr == (fp = arc_open(path, filename)))
		return(CANT_OPEN);

	struct filerec *dr_nofile = Dr + basefile;
	bool last_read_bm = false;
	next_ba = 0;
	for(int n_bm_done = 0; n_bm_done<MAX_IMAGES ;)
		{
		if(n_bm_done>0 && last_read_bm)
			{
			if(next_ba > 0)
				fseek(fp, next_ba, SEEK_SET);
			else
				goto norm_ret;
			}
		type = static_cast<short>(fgetWord(fp));
		switch(type)
			{
			case BA_2chars:
				if(1 != fread(&ba_header, sizeof(struct bahead), 1, fp))
					goto corrupt;			// this isn't an error, just a warning
				next_ba = ba_header.offsetnext;
				last_read_bm = false;
				break;

			case BM_2chars:
				if(1 != fread(&bmp_header, sizeof(struct bmphead), 1, fp)
					|| bmp_header.headsize < 12)
					goto corrupt;			// this isn't an error, just a warning

				if(bmp_header.headsize < 16)
					{
					width = static_cast<short>(bmp_header.bmp_var.os2.width);
					height = static_cast<short>(bmp_header.bmp_var.os2.height);
					ncolor_bits = static_cast<unsigned int>(bmp_header.bmp_var.os2.bits_per_pixel);
					}
				else
					{
					width = static_cast<short>(bmp_header.bmp_var.win3.width);
					height = static_cast<short>(bmp_header.bmp_var.win3.height);
					ncolor_bits = static_cast<unsigned int>(bmp_header.bmp_var.win3.bits_per_pixel);
					}
				if(n_bm_done>0)
					dr_nofile->ExtraInfo += tr(", ");
				dr_nofile->ExtraInfo += tr("%1x%2x%3").arg(width).arg(height).arg(num_color_str(static_cast<int>(ncolor_bits)));
				setHasExtra(dr_nofile);
				last_read_bm = true;
				n_bm_done++;
				break;

			default:
				goto norm_ret;
			}
		}

norm_ret:
		fclose(fp);
		return(return_code);

corrupt:
		return_code = CORRUPT;
		Q_EMIT signal_logMessage(tr("%1%2 is corrupted or not a Windows or OS/2 BMP image file.")
				.arg(path, filename));
		goto norm_ret;
	}
