// This file is part of the disk cataloger Wssindex.  Wssindex is copyright
// 1985-2021 by Robert W. Babcock and WSS Division of DDC.  Effective 2010,
// Wssindex is distributed under the terms of the GNU General Public
// License.  Contact the author at wssddc@wssddc.com or see the project web
// page at <https://www.wssddc.com/wssindex/>.
//
// Wssindex uses the open source Qt toolkit from the Qt Project (qt-project.org)
// and The Qt Company (www.qt.io).
//
// Wssindex is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the
// Free Software Foundation, either version 3 of the License, or (at your
// option) any later version.
//
// Wssindex is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with Wssindex.  If not, see <https://www.gnu.org/licenses/>.

#include "OtherWssi/wssindex.h"
#include "OtherWssi/wssiproto.h"
#include "OtherWssi/wsxxxadd.h"
#include "Dialogs/indexwindow.h"

// Prototypes for local routines
static void read_color_table(unsigned char flag);

// gifadd - read a .GIF file and add information to wssindex database.
//	The image size and number of colors are converted into extra info
//	of the form WIDTHxHEIGHTxNCOLORS, plus monochrome if all colors in
//	color table have equal RGB components.
//
//	A GIF file with no image data will be reported as corrupt.  If there are
//	multiple images in the GIF, only the first one will be seen.

static FILE *fp;
static int ncolors;
static bool mono;

#define UBIT8(Z) (static_cast<unsigned char>(0x ## Z))

#pragma pack(1)

int IndexWindow::gifadd(QString path, QString filename, int /* dummy */)
	{
	struct gh
		{
		union
			{
			char gif[3];					// "GIF", 3 chars, no terminator
			unsigned short gif_2chars;
			};
		char version[3];					// "87a", "89a", etc., no terminator
		unsigned short screen_width;	// screen width
		unsigned short screen_height;	// screen height
		unsigned char flags;				// 1 bit global color table flag
												// 3 bits color resolution
												// 1 bit sort flag
												// 3 bits color table size
		unsigned short dont_care;
		} gif_header;

	char separator;		// should be 0x2c, 0x21 means an extension
								//	which needs to be skipped over
	struct lh
		{
		unsigned short x_position;		// x coordinate of upper left hand corner
		unsigned short y_position;		// y coordinate of upper left hand corner
		unsigned short width;			// image width
		unsigned short height;			// image height
		unsigned char flags;				// 1 bit local color table flag
												//	1 bit interleave flag
												//	1 bit sort flag
												//	2 bits reserved
												//	3 bits color table size
		} local_header;

	unsigned long basefile;
	int	return_code;
	struct filerec *dr_nofile;

	if(skip_extract(filename))
		return(SKIPPED);

	basefile = Nofile - 1;
	return_code = EXTRACTED_OK;

	if(nullptr == (fp = arc_open(path, filename)))
		return(CANT_OPEN);

	if(1 != fread(&gif_header, sizeof(struct gh), 1, fp))
		goto corrupt;

	if(0xd8ff == gif_header.gif_2chars)
		{
		return_code = CORRUPT;
		Q_EMIT signal_logMessage(tr("%1%2 is probably a misnamed JPG image, not GIF.")
				.arg(path, filename));
		goto norm_ret;
		}

	if(0x5089 == gif_header.gif_2chars)
		{
		return_code = CORRUPT;
		Q_EMIT signal_logMessage(tr("%1%2 is probably a misnamed PNG image, not GIF.")
				.arg(path, filename));
		goto norm_ret;
		}

	if(0 != strncmp(gif_header.gif, "GIF", 3))
		goto corrupt;			// this isn't an error, just a warning

	ncolors = 0;
	mono = false;
	if(gif_header.flags & UBIT8(80))
		read_color_table(gif_header.flags);
// Viewers seem to accept extra zeroes inserted here, so allow for up to 16
	for(int i=0; i<16; i++)
		{
		if(0 == (separator = static_cast<char>(fgetc(fp))))
			continue;
		ungetc(separator, fp);
		break;
		}
	while(0x21 == (separator = static_cast<char>(fgetc(fp))))
		{
		fgetc(fp);
		int skip_length;
		while(0 != (skip_length = fgetc(fp)))
			{
			fseek(fp, skip_length, SEEK_CUR);
			if(WSSERROR_OR_EOF(fp))
				goto corrupt;
			}
		}

// Could check for 0x3b here which marks trailer.  If found, it would mean that
//	the GIF contains no image, which isn't very interesting

	if(0x2c != separator)
		goto corrupt;
	if(1 != fread(&local_header, sizeof(struct lh), 1, fp))
		goto corrupt;			// this isn't an error, just a warning

	if(local_header.flags & UBIT8(80))
		read_color_table(local_header.flags);

	dr_nofile = Dr + basefile;
	dr_nofile->ExtraInfo = tr("%1x%2x%3").arg(local_header.width).arg(local_header.height).arg(ncolors);
	if(mono)
		dr_nofile->ExtraInfo += tr(", monochrome");
	setHasExtra(dr_nofile);

norm_ret:
		fclose(fp);
		return(return_code);

corrupt:
		return_code = CORRUPT;
		Q_EMIT signal_logMessage(tr("%1%2 is corrupted or not a GIF image.")
				.arg(path, filename));
		goto norm_ret;
	}

// read_color_table - read the color table of a GIF file, leaving the file
//	pointer positioned after the end of the color table.  If all RGB triplets
//	in the color table have equal R,G,B values, set mono=true indicating a
//	monochrome image.  If a color table is found, also set ncolors.

static void read_color_table(unsigned char flag)
	{
	int i;

	if(0 == (flag & UBIT8(80)))
		return;
	ncolors = 1 << (1+(flag&7));
	mono = true;
	for(i=0; i<ncolors; i++)
		{
		int red = fgetc(fp);
		int green = fgetc(fp);
		int blue = fgetc(fp);
		if(red != green || green != blue)
			mono = false;
		}
	return;
	}
