// This file is part of the disk cataloger Wssindex.  Wssindex is copyright
// 1985-2021 by Robert W. Babcock and WSS Division of DDC.  Effective 2010,
// Wssindex is distributed under the terms of the GNU General Public
// License.  Contact the author at wssddc@wssddc.com or see the project web
// page at <https://www.wssddc.com/wssindex/>.
//
// Wssindex uses the open source Qt toolkit from the Qt Project (qt-project.org)
// and The Qt Company (www.qt.io).
//
// Wssindex is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the
// Free Software Foundation, either version 3 of the License, or (at your
// option) any later version.
//
// Wssindex is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with Wssindex.  If not, see <https://www.gnu.org/licenses/>.

#include "OtherWssi/wssindex.h"
#include "OtherWssi/globalsettings.h"
#include "OtherWssi/wssiproto.h"
#include "OtherWssi/wsxxxadd.h"
#include "Dialogs/indexwindow.h"
#include "Dialogs/mainwindow.h"

#include <fcntl.h>

// hapadd - read a .HAP file and add information to wssindex database.

#pragma pack(1)

int IndexWindow::hapadd(QString path, QString filename, int /* dummy */)
	{
	struct haphead
		{
		unsigned	long hapid;			// 0x46483391
		unsigned long commentl;		// header comment length
		char reserved[7];
		} hap_header;

	struct hapfile
		{
		unsigned long fileid;			// 0x574a688e
		unsigned long hapsize;			// size of the coded hapfile
		unsigned long CRC1;				// CRC input file (original)
		unsigned long int CRC2;		// CRC output file (compressed)
		char reserved[1];	// reserved, if zero assume 3.00
		char attri;			// orig file attributes
		unsigned short time;			// orig file time
		unsigned short date;			// orig file date
		unsigned long orisize;			// orig file size in bytes
		char fname[13];		// always 13 chars ASCIIZ
		char storage_method;	// compression method
		} hap_file;

	FILE *fp;						// file pointer for reading archive
	long				seek_offset;		// offset to skip to next header

	if(skip_extract(filename))
		return(SKIPPED);

	unsigned long basefile = Nofile - 1;
	int return_code = EXTRACTED_OK;

	if(nullptr == (fp = arc_open(path, filename)))
		return(CANT_OPEN);

	struct filerec *dr_nofile;
	if(1 != fread(&hap_header, sizeof(struct haphead), 1, fp)
		|| 0x46483391UL != hap_header.hapid)
		goto corrupt;

// Header allows for comments, but not yet implemented by HAP
	seek_offset = static_cast<long>(hap_header.commentl);
	for(;;)
		{
		fseek(fp, seek_offset, SEEK_CUR);
		if(1 != fread(&hap_file, sizeof(struct hapfile), 1, fp)
			|| 0x574a688eUL != hap_file.fileid)
			goto norm_ret;
		dr_nofile = Dr + Nofile;

		switch(Case_option_indexing)
			{
			case CASE_PRESERVE:
				break;
			case CASE_FORCE_UP:
				_strupr(hap_file.fname);
				break;
			case CASE_FORCE_LOW:
				_strlwr(hap_file.fname);
				break;
			}

		dr_nofile->Ddate = datextract(hap_file.date);
		dr_nofile->Dtime = timextract(hap_file.time);
		dr_nofile->Dsize = hap_file.orisize;
		dr_nofile->Volno = static_cast<unsigned long>(Diskno);
		dr_nofile->Subdirno = static_cast<unsigned long>(basefile);
		dr_nofile->Comment = QString();
		dr_nofile->Category = QString();
		dr_nofile->Bitflags = 0;
		dr_nofile->Diskbits = 0;
		setInArc(dr_nofile);
		unbad(hap_file.fname, 12);
		dr_nofile->Lfn = hap_file.fname;
		dr_nofile->ExtOffset = static_cast<unsigned short>(lastdot(dr_nofile->Lfn));
		setHasLFN(dr_nofile);

		if(mainWindowPtr->incfile_and_check_for_error(false))
			goto no_core;

// Space forward to next header, method depends on header type
		seek_offset = static_cast<long>(hap_file.hapsize);
		}

// Fall out of above loop only if file is corrupt.
corrupt:
		return_code = CORRUPT;
		Q_EMIT signal_logMessage(tr("%1%2 is corrupted or not an Hap&Pah archive.")
				.arg(path, filename));
// fall through to norm_ret

norm_ret:
		fclose(fp);
//		update_db_totals();
		qApp->processEvents();
		return(return_code);

no_core:
		return_code = DATABASE_FULL;
		QMessageBox::critical(this, tr("Error"), tr("Out of memory in hapadd"));
		Stop_indexing = true;
		goto norm_ret;
	}
