// This file is part of the disk cataloger Wssindex.  Wssindex is copyright
// 1985-2021 by Robert W. Babcock and WSS Division of DDC.  Effective 2010,
// Wssindex is distributed under the terms of the GNU General Public
// License.  Contact the author at wssddc@wssddc.com or see the project web
// page at <https://www.wssddc.com/wssindex/>.
//
// Wssindex uses the open source Qt toolkit from the Qt Project (qt-project.org)
// and The Qt Company (www.qt.io).
//
// Wssindex is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the
// Free Software Foundation, either version 3 of the License, or (at your
// option) any later version.
//
// Wssindex is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with Wssindex.  If not, see <https://www.gnu.org/licenses/>.

#include "OtherWssi/wssindex.h"
#include "OtherWssi/globalsettings.h"
#include "OtherWssi/wssiproto.h"
#include "OtherWssi/wsxxxadd.h"
#include "Dialogs/indexwindow.h"
#include "Dialogs/mainwindow.h"

#include <fcntl.h>

#define HP_2chars ('H' + ('P'<<8))
#define ST_2chars ('S' + ('T'<<8))

// hypadd - read a .HYP file and add information to wssindex database.
//
//	HYPER files do not have comments.

#pragma pack(1)
#define TEMPBUF_SIZE 4096

int IndexWindow::hypadd(QString path, QString filename, int /* dummy */)
	{
	char tempbuf[TEMPBUF_SIZE];
	struct head
		{
		char cz;							// control-z
		unsigned short id;			// 'HP' or 'ST'
		char version;
		unsigned long compressed_size;		// compressed file size (bytes)
		unsigned long uncompressed_size;		// uncompressed size
		unsigned short time_stamp;	// original file time, usual MS-DOS format
		unsigned short date_stamp;	// original file date, usual MS-DOS format
		long crc;						// checksum
		char file_attr;				// MS-DOS attribute byte
		unsigned char file_name_length;		// number of file name characters
		} hyp;

	char *full_name;	// name of file in archive, may include a path

	FILE *fp;
	struct filerec *dr_nofile;

	if(skip_extract(filename))
		return(SKIPPED);

	int return_code = EXTRACTED_OK;
	unsigned long basefile = Nofile - 1;

	if(nullptr == (fp = arc_open(path, filename)))
		return(CANT_OPEN);

	for(;;)
		{
		if(1 != fread(&hyp, sizeof(struct head), 1, fp))
			{
			if(feof(fp))
				goto norm_ret;
			else
				goto corrupt;
			}
		if(HP_2chars != hyp.id && ST_2chars != hyp.id)
			goto corrupt;
		if(1 != fread(tempbuf, hyp.file_name_length, 1, fp))
			goto corrupt;

		dr_nofile = Dr + Nofile;

		full_name = tempbuf;
		full_name[static_cast<int>(hyp.file_name_length)] = '\0';
		for(int i=0; i<hyp.file_name_length; i++)
			if(*(tempbuf+i) == '\\')
				full_name = tempbuf+i+1;

		switch(Case_option_indexing)
			{
			case CASE_PRESERVE:
				break;
			case CASE_FORCE_UP:
				_strupr(full_name);
				break;
			case CASE_FORCE_LOW:
				_strlwr(full_name);
				break;
			}

		dr_nofile->Ddate = datextract(hyp.date_stamp);
		dr_nofile->Dtime = timextract(hyp.time_stamp);
		dr_nofile->Dsize = hyp.uncompressed_size;
		dr_nofile->Volno = Diskno;
		dr_nofile->Subdirno = basefile;
		dr_nofile->Comment = QString();
		dr_nofile->Category = QString();
		dr_nofile->Bitflags = 0;
		dr_nofile->Diskbits = 0;
		setInArc(dr_nofile);
		unbad(full_name, 0);
		dr_nofile->Lfn = full_name;
		dr_nofile->ExtOffset = static_cast<unsigned short>(lastdot(dr_nofile->Lfn));
		setHasLFN(dr_nofile);

		if(mainWindowPtr->incfile_and_check_for_error(false))
			goto no_core;

// space forward to next header
		if(0 != fseek(fp, static_cast<long>(hyp.compressed_size), SEEK_CUR))
			goto corrupt;		// this is an error, but not a severe one
		}

// never fall out of above loop
corrupt:
		return_code = CORRUPT;
		Q_EMIT signal_logMessage(tr("%1%2 is corrupted or not a HYPER archive.")
				.arg(path, filename));
// Fall through to norm_ret

norm_ret:
		fclose(fp);
//		update_db_totals();
		qApp->processEvents();
		return(return_code);

no_core:
		return_code = DATABASE_FULL;
		QMessageBox::critical(this, tr("Error"), tr("Out of memory in hapadd"));
		Stop_indexing = true;
		goto norm_ret;
	}
