// This file is part of the disk cataloger Wssindex.  Wssindex is copyright
// 1985-2021 by Robert W. Babcock and WSS Division of DDC.  Effective 2010,
// Wssindex is distributed under the terms of the GNU General Public
// License.  Contact the author at wssddc@wssddc.com or see the project web
// page at <https://www.wssddc.com/wssindex/>.
//
// Wssindex uses the open source Qt toolkit from the Qt Project (qt-project.org)
// and The Qt Company (www.qt.io).
//
// Wssindex is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the
// Free Software Foundation, either version 3 of the License, or (at your
// option) any later version.
//
// Wssindex is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with Wssindex.  If not, see <https://www.gnu.org/licenses/>.

#include "OtherWssi/wssindex.h"
#include "OtherWssi/wsxxxadd.h"
#include "Dialogs/indexwindow.h"

// _rotl works on int, not short
#define _rotl_short(a,b) (static_cast<short>(((a)<<(b)) | ((static_cast<unsigned short>((a))>>(16-(b))))))

// num_color_str - returns pointer to a static buffer with nbits converted
//	to K or M.
QString IndexWindow::num_color_str(int nbits)
	{
	if(nbits < 10)
		return (tr("%1").arg(1<<nbits));
	else if(nbits < 20)
		return (tr("%1K").arg(1<<(nbits-10)));
	else
		return (tr("%1M").arg(1<<(nbits-20)));
	}

// jpgadd - read a .JPG file and add information to wssindex database.
//	The image size and number of colors are converted into extra info
//	of the form WIDTHxHEIGHTxNBITS.
//
//	If there are multiple images in the JPG file, only the first one will
//	be seen (possible?).

// extracted from rdjpgcom.c
// JPEG markers consist of one or more 0xFF bytes, followed by a marker
// code byte (which is not an FF).  Here are the marker codes of interest
// in this program.  (See jdmarker.c for a more complete list.)

#define M_SOF0  0xC0FFU            // Start Of Frame N
#define M_SOF1  0xC1FFU            // N indicates which compression process
#define M_SOF2  0xC2FFU            // Only SOF0-SOF2 are now in common use
#define M_SOF3  0xC3FFU
#define M_SOF5  0xC5FFU            // NB: codes C4 and CC are NOT SOF markers
#define M_SOF6  0xC6FFU
#define M_SOF7  0xC7FFU
#define M_SOF9  0xC9FFU
#define M_SOF10 0xCAFFU
#define M_SOF11 0xCBFFU
#define M_SOF13 0xCDFFU
#define M_SOF14 0xCEFFU
#define M_SOF15 0xCFFFU
#define M_SOI   0xD8FFU            // Start Of Image (beginning of datastream/
#define M_EOI   0xD9FFU            // End Of Image (end of datastream)
#define M_SOS   0xDAFFU            // Start Of Scan (begins compressed data)
#define M_APP0  0xE0FFU            // Application-specific marker, type N
#define M_APP12 0xECFFU            // (we don't bother to list all 16 APPn's)
#define M_COM   0xFEFFU            // COMment

#pragma pack(1)

int IndexWindow::jpgadd(QString path, QString filename, int /* dummy */)
	{
	struct jpeg_block						// inverted byte order for all fields
		{
		unsigned short		type;						// block type code, 0xff..
		unsigned short		length;					// length of block, including length but not type
		char					bits_per_component;	// bits for each color plane
		unsigned short		height;					// image height
		unsigned short		width;					// image width
		char 					components;				// number of image planes, 1 or 3
		} jpeg;
	FILE *fp;
	short height, width;

	if(skip_extract(filename))
		return(SKIPPED);

	if(nullptr == (fp = arc_open(path, filename)))
		return(CANT_OPEN);

	int return_code = EXTRACTED_OK;
	unsigned long basefile = Nofile - 1;
	struct filerec *dr_nofile = Dr + basefile;

	if(1 != fread(&jpeg, sizeof(unsigned short), 1, fp))
		goto corrupt;			// this isn't an error, just a warning
	if(0x4947 == jpeg.type)	// GI
		{
		return_code = CORRUPT;
		Q_EMIT signal_logMessage(tr("%1%2 is probably a misnamed GIF image, not JPG.")
				.arg(path, filename));
		goto norm_ret;
		}

	if(0x5089 == jpeg.type)	// .P
		{
		return_code = CORRUPT;
		Q_EMIT signal_logMessage(tr("%1%2 is probably a misnamed PNG image, not JPG.")
				.arg(path, filename));
		goto norm_ret;
		}

	if(M_SOI != jpeg.type)
		goto corrupt;			// this isn't an error, just a warning

	for(;;)
		{
		if(1 != fread(&jpeg, sizeof(struct jpeg_block), 1, fp))
			goto corrupt;

		if(0xff != (jpeg.type & 0xff))
			goto corrupt;

		switch(jpeg.type)
			{
			case M_SOF0:
			case M_SOF1:
			case M_SOF2:
			case M_SOF3:
// c4 is Define Huffman Table
			case M_SOF5:
			case M_SOF6:
			case M_SOF7:
// c8 is JPEG Extensions
			case M_SOF9:
			case M_SOF10:
			case M_SOF11:
// cc is Define Arithmetic Coding
			case M_SOF13:
			case M_SOF14:
			case M_SOF15:
				goto header_found;
			default:
				if((jpeg.length = static_cast<unsigned short>(_rotl_short(jpeg.length, 8))) == 0)
					goto corrupt;
				fseek(fp, static_cast<unsigned long>(jpeg.length) - 8, SEEK_CUR);
				break;
			}
		}

header_found:
	width = _rotl_short(jpeg.width, 8);
	height = _rotl_short(jpeg.height, 8);

	if(1 == jpeg.components)
		dr_nofile->ExtraInfo = tr("%1x%2x%3, monochrome").arg(width).arg(height)
		.arg(num_color_str(static_cast<int>(jpeg.bits_per_component) * jpeg.components));
	else
		dr_nofile->ExtraInfo = tr("%1x%2x%3").arg(width).arg(height)
		.arg(num_color_str(static_cast<int>(jpeg.bits_per_component) * jpeg.components));
	setHasExtra(dr_nofile);

norm_ret:
		fclose(fp);
		return(return_code);

corrupt:
		return_code = CORRUPT;
		Q_EMIT signal_logMessage(tr("%1%2 is corrupted or not a JPG image.")
				.arg(path, filename));
		goto norm_ret;
	}
