// This file is part of the disk cataloger Wssindex.  Wssindex is copyright
// 1985-2021 by Robert W. Babcock and WSS Division of DDC.  Effective 2010,
// Wssindex is distributed under the terms of the GNU General Public
// License.  Contact the author at wssddc@wssddc.com or see the project web
// page at <https://www.wssddc.com/wssindex/>.
//
// Wssindex uses the open source Qt toolkit from the Qt Project (qt-project.org)
// and The Qt Company (www.qt.io).
//
// Wssindex is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the
// Free Software Foundation, either version 3 of the License, or (at your
// option) any later version.
//
// Wssindex is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with Wssindex.  If not, see <https://www.gnu.org/licenses/>.

// read an MP3 sound file and add information to wssindex database.

#include "OtherWssi/wssindex.h"
#include "OtherWssi/wsxxxadd.h"
#include "Dialogs/indexwindow.h"
#include "Non_member_functions/customio.h"

#include <io.h>
#include <sstream>
#include <iomanip>

#include <QMessageBox>

#define EOS '\0'	// standard end of string
#define TEMPBUF_SIZE 256

#define WAVE_FORMAT_PCM     1		// from mmsystem.h
#define IBM_FORMAT_MULAW	0x0101
#define IBM_FORMAT_ALAW		0x0102
#define IBM_FORMAT_ADPCM	0x0103

#define MP3e	0x4d503365
#define ID33	0x49443303
#define APIC	0x41504943UL
#define COMM	0x434f4d4dUL
#define MCDI	0x4d434449UL
#define TALB	0x54414c42UL
#define TCON	0x54434f4eUL
#define TIT2	0x54495432UL
#define TLEN	0x544c454eUL
#define TPE1	0x54504531UL
#define TYER	0x54594552UL
#define Xing	0x58696E67UL

#define UBIT8(Z) (static_cast<unsigned char>(0x ## Z))

#define N_GENRE	148
static const char *genre[N_GENRE] = {"Blues", "Classic Rock", "Country", "Dance", "Disco",
	"Funk", "Grunge", "Hip-Hop", "Jazz", "Metal", "New Age", "Oldies", "Other",
	"Pop","R&B", "Rap", "Reggae", "Rock", "Techno", "Industrial", "Alternative",
	"Ska", "Death Metal", "Pranks", "Soundtrack", "Euro-Techno", "Ambient",
	"Trip-Hop", "Vocal", "Jazz+Funk", "Fusion", "Trance", "Classical",
	"Instrumental", "Acid", "House", "Game", "Sound Clip", "Gospel", "Noise",
	"AlternRock", "Bass", "Soul", "Punk", "Space", "Meditative",
	"Instrumental Pop", "Instrumental Rock", "Ethnic", "Gothic", "Darkwave",
	"Techno-Industrial", "Electronic", "Pop-Folk", "Eurodance", "Dream",
	"Southern Rock", "Comedy", "Cult", "Gangsta", "Top 40", "Christian Rap",
	"Pop/Funk", "Jungle", "Native American", "Cabaret", "New Wave",
	"Psychadelic", "Rave", "Showtunes", "Trailer", "Lo-Fi", "Tribal",
	"Acid Punk", "Acid Jazz", "Polka", "Retro", "Musical", "Rock & Roll",
	"Hard Rock", "Folk", "Folk-Rock", "National Folk", "Swing", "Fast Fusion",
	"Bebob", "Latin", "Revival", "Celtic", "Bluegrass", "Avantgarde",
	"Gothic Rock", "Progressive Rock", "Psychedelic Rock", "Symphonic Rock",
	"Slow Rock", "Big Band", "Chorus", "Easy Listening", "Acoustic", "Humour",
	"Speech", "Chanson", "Opera", "Chamber Music", "Sonata", "Symphony",
	"Booty Brass", "Primus", "Porn Groove", "Satire", "Slow Jam", "Club",
	"Tango", "Samba", "Folklore", "Ballad", "Poweer Ballad", "Rhythmic Soul",
	"Freestyle", "Duet", "Punk Rock", "Drum Solo", "A Capela", "Euro-House",
	"Dance Hall", "Goa", "Drum & Bass", "Club-House", "Hardcore", "Terror",
	"Indie", "BritPop", "Negerpunk", "Polsk Punk", "Beat",
	"Christian Gangsta Rap", "Heavy Metal", "Black Metal", "Crossover",
	"Contemporary Christian", "Christian Rock","Merengue", "Salsa",
	"Thrash Metal", "Anime", "Jpop", "Synthpop"};

// Bitrate for version 1,2 (2.5 same as 2) and layer 1,2,3.
//	0 means free format determined by application.

static int mp3_v1_L1[] = {  0,  32,  64,  96, 128, 160, 192, 224,
								  256, 288, 320, 352, 384, 416, 448};
static int mp3_v1_L2[] = {  0,  32,  48,  56,  64,  80,  96, 112,
								  128, 160, 192, 224, 256, 320, 384};
static int mp3_v1_L3[] = {  0,  32,  40,  48,  56,  64,  80,  96,
								  112, 128, 160, 192, 224, 256, 320};
static int mp3_v2_L1[] = {  0,  32,  48,  56,  64,  80,  96, 112,
								  128, 144, 160, 176, 192, 224, 256};
static int mp3_v2_L23[]= {  0,   8,  16,  24,  32,  40,  48,  56,
									64,  80,  96, 112, 128, 144, 160};

// Sampling rate frequency index in Hz

static int mp3_samp_v1[] = {44100, 48000, 32000, 0};
static int mp3_samp_v2[] = {22050, 24000, 16000, 0};
static int mp3_samp_v25[] = {11025, 12000, 8000, 0};

static long last_sync_point;
#define GENRE_SIZE 128

unsigned char BOMfffe[3] = {0xff, 0xfe, 0};
unsigned char BOMfeff[3] = {0xfe, 0xff, 0};

int IndexWindow::mp3add(QString path, QString filename, int /* dummy */)
	{
	char tempbuf[TEMPBUF_SIZE];
	QString genre_id3;
	unsigned long header, samprate, bitrate, frame_bytes, frames;
	FILE *fp;
	int mpeg_version, layer_desc, bitrate_index, samprate_index;
	int padding, channel_mode, minutes;
	unsigned int n_genre;
	unsigned long frame_id;
	unsigned long frame_size;
	unsigned char id3_flags;
	bool extended_header, unsync, use_id3_seconds, variable_bitrate;
	bool id3_title, id3_artist, id3_album, id3_year, id3_comment, id3_genre;
	bool corruptMp3 = false;
	unsigned long bookmark, header_skip, id3_size, vbr_flags, vbr_offset, vbr_sync;
	unsigned int next_sync;
	std::stringstream ss;
	static QRegularExpression tconParenSplitter = QRegularExpression("[()]");

	if(skip_extract(filename))
		return(SKIPPED);

	unsigned long basefile = Nofile - 1;
	int return_code = EXTRACTED_OK;

	if(nullptr == (fp = arc_open(path, filename)))
		return(CANT_OPEN);
	unsigned long filelen = static_cast<unsigned long>(_filelength(_fileno(fp)));

	struct filerec *dr_nofile = Dr + basefile;
	int corrupt_counter = 0;
	int seconds32 = 0;
	int secondscalc = 0;
	int secondsvbr = 0;
	int seconds = 0;
	unsigned long id3_max = 0;
	bool has_tag = use_id3_seconds = false;
	id3_title = id3_artist = id3_album = id3_year = id3_comment = id3_genre = false;
	QString tag_comment, tag_album, tag_title, tag_artist, tag_year, play_time;
	QString exInfo;
	header = find_sync(fp);

	if(ID33 == header)
		{
// ID3 tag found
		/* id3_minor = */ fgetc(fp);
		id3_flags = static_cast<unsigned char>(fgetc(fp));
		extended_header = (id3_flags & 0x40) != 0;
		unsync = (id3_flags & UBIT8(80)) != 0;
		id3_size = fget_7_long(fp);	// includes extended header, if any
		id3_max = static_cast<unsigned long>(ftell(fp)) + id3_size;
		if(extended_header)
			{
			bookmark = static_cast<unsigned long>(ftell(fp));
			header_skip = fget_unsync_long(fp, unsync);
			fseek(fp, static_cast<long>(bookmark+header_skip+4), SEEK_SET);
			}
// Use id3_max-4 as limit rather than id3_max so that read of frame_id
//	cannot go past end of id3 header.  Some mp3s have a few extra zero
//	bytes before the sync record.

		while(static_cast<unsigned long>(ftell(fp)) < id3_max-4)
			{
			frame_id = fgetSwappedLong(fp);
// This isn't  a real tag, it's filler after all the tags
			if(0x4d503365 == frame_id)		// MP3e
				frame_id = 0;

// A frame id with a leading byte of zero is illegal according to all docs
//	I've seen, but it does happen and other utilities do not complain.
//	Ignore all tags after this because there's no length specifier either.
			if(0 == (frame_id & 0xff000000UL))
				{
				fseek(fp, static_cast<long>(id3_max), SEEK_SET);
				break;
				}
			frame_size = fget_guess_long(fp, id3_max);
			/* frame_flags = */ fgetSwappedWord(fp);
			if(0 == frame_size)			// This does happen
				continue;
			QString extractedText;
			switch(frame_id)
				{
				case COMM:		// COMM - comment
// Unlike most other frames, COMM has a 3-byte language specifier after
//	the text encoding byte.  There's a null before the text, so the minimum
//	legal frame size is 5.
				tag_comment = id3text(fp, frame_size, true, corruptMp3);
				if(tag_comment.length() > 0)
					id3_comment = true;
				break;

				case TALB:		// TALB - album
					tag_album = id3text(fp, frame_size, false, corruptMp3);
					if(tag_album.length() > 0)
					if(corruptMp3)
						goto corrupt;
					id3_album = true;
					break;

				case TCON:		// TCON - content type
// Content type could be a string, a number in parenthesis indexing into
//	the standard genre list, or some combination of both.
					genre_id3 = id3text(fp, frame_size, false, corruptMp3);
					if(corruptMp3)
						goto corrupt;
					if(genre_id3.startsWith("("))
						{
						n_genre = extractedText.split(tconParenSplitter).at(0).toUInt();
						if(n_genre < N_GENRE)
							genre_id3 = genre[n_genre];
						}
					id3_genre = true;
					break;

				case TIT2:		// TIT2 - title
					tag_title = id3text(fp, frame_size, false, corruptMp3);
					if(corruptMp3)
						goto corrupt;
					id3_title = true;
					break;

				case TLEN:		// TLEN - length in milliseconds
					extractedText = id3text(fp, frame_size, false, corruptMp3);
					if(corruptMp3)
						goto corrupt;
					seconds32 = extractedText.toInt()/1000;
					use_id3_seconds = true;
					break;

				case TPE1:		// TPE1 - Lead performer
					tag_artist = id3text(fp, frame_size, false, corruptMp3);
					if(corruptMp3)
						goto corrupt;
					id3_artist = true;
					break;

				case TYER:		// TYER - year
					tag_year = id3text(fp, frame_size, false, corruptMp3);
					if(corruptMp3)
						goto corrupt;
					id3_year = true;
					break;

// Some software seems to mess up the length of MCDI frames if unsync mechanism
//	needs to be used.  Detect and correct for this???
// This could happen elsewhere, but MCDI seems the most common place.
//				case MCDI:			// MCDI
//					Q_EMIT signal_logMessage(tr("%1%2 has MCDI tag.")
//							.arg(path).arg(filename));

				default:						// Unrecognized, skip frame
					fseek(fp, static_cast<long>(frame_size), SEEK_CUR);
					break;

				}
			}
		header = find_sync(fp);
		}

 header_retry:
	if(0xffe00000UL != (header & 0xffe00000UL))
		goto maybe_corrupt;

	samprate = bitrate = n_genre = 0;	// to quite gcc warnings
	mpeg_version = (header & 0x180000UL)>>19;
	layer_desc = (header & 0x60000UL)>>17;
//	crc_bytes = 2 - ((header & 10000UL)>>15);		// 0 or 2
	bitrate_index = (header & 0xf000UL)>>12;
	samprate_index = (header & 0xc00UL)>>10;
	padding = (header & 0x200UL)>>9;					// 1 or 0
	channel_mode = (header & 0xc0UL)>>6;

	switch(mpeg_version)
		{
		case 1:		// reserved
			goto maybe_corrupt;
		case 0:		// MPEG version 2.5
			samprate = static_cast<unsigned long>(mp3_samp_v25[samprate_index]);
			goto ver2_layer;
		case 2:		// MPEG version 2
			samprate = static_cast<unsigned long>(mp3_samp_v2[samprate_index]);
	ver2_layer:
			switch(layer_desc)
				{
				case 0:	// reserved
					goto maybe_corrupt;
				case 1:	// Layer III
				case 2:	// Layer II
					bitrate = static_cast<unsigned long>(mp3_v2_L23[bitrate_index]);
					break;
				case 3:	// Layer I
					bitrate = static_cast<unsigned long>(mp3_v2_L1[bitrate_index]);
					break;
				}
			break;
		case 3:		// MPEG version 1
			samprate = static_cast<unsigned long>(mp3_samp_v1[samprate_index]);
			switch(layer_desc)
				{
				case 0:	// reserved
					goto maybe_corrupt;
				case 1:	// Layer III
					bitrate = static_cast<unsigned long>(mp3_v1_L3[bitrate_index]);
					break;
				case 2:	// Layer II
					bitrate = static_cast<unsigned long>(mp3_v1_L2[bitrate_index]);
					break;
				case 3:	// Layer I
					bitrate = static_cast<unsigned long>(mp3_v1_L1[bitrate_index]);
					break;
				}
			break;
		}

// Check for a variable bit rate header
	if(3 == mpeg_version)
		vbr_offset = (channel_mode == 3) ? 17 : 32;
	else
		vbr_offset = (channel_mode == 3) ? 9 : 17;
	fseek(fp, static_cast<long>(vbr_offset), SEEK_CUR);
	vbr_sync = fgetSwappedLong(fp);
	vbr_offset += 4;
	if(Xing == vbr_sync)
		{
		variable_bitrate = true;
		vbr_flags = fgetSwappedLong(fp);
		frames = fgetSwappedLong(fp);
//		vbr_offset += 8;
		if(!(vbr_flags & 1))
			frames = 0;
		if(frames)
			bitrate = (((filelen - id3_max)/frames) * samprate)/((3 == layer_desc) ? 12000UL : 144000UL);
		if(!use_id3_seconds)
			{
			secondsvbr = static_cast<int>(8*(filelen - id3_max) / (1000*bitrate));
			}
		}
	else
		{
		variable_bitrate = false;
		if(0 == samprate || 0 == bitrate)
			goto maybe_corrupt;
		else
			{
			if(3 == layer_desc)		// Layer 1
				{
				if(3 == mpeg_version)	// mpeg 1?
					frame_bytes = ((12000 * bitrate) / samprate + static_cast<unsigned long>(padding)) * 4;
				else
					frame_bytes = ((6000 * bitrate) / samprate + static_cast<unsigned long>(padding)) * 4;
//				samples_per_frame = 384;
				}
			else							// Layer 2 or 3
				{
				if(3 == mpeg_version)	// mpeg 1?
					frame_bytes = (144000 * bitrate) / samprate + static_cast<unsigned long>(padding);
				else
					frame_bytes = (72000 * bitrate) / samprate + static_cast<unsigned long>(padding);
//				samples_per_frame = 1152;
				}
			}

// Check whether next frame starts with frame sync as a partial guard
//	against syncing on random data.  Don't do this for VBR since finding
//	"Xing" is a pretty good check.
		fseek(fp, static_cast<long>(frame_bytes - 4 - vbr_offset), SEEK_CUR);
		next_sync = fgetSwappedWord(fp);
//sprintf(tempbuf, "Sync offset = %lx, value = %x", ftell(fp)-2, next_sync);
//string_new(tempbuf, true);
		if(0xffe0 != (next_sync & 0xffe0))
			goto maybe_corrupt;

		if(filelen < 128)		// could be smaller, but then not not interesting
			goto corrupt;

		if(bitrate > 0 && samprate > 0)
			secondscalc = static_cast<int>(8*(filelen - id3_max) / (1000*bitrate));
		}

// If ID3v2 tag at beginning of tile has not supplied enough information,
// look for ID3v1 tag which is 128 bytes at end of file.
	if(!(id3_title && id3_artist && id3_album && id3_year && id3_comment && id3_genre))
		{
		fseek(fp, -128, SEEK_END);
		fread(tempbuf, 128, 1, fp);
		has_tag = (0 == strncmp("TAG", tempbuf, 3));

		if(has_tag)
			{
			if(!id3_title)
				{
				charToQstring(tempbuf+3, 30, tag_title);
				tag_title = tag_title.trimmed();
				}

			if(!id3_artist)
				{
				charToQstring(tempbuf+33, 30, tag_artist);
				tag_artist = tag_artist.trimmed();
				}

			if(!id3_album)
				{
				charToQstring(tempbuf+63, 30, tag_album);
				tag_album = tag_album.trimmed();
				}

			if(!id3_year)
				{
				charToQstring(tempbuf+93, 4, tag_year);
				tag_year = tag_year.trimmed();
				}

			if(!id3_comment)
				{
				charToQstring(tempbuf+97, 30, tag_comment);
				tag_comment = tag_comment.trimmed();
				}

			n_genre = static_cast<unsigned char>(tempbuf[127]);
			}
		}

// 3 possible sources for playing time:
// seconds32 from TLEN record
// secondsvbr calculated from Xing info
// secondscalc calculated from bitrate and samprate
	if(seconds32 > 0)
		seconds = seconds32;
	else if(secondsvbr > 0)
		seconds = secondsvbr;
	else
		seconds = secondscalc;

	if(seconds >= 100)
		{
		minutes = seconds/60;
		seconds -= 60*minutes;
		play_time = QString("%1 kbps %2:%3").arg(bitrate).arg(minutes).arg(seconds, 2, 10, QChar('0'));
		if(variable_bitrate)
			play_time += " VBR";
		}
	else if(seconds > 0)
		{
		play_time = QString("%1 kbps").arg(bitrate);
		if(variable_bitrate)
			play_time += " VBR";
		}

	exInfo = play_time;
	if(!tag_title.isEmpty())
		exInfo = QString("%1, %2").arg(exInfo, tag_title);
	if(!tag_artist.isEmpty())
		exInfo = QString("%1, %2").arg(exInfo, tag_artist);
	if(!tag_album.isEmpty())
		exInfo = QString("%1, %2").arg(exInfo, tag_album);
	if(!tag_year.isEmpty())
		exInfo = QString("%1, %2").arg(exInfo, tag_year);
	if(!tag_comment.isEmpty())
		exInfo = QString("%1, %2").arg(exInfo, tag_comment);
	if(!id3_genre && has_tag && (n_genre<N_GENRE))
		exInfo = QString("%1, %2").arg(exInfo, genre[n_genre]);

	if(!exInfo.isEmpty())
		extra_arch(exInfo, dr_nofile);

	goto norm_ret;

maybe_corrupt:
	corrupt_counter++;
	if(corrupt_counter < 20)
		{
		fseek(fp, last_sync_point+1, SEEK_SET);
		header = find_sync(fp);
		goto header_retry;
		}

corrupt:
	return_code = CORRUPT;
	Q_EMIT signal_logMessage(tr("%1%2 is corrupted or not a properly tagged MP3 sound file."
			"  Last file position was %L3").arg(path, filename).arg(ftell(fp)));
norm_ret:
	fclose(fp);
	return(return_code);
	}

// Convert a char array which may not be null terminate to a QString
void IndexWindow::charToQstring(char *ch, int size, QString &str)
	{
	for(int i=0; i<size; i++)
		{
		if(!ch[i])
			break;
		str.append(ch[i]);
		}
	}

// Read possibly unicode text from an mp3 frame.
// Frame id, size and flags have just been read from fp.
// On return, fp is positioned for reading the next frame.
#define CHARBUFSIZE	2048UL
QString IndexWindow::id3text(FILE *fp, unsigned long frame_size, bool hasShortDesc, bool &corruptMp3)
	{
	long tlen;
	long initialPos = ftell(fp);
	QString retString = "";
	char charbuf[CHARBUFSIZE];
	corruptMp3 = false;
// 0=ISO-8859-1
// 1=UTF-16 encoded Unicode with BOM
// 2=UTF-16BE encoded Unicode without BOM, in ID3v2.4.
// 3=UTF-8 encoded Unicode, in ID3v2.4.
	int textEncoding = fgetc(fp);
	if(1 == frame_size)
		{
		retString = "";
		goto position_and_return;
		}
	tlen = qMin(CHARBUFSIZE, frame_size);
	if(textEncoding < 4)
		{
		if(hasShortDesc)
			{
			fseek(fp, 3, SEEK_CUR);	// skip over language
			tlen -= 3;
// Read short content description which is terminated by either one or two zeroes
// and the characters being read may be one or two bytes
			int shortCount;
			for(shortCount=0;;)
				{
				int tval = getc(fp);
				int tval2 = 0;
				charbuf[shortCount++] = tval;
				tlen--;
				if(textEncoding==1 || textEncoding==2)
					{
					tval2 = getc(fp);
					charbuf[shortCount++] = tval2;
					tlen--;
					}
				if((tval2<<8) + tval == 0)
					break;
				}
			switch(textEncoding)
				{
				case 0:     // ISO-8859-1, i.e. ascii
// Special case where what isn't even supposed to be unicode is only a BOM
					if(strncmp(charbuf, reinterpret_cast<char *>(BOMfeff), 3) == 0 || strncmp(charbuf, reinterpret_cast<char *>(BOMfffe), 3) == 0)
						{
						retString = QString();
						goto position_and_return;
						}
					if(shortCount > 1)
						retString = QString::fromLatin1(charbuf);
					break;
				case 1:     // UTF-16 encoded Unicode with BOM
				case 2:     // UTF-16BE encoded Unicode without BOM
					if(shortCount > 2)
						{
						retString = QString::fromUtf16(reinterpret_cast<unsigned short *>(charbuf),
								  static_cast<int>((shortCount-1)/2));
						}
					break;
				case 3:     // UTF-8 encoded Unicode
					if(shortCount > 1)
						retString = QString::fromUtf8(charbuf);
					break;
				}
			if(tlen > 1)
			retString += ": ";
			}
		if(tlen < 2)
			{
			goto position_and_return;
			}
		if(1 != fread(charbuf, tlen-1, 1, fp))
			{
			corruptMp3 = true;
			return retString;
			}
		charbuf[tlen-1] = EOS;
		switch(textEncoding)
			{
		   case 0:     // ISO-8859-1, i.e. ascii
				retString += QString::fromLatin1(charbuf);
    			break;
			case 1:     // UTF-16 encoded Unicode with BOM
		   case 2:     // UTF-16BE encoded Unicode without BOM
				charbuf[tlen-2] = EOS;
    			retString += QString::fromUtf16(reinterpret_cast<unsigned short *>(charbuf),
                    static_cast<int>((tlen-1)/2));
	         break;
         case 3:     // UTF-8 encoded Unicode
    			retString += QString::fromUtf8(charbuf);
    			break;
			}
		}
	else
		corruptMp3 = true;

position_and_return:
	fseek(fp, initialPos+static_cast<long>(frame_size), SEEK_SET);
	return retString;
	}

// Get a long from mp3 id3 header where only 7 of 8 bits are significant
unsigned long IndexWindow::fget_7_long(FILE *fp)
	{
	unsigned long b0, b1, b2, b3;

	b3 = fgetc(fp) & 0x7f;
	b2 = fgetc(fp) & 0x7f;
	b1 = fgetc(fp) & 0x7f;
	b0 = fgetc(fp) & 0x7f;
	return (b3 << 21) + (b2 << 14) + (b1 << 7) + b0;
	}

// ID3 tag document version 2.3.0 does not indicate that the size field
//	has high-bits always zero, but 2.4.0 document does.  Also, it is claimed
//	(and verified by mp3s I've seen) that iTunes gets this wrong.  This is
//	a pain because there's no 100% reliable way to tell which length is
//	right.
//	1. Get both length values.
//	2. If they are the same, we're done.
//	3. If any high bits are set, use the 32-bit length.
//	4. If 32-bit length runs past end of tags, use 28-bit length.
//	5. If 28-bit length points to what looks like a valid frame, use it.
//	6. Use 32-bit length.

unsigned long IndexWindow::fget_guess_long(FILE *fp, unsigned long id3_max)
	{
	unsigned long b0, b1, b2, b3, b4, b5, b6, b7, L28, L32, retlen, save_pos;

	b3 = static_cast<unsigned long>(fgetc(fp));
	b2 = static_cast<unsigned long>(fgetc(fp));
	b1 = static_cast<unsigned long>(fgetc(fp));
	b0 = static_cast<unsigned long>(fgetc(fp));
	L32 = (b3 << 24) + (b2 << 16) + (b1 << 8) + b0;
	b7 = b3 & 0x7f;
	b6 = b2 & 0x7f;
	b5 = b1 & 0x7f;
	b4 = b0 & 0x7f;
	L28 = (b7 << 21) + (b6 << 14) + (b5 << 7) + b4;
	if(L28 == L32)
		return L28;
	if(b0>127 || b1>127 || b2>127 || b3>127)
		return L32;
	save_pos = static_cast<unsigned long>(ftell(fp));
	if(save_pos+L32 > id3_max)
		return L28;
	fseek(fp, static_cast<long>(save_pos+L28+2), SEEK_SET);	// +2 because flags haven't been read yet
	retlen = L28;
// Valid tag is either 4 uppercase letters or 3 plus a number
// so last character gets different test.
	unsigned long tc;
	for(int i=0; i<3; i++)
		{
		tc = static_cast<unsigned long>(fgetc(fp));
		if(tc < 'A' || tc > 'Z')
			{
			retlen = L32;
			goto position_and_return;
			}
		}
	tc = static_cast<unsigned long>(fgetc(fp));
	if(tc < '0' || (tc>'9' && tc < 'A') || tc > 'Z')
		{
		retlen = L32;
		}
position_and_return:
	fseek(fp, static_cast<long>(save_pos), SEEK_SET);
	return retlen;
	}

// Get a long from mp3 id3 header which may have extra zeroes inserted
//	to avoid looking like a false synchronization (0xff followed by byte
//	with first 3 bits set).  Note that if unsync is true, the number of
//	bytes read by the routine depends on the file contents;
unsigned long IndexWindow::fget_unsync_long(FILE *fp, bool unsync)
	{
	bool ff_found;
	unsigned long result;
	unsigned char tchar;
	int i;

	ff_found = false;
	result = 0;
	for(i=0; i<4; i++)
		{
		tchar = static_cast<unsigned char>(fgetc(fp));
		if(ff_found && ((UBIT8(e0) & tchar) == UBIT8(e0)))
			{
			tchar = static_cast<unsigned char>(fgetc(fp));
			}
		result = tchar + (result << 8);
		ff_found = unsync && (UBIT8(ff) == tchar);
		}
	return(result);
	}

// find_sync - find the start of an MP3 frame.  Look for
//	byte with all bits set followed by byte with 3 high-order bits set.
//	Also look for ID3 tag.  Return whichever is found first.
//	last_sync_point bookmarks the first byte of the frame sync or ID3 tag.
#define FOUND_NONE	0
#define FOUND_I		1
#define FOUND_ID		2
#define FOUND_ID3		3

unsigned long IndexWindow::find_sync(FILE *fp)
	{
	bool ff_found;
	unsigned char tchar;
	int i, id3_state;
	unsigned long result;

	ff_found = false;
	id3_state = FOUND_NONE;
// Searching for sync in 1st 2K bytes is not enough in my collection
	for(i=0; i<4096; i++)
		{
		tchar = static_cast<unsigned char>(fgetc(fp));
		if(ff_found && ((UBIT8(e0) & tchar) == UBIT8(e0)))
			{
			result = fgetSwappedWord(fp) + ((0xff00UL + tchar)<<16);
			last_sync_point = ftell(fp) - 4;
			return(result);
			}
		ff_found = (UBIT8(ff) == tchar);
		switch(id3_state)
			{
			case FOUND_NONE:
				if('I' == tchar)
					id3_state = FOUND_I;
				continue;

			case FOUND_I:
				if('D' == tchar)
					id3_state = FOUND_ID;
				else
					id3_state = FOUND_NONE;
				continue;

			case FOUND_ID:
				if('3' == tchar)
					id3_state = FOUND_ID3;
				else
					id3_state = FOUND_NONE;
				continue;

			case FOUND_ID3:
// If it's ID3 version 4 or less, say it's version 3
				if(tchar <= 4)
					{
					last_sync_point = ftell(fp) - 4;
					return(0x49443303);
					}
				else
					id3_state = FOUND_NONE;
				continue;
			}
		}
	return(0);
	}
