// This file is part of the disk cataloger Wssindex.  Wssindex is copyright
// 1985-2021 by Robert W. Babcock and WSS Division of DDC.  Effective 2010,
// Wssindex is distributed under the terms of the GNU General Public
// License.  Contact the author at wssddc@wssddc.com or see the project web
// page at <https://www.wssddc.com/wssindex/>.
//
// Wssindex uses the open source Qt toolkit from the Qt Project (qt-project.org)
// and The Qt Company (www.qt.io).
//
// Wssindex is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the
// Free Software Foundation, either version 3 of the License, or (at your
// option) any later version.
//
// Wssindex is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with Wssindex.  If not, see <https://www.gnu.org/licenses/>.

// The UnRAR DLL source code was used as a basis for this routine.

#include "OtherWssi/wssindex.h"
#include "OtherWssi/globalsettings.h"
#include "OtherWssi/wssiproto.h"
#include "OtherWssi/wsxxxadd.h"
#include "Dialogs/indexwindow.h"
#include "Non_member_functions/customio.h"
#include "Dialogs/mainwindow.h"
#include "3rdparty/unrar.h"

#include <QtDebug>

//int CALLBACK CallbackProc(UINT msg, LPARAM UserData, LPARAM P1, LPARAM P2);
enum { EXTRACT, TEST, PRINT, LIST };

// raradd - read an .RAR file and add information to wssindex database.
//
//	RAR comments are compressed and so not processed.
//	Self-extracting RAR files compress the file names and so are not processed.

#pragma pack(1)
#define TEMPBUF_SIZE 4096
#define CMTBUF_SIZE	16384

int IndexWindow::raradd(QString path, QString filename, int /* dummy */)
	{
	HANDLE hArcData;
	int RHCode,PFCode;
	char CmtBuf[CMTBUF_SIZE];
	struct RARHeaderDataEx HeaderData;
	struct RAROpenArchiveDataEx OpenArchiveData;
	wchar_t RedirName[1024];

	if(skip_extract(filename))
		return(SKIPPED);

// The file extension being processed matches RAR or R00-R99

	int return_code = EXTRACTED_OK;
	unsigned long basefile = Nofile -1 ;
	struct filerec *dr_nofile = Dr + basefile;

	QString name_with_path = path + filename;
	wchar_t* ArcNameW = new wchar_t[static_cast<qulonglong>(name_with_path.length()+1)];
	int nameLength = name_with_path.toWCharArray(ArcNameW);
	ArcNameW[nameLength] = 0;
	memset(&HeaderData, 0, sizeof(HeaderData));
	memset(&OpenArchiveData, 0, sizeof(OpenArchiveData));
	OpenArchiveData.ArcNameW = ArcNameW;
	OpenArchiveData.CmtBuf = CmtBuf;
	OpenArchiveData.CmtBufSize = sizeof(CmtBuf);
	OpenArchiveData.OpenMode = RAR_OM_LIST;
	OpenArchiveData.Callback = nullptr;
//	OpenArchiveData.Callback = CallbackProc;
	OpenArchiveData.UserData = LIST;
	hArcData = RAROpenArchiveEx(&OpenArchiveData);

	if(0 != OpenArchiveData.OpenResult)
		return(CANT_OPEN);

	if(1 == OpenArchiveData.CmtState)
		if(com_arch(CmtBuf, dr_nofile, CMTBUF_SIZE))
			goto no_core;

	HeaderData.RedirName = RedirName;
	HeaderData.RedirNameSize = sizeof(RedirName)/sizeof(RedirName[0]);
	while(0 == (RHCode = RARReadHeaderEx(hArcData, &HeaderData)))
		{
		qlonglong UnpSize = HeaderData.UnpSize + ((static_cast<qlonglong>(HeaderData.UnpSizeHigh)) << 32);
		dr_nofile = Dr + Nofile;
		dr_nofile->Dsize = static_cast<quint64>(UnpSize);
		time1600(dr_nofile, HeaderData.MtimeLow, HeaderData.MtimeHigh);

		char *tptr;
		if(Store_arc_path)
			tptr = HeaderData.FileName;
		else
			tptr = no_path(HeaderData.FileName, static_cast<int>(strlen(HeaderData.FileName)));
		switch(Case_option_indexing)
			{
			case CASE_PRESERVE:
				break;
			case CASE_FORCE_UP:
				_strupr(tptr);
				break;
			case CASE_FORCE_LOW:
				_strlwr(tptr);
				break;
			}

		dr_nofile->Volno  = Diskno;
		dr_nofile->Subdirno = static_cast<unsigned long>(basefile);
		dr_nofile->Category = QString();
		dr_nofile->Comment = QString();
		dr_nofile->Bitflags = 0;
		dr_nofile->Diskbits = 0;
		setInArc(dr_nofile);
//		unbad(tptr, 0);
		dr_nofile->Lfn = tptr;
		dr_nofile->ExtOffset = static_cast<unsigned short>(lastdot(dr_nofile->Lfn));
		setHasLFN(dr_nofile);

		if(mainWindowPtr->incfile_and_check_for_error(false))
			goto no_core;
		PFCode = RARProcessFileW(hArcData, RAR_SKIP, nullptr, nullptr);
		if(0 != PFCode)
			goto corrupt;
		}
// fall out of above loop when finished with a rar file

	if(ERAR_BAD_DATA == RHCode)
		goto corrupt;

norm_ret:
		RARCloseArchive(hArcData);
		delete [] ArcNameW;
//		update_db_totals();
		qApp->processEvents();
		return(return_code);

corrupt:
		return_code = CORRUPT;
		Q_EMIT signal_logMessage(tr("%1%2 is corrupted or not a RAR archive.")
				.arg(path, filename));
		goto norm_ret;

no_core:
		return_code = DATABASE_FULL;
		QMessageBox::critical(this, tr("Error"), tr("Out of memory in raradd"));
		Stop_indexing = true;
		goto norm_ret;
	}

//int CALLBACK CallbackProc(UINT msg, LPARAM /* UserData */, LPARAM /* P1 */, LPARAM /* P2 */)
//	{
//	return(0);
//	}
