// This file is part of the disk cataloger Wssindex.  Wssindex is copyright
// 1985-2021 by Robert W. Babcock and WSS Division of DDC.  Effective 2010,
// Wssindex is distributed under the terms of the GNU General Public
// License.  Contact the author at wssddc@wssddc.com or see the project web
// page at <https://www.wssddc.com/wssindex/>.
//
// Wssindex uses the open source Qt toolkit from the Qt Project (qt-project.org)
// and The Qt Company (www.qt.io).
//
// Wssindex is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the
// Free Software Foundation, either version 3 of the License, or (at your
// option) any later version.
//
// Wssindex is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with Wssindex.  If not, see <https://www.gnu.org/licenses/>.

#include "OtherWssi/wssindex.h"
#include "OtherWssi/wssiproto.h"
#include "OtherWssi/wsxxxadd.h"
#include "Dialogs/indexwindow.h"
#include "Non_member_functions/customio.h"

#define II_2chars ('I' + ('I'<<8))
#define MM_2chars ('M' + ('M'<<8))

// tifadd - read a .TIF file and add information to wssindex database.
//	The image size and number of colors are converted into extra info
//	of the form WIDTHxHEIGHTxNCOLORS, plus monochrome if all colors in
//	color table have equal RGB components.
//
//	If there are multiple images in the TIF file, only the first one will
//	be seen.

static struct th
	{
	unsigned short byte_order;	// II for Intel order, MM for Motorola order
	unsigned short version;		// always 42
	long ifd_offset;				// offset from start of file to first image
										//	file directory
	} tif_header;

static struct ifd
	{
	unsigned short tag;				// identifying number
	unsigned short fieldtype;		// 1=byte, 2=asciiz, 3=short, 4=long,
										//	5=rational, 2 longs numerator/denominator
	long count;						// number of values
	long value_offset;			// offset from start of file to value, or value
										//	itself if it fits in 4 bytes
	} tif_ifd;

int IndexWindow::tifadd(QString path, QString filename, int /* dummy */)
	{
	int bitspersample,
			height,
			photo_interp,
			samplesperpixel,
			width;
	short nifd;

	struct filerec *dr_nofile;

	FILE *fp;
	int ncolor_bits;
	bool mono;

	if(skip_extract(filename))
		return(SKIPPED);

	unsigned long basefile = Nofile - 1;
	int return_code = EXTRACTED_OK;

	if(nullptr == (fp = arc_open(path, filename)))
		return(CANT_OPEN);

	if(1 != fread(&tif_header, sizeof(struct th), 1, fp)
		|| (tif_header.byte_order != II_2chars
			 && tif_header.byte_order != MM_2chars))
		goto corrupt;			// this isn't an error, just a warning

	swap2_tif(reinterpret_cast<char *>(&tif_header.version));
	if(tif_header.version != 42)
		goto corrupt;

	swap4_tif(reinterpret_cast<char*>(&tif_header.ifd_offset));
	fseek(fp, tif_header.ifd_offset, SEEK_SET);
	nifd = static_cast<short>(fgetWord(fp));
	swap2_tif(reinterpret_cast<char *>(&nifd));
	width = height = bitspersample = photo_interp = samplesperpixel = 0;
	mono = false;
	for(int i=0; i<nifd; i++)
		{
		if(1 != fread(&tif_ifd, sizeof(struct ifd), 1, fp))
			goto corrupt;
		swap2_tif(reinterpret_cast<char *>(&tif_ifd.tag));
		switch(tif_ifd.tag)
			{
			case 256:				// ImageWidth
				width = static_cast<short>(tif_decode());
				break;

			case 257:				// ImageHeight
				height = static_cast<short>(tif_decode());
				break;

			case 258:		// BitPerSample, fails if count > 1
				bitspersample = static_cast<short>(tif_decode());
				break;

			case 262:				// PhotometricInterpretation
				photo_interp = static_cast<short>(tif_decode());
				if(photo_interp < 2)
					mono = true;
				break;

			case 277:				// SamplesPerPixel
				samplesperpixel = static_cast<short>(tif_decode());
				break;

			case 290:				// GrayResponseUnit
			case 291:				// GrayResponseCurve
				mono = true;
				break;

			default:
				break;
			}
		}

	ncolor_bits = bitspersample*samplesperpixel;

	dr_nofile = Dr + basefile;
	if(mono)
		dr_nofile->ExtraInfo = tr("%1x%2x%3, monochrome").arg(width).arg(height).arg(num_color_str(ncolor_bits));
	else
		dr_nofile->ExtraInfo = tr("%1x%2x%3").arg(width).arg(height).arg(num_color_str(ncolor_bits));
	setHasExtra(dr_nofile);

norm_ret:
		fclose(fp);
		return(return_code);

corrupt:
		return_code = CORRUPT;
		Q_EMIT signal_logMessage(tr("%1%2 is corrupted or not a TIF image.")
				.arg(path, filename));
		goto norm_ret;
	}

// tif_decode - decode a single byte, short or long value in tif header.  If
//	there are multiple values, or if the type is not byte, short or long,
//	return zero.

long IndexWindow::tif_decode(void)
	{
	union foo
		{
		long long_val;
		short short_val[2];
		char byte_val[4];
		} switcher;

	swap4_tif(reinterpret_cast<char *>(&tif_ifd.count));
	if(tif_ifd.count > 1)
		return(0);

	switcher.long_val = tif_ifd.value_offset;
	swap2_tif(reinterpret_cast<char *>(&tif_ifd.fieldtype));

	switch(tif_ifd.fieldtype)
		{
		case 1:				// byte
			return(switcher.byte_val[0]);

		case 3:				// short
			swap2_tif(reinterpret_cast<char *>(&switcher.short_val[0]));
			return(switcher.short_val[0]);

		case 4:				// long
			swap4_tif(reinterpret_cast<char *>(&switcher.long_val));
			return(switcher.long_val);

		default:
			return(0);
		}
	}

// swap2_tif - interchange two bytes if Motorola format
void IndexWindow::swap2_tif(char *byte_val)
	{
	if(MM_2chars == tif_header.byte_order)
		swap2(byte_val);

	return;
	}

// swap4_tif - reverse 4 bytes if Motorola format
void IndexWindow::swap4_tif(char *byte_val)
	{
	if(MM_2chars == tif_header.byte_order)
		swap4(byte_val);

	return;
	}
