// This file is part of the disk cataloger Wssindex.  Wssindex is copyright
// 1985-2021 by Robert W. Babcock and WSS Division of DDC.  Effective 2010,
// Wssindex is distributed under the terms of the GNU General Public
// License.  Contact the author at wssddc@wssddc.com or see the project web
// page at <https://www.wssddc.com/wssindex/>.
//
// Wssindex uses the open source Qt toolkit from the Qt Project (qt-project.org)
// and The Qt Company (www.qt.io).
//
// Wssindex is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the
// Free Software Foundation, either version 3 of the License, or (at your
// option) any later version.
//
// Wssindex is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with Wssindex.  If not, see <https://www.gnu.org/licenses/>.

#include "OtherWssi/wssindex.h"
#include "OtherWssi/globalsettings.h"
#include "OtherWssi/wssiproto.h"
#include "OtherWssi/wsxxxadd.h"
#include "Dialogs/indexwindow.h"
#include "Dialogs/mainwindow.h"
#include "Non_member_functions/customio.h"

#include <fcntl.h>
#include <io.h>
#include <stdio.h>

#undef CR
#undef END
#undef FILENAME
#define filename zip_filename
#define match zip_match
	#include "3rdparty/InfoZip/unzip.h"
#undef filename
#undef match

// See qplatformdefs.h in ...\qt\mkspecs\whatever for appropriate
// definitions of QT_FSEEK and QT_FTELL to support files larger than 2GB.
#ifdef BUILT_WITH_MinGW
	#define QT_FSEEK	fseeko64
	#define QT_FTELL	ftello64
#endif

#if defined(BUILT_WITH_MSVC09) || defined(BUILT_WITH_MSVC10) || defined(BUILT_WITH_MSVC11) || defined(BUILT_WITH_MSVC12) || defined(BUILT_WITH_MSVC14) || defined(BUILT_WITH_MSVC15) || defined(BUILT_WITH_MSVC16)
	#define QT_FSEEK	_fseeki64
	#define QT_FTELL	_ftelli64
#endif

// zipadd - read a .ZIP file and add information to wssindex database.

static unsigned short this_disk;

// NUM_END_BLOCKS is the number of 512 byte blocks to read, starting from
//	the end of file, when looking for the end of central directory record.
//	Ideally, we could just seek to the expected offset from the end of the
//	file, but there could be some garbage appended.  Other zip-handling
// programs seem to be able to cope with appended junk.
#define NUM_END_BLOCKS	8

#define LOCAL_SIG		0x0403
#define CENTRAL_SIG	0x0201
#define END_SIG		0x0605
#define END_SIG64		0x0606
#define LOC_SIG64		0x0706
#define PK_SIG			0x4b50	// 'PK'
#define END_MARKER	0x6054b50UL		// standard Zip file
#define LOC_MARKER64	0x7064b50UL		// Zip64 format

#define ZL		zip_head.local_u
#define ZC		zip_head.central_u
#define ZE		zip_head.endzip_u
#define ZLOC64	zip_head.loc_64_u
#define ZE64	zip_head.end_record64_u

#pragma pack(1)

struct lh
	{
	unsigned short pk_sig;			// 'PK'
	unsigned short sig;				// 0x403
	unsigned short ext_ver;			// PKUNZIP version needed to extract
	unsigned short bit_flags;		// general purpose flags
	short comp_type;					// compression method
	unsigned short time;				// time stamp, DOS format
	unsigned short date;				// date stamp, DOS format
// Some compilers may want to insert padding here.
	unsigned long crc;				// cyclic redundancy check, not used here
	long compressed_size;			// compressed file size (bytes)
	long uncompressed_size;			// uncompressed size
	unsigned short name_length;	// file name length
	unsigned short extra_length;	// extra field length
// file name and extra field go here
	};

struct ch
	{
	unsigned short pk_sig;			// 'PK'
	unsigned short sig;				// 0x201
	unsigned short made_ver;		// PKZIP version zipfile was made by
	unsigned short ext_ver;			// PKUNZIP version needed to extract
	unsigned short bit_flags;		// general purpose flags
	short comp_type;					// compression method
	unsigned short time;				// time stamp, DOS format
	unsigned short date;				// date stamp, DOS format
	unsigned long crc;				// cyclic redundancy check, not used here
	unsigned long compressed_size;	// compressed file size (bytes)
	unsigned long uncompressed_size;	// uncompressed size
	unsigned short name_length;	// file name length
	unsigned short extra_length;	// extra field length
	unsigned short com_length;		// file comment length
	unsigned short disk_start;		// disk number on which this file starts
	unsigned short int_attr;		// low bit 0 for ascii, 1 for binary
// Some compilers may want to insert padding here.

	unsigned long ext_attr;			// low byte is directory attribute byte
	unsigned long loc_head;			// offset to local header from start of disk
											//	on which this file begins
// file name, extra field, and file comment go here
		};

struct ez
	{
	unsigned short pk_sig;			// 'PK'
	unsigned short sig;				// 0x605
	unsigned short thisno;			// number of this disk
	unsigned short cent_disk;		// number of disk with start of central dir
	unsigned short dir_count;		// number of central dir entries on this disk
	unsigned short total_count;	// total number of central dir entries
	unsigned long central_size;	// size of central dir
	unsigned long central_off;		// offset of central dir on starting disk
	unsigned short com_length;		// zipfile comment length
// zipfile comment goes here
	};

struct ez64
	{
	unsigned short pk_sig;				// 'PK'
	unsigned short sig;					// 0x606
	quint64 ez64_size;		// size of zip64 end of central directory record
	unsigned short ver_madeby;			// version made by
	unsigned short ver_needed;			// version needed to extract
	unsigned int thisno;					// number of this disk
	unsigned int cent_disk;				// number of disk with start of central dir
	quint64 dir_count;		// number of central dir entries on this disk
	quint64 total_count;		// total number of central dir entries
	quint64 central_size;	// size of central dir
	quint64 central_off;		// offset of central dir on starting disk
// zip64 extensible data sector goes here
	};

struct extra64
	{
	unsigned short tag;					// 0x0001 for zip64
	unsigned short extra_size;			// size of this extra block
	quint64 uncomp_size;		// uncompressed file size
	quint64 comp_size;		// compressed file size
	quint64 local_offset;	// offset of local header record
	unsigned int diskno;					// number of disk on which this file starts
	};

struct loc64
	{
	unsigned long pk_sig;			// LOC_SIG64
	unsigned int cent_disk;			// number of disk with start of central dir
	qint64 central_off;				// offset of central dir on starting disk
	unsigned int disk_count;		// total number of disks
	};

static union
	{
	struct lh local_u;
	struct ch central_u;
	struct ez endzip_u;
	struct loc64 loc_64_u;
	struct ez64 end_record64_u;
	} zip_head;
#define TEMPBUF_SIZE 4096

int IndexWindow::zipadd(QString path, QString filename, int mode)
	{
	char tempbuf[TEMPBUF_SIZE];
//	struct ez end_record;
//	struct ez64 end_record64;
	struct loc64 loc_record64;
	unsigned long diz_fileno = 0;	// initialize to quiet warning
	char *full_name;
	FILE *fp;
	int length;
	long long abs_pos_next, offset_to_next, save_place;
	unsigned int central_disk;
	unsigned short extra_skip_length;
	quint64 uncompressed_size = 0;	// initialized to quiet compiler warning
	bool continued,						// central directory split
		  is_zip64 = false,				// Zip64 format
		  multi_disk = false;			// multi-volume zip file, initialize
												//	to quiet compiler warning
	short diz_disk;
	unsigned long diz_loc_header = 0;	// initialize to quiet warning

	if(skip_extract(filename))
		return(SKIPPED);

	diz_disk = -1;
	this_disk = static_cast<unsigned short>(-1);
	unsigned long basefile = Nofile - 1;

	int return_code = EXTRACTED_OK;

	if(nullptr == (fp = arc_open(path, filename)))
		return(CANT_OPEN);

	struct filerec *dr_nofile = Dr + basefile;

	continued = false;
	int handle = _fileno(fp);

	offset_to_next = _filelengthi64(handle) - 512;
	for(int i=0; i<NUM_END_BLOCKS; i++)
		{
		if(offset_to_next < 0)
			offset_to_next = 0;
		if(0 != QT_FSEEK(fp, offset_to_next, SEEK_SET)
			|| 0 >= (length = static_cast<int>(fread(tempbuf, 1, 512, fp))))
			break;
		length -= 3;
// Find end of zip marker.  If the last file in the ZIP is itself an
// uncompressed ZIP, it will have an end marker near the desired
// marker.  Must scan backwards to avoid this possibility.
		for(int j=length; j>0;)
			{
			j--;
			if(END_MARKER != *(reinterpret_cast<unsigned long *>(tempbuf+j)))
				continue;
// Found marker for end-of-central-directory record.
			QT_FSEEK(fp, j-length-3, SEEK_CUR);
			if(0 == QT_FTELL(fp))
				{
				Q_EMIT signal_logMessage(tr("%1%2 is completely empty.").arg(path, filename));
				goto norm_ret;
				}
			if(sizeof(struct ez) != fread(&ZE, 1, sizeof(struct ez), fp))
				goto corrupt;
			this_disk = ZE.thisno;
			multi_disk = this_disk > 0;
// Back up and see if there's a zip64 locator record
			QT_FSEEK(fp, -static_cast<qint64>(sizeof(struct ez)+sizeof(struct loc64)), SEEK_CUR);
			if(sizeof(struct loc64) != fread(&loc_record64, 1, sizeof(struct loc64), fp))
				goto corrupt;
			if(LOC_MARKER64 == loc_record64.pk_sig)
				{
				is_zip64 = true;
				QT_FSEEK(fp, loc_record64.central_off, SEEK_SET);
				if(sizeof(struct ez64) != fread(&zip_head, 1, sizeof(struct ez64), fp))
					goto corrupt;
				central_disk = ZE64.cent_disk;
// If central_off is wrong, it's too late, we've already used it
				abs_pos_next = static_cast<qint64>(ZE64.central_off);
				}
			else
				{
// central_off may be wrong in self-extractors, so use size field instead
//				abs_pos_next = ZE.central_off;
				abs_pos_next = ftell(fp) - static_cast<long>(ZE.central_size);
				central_disk = ZE.cent_disk;
				}
// Found end-of-central-directory record, does central directory start on
//	another disk?  Is this a multi-volume zip file?
			if(this_disk <  central_disk)
				{
				Q_EMIT signal_logMessage(tr("%1%2 has bad disk number for central directory, ignoring.").arg(path, filename));
				}
			else if(this_disk != central_disk)
				{
				fclose(fp);
				QMessageBox msgBox;
				msgBox.setText(tr("ZIP file directory for %1 spans disks, please insert previous disk").arg(filename));
				msgBox.addButton(tr("Ready"), QMessageBox::ActionRole);
				QPushButton *abortButton = msgBox.addButton(tr("Skip this file"), QMessageBox::ActionRole);
				msgBox.exec();

				if(msgBox.clickedButton() == abortButton)
					return(SKIPPED);
				if(nullptr == (fp = arc_open(path, filename)))
					return(CANT_OPEN);
				this_disk--;
				continued = true;
				}

			goto central_read;
			}
		offset_to_next -= 508;
		}

// Didn't find central directory; error unless looking for possible
//	self-extracting file.

	if(0 == QFA("z").compare(dr_nofile->Lfn.midRef(dr_nofile->ExtOffset,1), Qt::CaseInsensitive))
		goto corrupt;
	else
		{
		return_code = NOT_RECOGNIZED;
		goto all_done_ret;
		}

// central directory found; extract info

central_read:
	clearerr(fp);
	if(QT_FSEEK(fp, abs_pos_next, SEEK_SET))
		goto corrupt;
	offset_to_next = 0;
	for(;;)
		{
		if(QT_FSEEK(fp, offset_to_next, SEEK_CUR))
			goto corrupt;
		save_place = QT_FTELL(fp);
		length = static_cast<int>(zread(&zip_head, 1, sizeof(struct ch), &fp, &continued, path,
							filename));
		if(ZC.pk_sig != PK_SIG)
			goto corrupt;
		switch(ZC.sig)
			{
			case CENTRAL_SIG:
				if(sizeof(struct ch) != length
					|| (length = ZC.name_length) > 512
					|| 1 != zread(tempbuf, static_cast<size_t>(length), 1, &fp, &continued, path,
								filename))
					{
					goto corrupt;		// this is an error, but not a severe one
					}
// Do we need to look at extra field to get a size too big for 32-bits?
				extra_skip_length = ZC.extra_length;
				if((is_zip64 || 45==ZC.ext_ver) && 0xffffffffUL == ZC.uncompressed_size && extra_skip_length > 0)
					{
					while(extra_skip_length > 0)
						{
						short extra_header_id = static_cast<short>(getshort(fp));
						short extra_data_size = static_cast<short>(getshort(fp));
						if(1 != extra_header_id)
							{
							extra_skip_length -= extra_data_size + 4;
							QT_FSEEK(fp, extra_data_size, SEEK_CUR);
							continue;
							}
						uncompressed_size = get64(fp);
						QT_FSEEK(fp, extra_data_size-8, SEEK_CUR);
						break;
						}
					}
				else
					{
					uncompressed_size = ZC.uncompressed_size;
					QT_FSEEK(fp, static_cast<long long>(extra_skip_length), SEEK_CUR);
					}

				if(!(extractMembers & mode))
					length = 0;
				else
					{
					dr_nofile = Dr + Nofile;
					full_name = tempbuf;
					*(full_name+length) = EOS;
					if(!Store_arc_path)
						full_name = no_path(tempbuf, length);

// ZIP files created under Unix may have entries for directories which
//	have nothing left after directory name is stripped off

					length = 0;
					if(*full_name)
						{
						switch(Case_option_indexing)
							{
							case CASE_PRESERVE:
								break;
							case CASE_FORCE_UP:
								_strupr(full_name);
								break;
							case CASE_FORCE_LOW:
								_strlwr(full_name);
								break;
							}

						dr_nofile->Ddate = datextract(ZC.date);
						dr_nofile->Dtime = timextract(ZC.time);
						dr_nofile->Dsize = uncompressed_size;
						dr_nofile->Volno = Diskno;
						dr_nofile->Subdirno = static_cast<unsigned long>(basefile);
						dr_nofile->Category = QString();
						dr_nofile->Comment = QString();
						dr_nofile->Bitflags = 0;
						dr_nofile->Diskbits = 0;
						setInArc(dr_nofile);
						unbad(full_name, 0);
						dr_nofile->Lfn = full_name;
						dr_nofile->ExtOffset = static_cast<unsigned short>(lastdot(dr_nofile->Lfn));
						setHasLFN(dr_nofile);

						if((extractMemberComments & mode) && (length = ZC.com_length) > 0)
							{
							if(length > TEMPBUF_SIZE-1)
								length = TEMPBUF_SIZE-1;
							if(1 != zread(tempbuf, static_cast<size_t>(length), 1, &fp, &continued, path,
									filename))
								goto corrupt;			// this is a severe error
							if(com_arch(tempbuf, dr_nofile, length))
								goto no_core;
							}
						if(Extract_diz && 0 == QFA("FILE_ID.DIZ").compare(dr_nofile->Lfn, Qt::CaseInsensitive))
							{
							diz_disk = static_cast<short>(ZC.disk_start);
							diz_loc_header = ZC.loc_head;
							if(DIZ_SELF == Extract_diz)
								diz_fileno = Nofile;
							else
								diz_fileno = basefile;
							}
						if(mainWindowPtr->incfile_and_check_for_error(false))
							goto no_core;
						}
					}

// skip over comment if not reading or if it was too long

				offset_to_next = ZC.com_length - length;
				break;

			case END_SIG64:
// Skip over zip64 end of central directory record
				offset_to_next = static_cast<qint64>(ZE64.ez64_size) + 12 - static_cast<qint64>(sizeof(struct ch));
				break;

			case LOC_SIG64:
// Skip over zip64 end of central directory locator record
// Last read probably got less than sizeof(struct ch) bytes, so offset_to_next
// is not sizeof(struct loc64)-sizeof(struct ch)
				offset_to_next = static_cast<int>(sizeof(struct loc64)) - length;
				break;

			case END_SIG:
				if(!(extractArcComment & mode) || (length = ZE.com_length) == 0)
					goto norm_ret;
				clearerr(fp);		// last read may have tried to go past EOF
				QT_FSEEK(fp, save_place + static_cast<qint64>(sizeof(struct ez)), SEEK_SET);
				dr_nofile = Dr + basefile;
				if(length > TEMPBUF_SIZE-1)
					length = TEMPBUF_SIZE-1;
				if(1 != zread(tempbuf, static_cast<size_t>(length), 1, &fp, &continued, path,
								filename))
					goto corrupt;			// this is a severe error

				if(com_arch(tempbuf, dr_nofile, length))
					goto no_core;
				goto norm_ret;

			default:
//				QString str;
//				QMessageBox::information(this, "Debug", QFA("%1 corrupt because sig = %2").arg(filename).arg(str.setNum(ZC.sig,16)));
				goto corrupt;
			}
		}

// never fall out of above loop

corrupt:
		return_code = CORRUPT;
		Q_EMIT signal_logMessage(tr("%1%2 is corrupted, continued on another disk or not a ZIP archive.")
				.arg(path, filename));
		fclose(fp);
//		update_db_totals();
		qApp->processEvents();
		return(return_code);

norm_ret:
		if(diz_disk >= 0)
			{
			if(ExtractDiz(diz_fileno, diz_disk, diz_loc_header, &fp, path,
								filename))
				goto no_core;
			}

all_done_ret:
		fclose(fp);
		if(multi_disk && this_disk != ZE.thisno)
			QMessageBox::information(this, tr("Info"),
				tr("Insert original disk and hit OK to continue indexing"), QMessageBox::Ok);
//		update_db_totals();
		qApp->processEvents();
		return(return_code);

no_core:
		return_code = DATABASE_FULL;
		QMessageBox::critical(this, tr("Error"), tr("Out of memory in zipadd"));
		Stop_indexing = true;
		goto norm_ret;
	}

// zread - read a ZIP file, prompt for necessary disk swaps if spanned

size_t IndexWindow::zread(void *buffer, size_t size, size_t n, FILE **fp,
							bool *continued, QString path, QString filename)
	{
	size_t length, wanted;

	if(0 == size)
		return(0);
	wanted = size * n;
	length = fread(buffer, 1, wanted, *fp);
	if(!*continued || wanted == length)
		return(length/size);

	fclose(*fp);
	int ret = QMessageBox::information(this, tr("Info"),
			tr("ZIP file directory for %1%2 spans disks, insert next disk").arg(path, filename),
			QMessageBox::Ok | QMessageBox::Cancel);
	if(QMessageBox::Cancel == ret)
		return(0);
	this_disk++;
	if(nullptr == (*fp = arc_open(path, filename)))
		return(0);
	wanted -= length;
	length += fread(reinterpret_cast<char *>(buffer)+length, 1, wanted, *fp);
	*continued = false;
	return(length/size);
	}

// arc_open - open archive file for reading.  Returns stream pointer, or NULL
//	if file can't be opened.

FILE *IndexWindow::arc_open(QString path, QString filename)
	{
	FILE *fp;
	QString name_with_path = path;

	name_with_path += filename;

	if(nullptr == (fp = _wfopen(name_with_path.toStdWString().data(), L"rb")))
		{
		Q_EMIT signal_logMessage(tr("Can't open %1 for info extraction.").arg(name_with_path));
		return nullptr;
		}

	return(fp);
	}

// Extract info from file_id.diz.  Returns WSSERROR (true) if out of memory,
//	otherwise returns NO_WSSERROR (false).

bool IndexWindow::ExtractDiz(unsigned long filenum, short diz_disk,
			unsigned long diz_loc_head, FILE **fp, QString path, QString filename)
	{
	unsigned int length, rc;
	struct filerec *dr_filenum;
	char tempbuf[TEMPBUF_SIZE];

	if(this_disk != diz_disk)
		{
		fclose(*fp);
		int ret = QMessageBox::information(this, tr("Info"),
			tr("FILE_ID.DIZ for %1%2 is on another disk; insert disk %2 of set")
			.arg(path, filename).arg(diz_disk+1),	QMessageBox::Ok | QMessageBox::Cancel);
		if(QMessageBox::Cancel == ret)
			{
			Stop_indexing = true;
			return(false);
			}
		this_disk = static_cast<unsigned short>(diz_disk);
		if(nullptr == (*fp = arc_open(path, filename)))
			{
			Stop_indexing = true;
			return(false);
			}
		}
	clearerr(*fp);		// last read may have tried to go past EOF
	if(0 != QT_FSEEK(*fp, static_cast<qint64>(diz_loc_head), SEEK_SET)
		|| 0 == (length = static_cast<unsigned int>(fread(&ZL, 1, sizeof(struct lh), *fp)))
		|| ZL.pk_sig != PK_SIG
		|| ZL.sig != LOCAL_SIG)
		goto fileid_error;
	QT_FSEEK(*fp, static_cast<qint64>(ZL.name_length+ZL.extra_length), SEEK_CUR);
	length = static_cast<unsigned int>(ZL.uncompressed_size);
	if(length > TEMPBUF_SIZE-1)
		length = TEMPBUF_SIZE-1;
	dr_filenum = Dr + filenum;
	switch(ZL.comp_type)
		{
		case STORED:
			if(length != fread(tempbuf, 1, length, *fp))
				goto fileid_error;
			rc = 0;
			break;
		case SHRUNK:
			rc = static_cast<unsigned int>(wssUnshrink(*fp, ZL.compressed_size, tempbuf, TEMPBUF_SIZE));
			break;
// various REDUCEDn cases are only created by pkzip 0.9 or earlier, not
//	likely to be used for FILE_ID.DIZ

		case REDUCED1:
		case REDUCED2:
		case REDUCED3:
		case REDUCED4:
			rc = 1;
			break;
		case IMPLODED:
			rc = static_cast<unsigned int>(wssExplode(*fp, ZL.bit_flags, length, tempbuf, TEMPBUF_SIZE));
			break;
		case DEFLATED:
			rc = static_cast<unsigned int>(wssInflate(*fp, ZL.compressed_size, tempbuf, TEMPBUF_SIZE));
			break;
		default:
			goto fileid_error;
		}
	if(0 == rc)
		{
		if(com_arch(tempbuf, dr_filenum, static_cast<int>(length)))
			goto no_core;
		}
	else
		{
		Q_EMIT signal_logMessage(tr("Can't decompress FILE_ID.DIZ in %1%2.")
				.arg(path, filename));
		}

	return(false);

fileid_error:
	Q_EMIT signal_logMessage(tr("Error reading FILE_ID.DIZ in %1%2.")
			.arg(path, filename));
	return(false);

no_core:
	QMessageBox::critical(this, tr("Error"), tr("Out of memory extracting comment from FILE_ID.DIZ"));
	return(true);
	}


// skip_extract - ask whether archive information should be extracted
bool IndexWindow::skip_extract(QString file_name)
	{
	if(Query_extract)
		{
		if(QueryOverride)
			return SkipOverrideAnswer;

		QMessageBox msgBox;
		msgBox.setText(tr("Extract file/comment info from %1?").arg(file_name));
		msgBox.setStandardButtons(QMessageBox::Yes | QMessageBox::No);
		QPushButton *yesAllButton = msgBox.addButton(tr("Yes to all"), QMessageBox::ActionRole);
		msgBox.addButton(tr("No to all"), QMessageBox::ActionRole);
		int ret = msgBox.exec();
		if(QMessageBox::Yes == ret)
			return true;
		if(QMessageBox::No == ret)
			return false;
		QueryOverride = true;
		if(msgBox.clickedButton() == yesAllButton)
			SkipOverrideAnswer = false;
		else
			SkipOverrideAnswer = true;
		return SkipOverrideAnswer;
		}
	return(false);
	}
