// This file is part of the disk cataloger Wssindex.  Wssindex is copyright
// 1985-2021 by Robert W. Babcock and WSS Division of DDC.  Effective 2010,
// Wssindex is distributed under the terms of the GNU General Public
// License.  Contact the author at wssddc@wssddc.com or see the project web
// page at <https://www.wssddc.com/wssindex/>.
//
// Wssindex uses the open source Qt toolkit from the Qt Project (qt-project.org)
// and The Qt Company (www.qt.io).
//
// Wssindex is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the
// Free Software Foundation, either version 3 of the License, or (at your
// option) any later version.
//
// Wssindex is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with Wssindex.  If not, see <https://www.gnu.org/licenses/>.

#include "OtherWssi/wssindex.h"
#include "Dialogs/indexwindow.h"
#include "ui_indexwindow.h"
#include "ExtractSubs/extractinfo.h"
#include "OtherWssi/wssiproto.h"
#include "OtherWssi/wsxxxadd.h"

#include <QDateTime>
#include <QtDebug>

void IndexWindow::extract_info(QString path, struct filerec *dr_info)
	{
	if(0 == dr_info->Dsize)
		return;
	if(0 == dr_info->ExtOffset)
		return;
	QElapsedTimer timer;
	timer.start();
	QString ext = dr_info->Lfn.mid(dr_info->ExtOffset).toUpper();
	for(int i=0; i<NLIBTYPE; i++)
		{
		int mode = Extract[i].flag;
		if(ui->radioButton_extract_configured->isChecked() && (dontExtract==mode))
			continue;
		if(ui->radioButton_extract_all->isChecked())
			mode = extractAll;
		if(Extract[i].nameRegexp.indexIn(ext) >= 0)
			{
			switch(Extract[i].routine)
				{
// Archives
				case zip7Extract:
					zip7add(path, dr_info->Lfn, mode);
					goto checkTime;
				case arcExtract:
				case arkExtract:
				case pakExtract:
				case pkaExtract:
				case sdnExtract:
					arcadd(path, dr_info->Lfn, mode);
					goto checkTime;
				case arjExtract:
					arjadd(path, dr_info->Lfn, mode);
					goto checkTime;
				case cabExtract:
					cabadd(path, dr_info->Lfn, mode);
					goto checkTime;
				case hapExtract:
					hapadd(path, dr_info->Lfn, mode);
					goto checkTime;
				case hypExtract:
					hypadd(path, dr_info->Lfn, mode);
					goto checkTime;
				case lbrExtract:
					lbradd(path, dr_info->Lfn, mode);
					goto checkTime;
				case lhaExtract:
				case lzsExtract:
					lhaadd(path, dr_info->Lfn, mode);
					goto checkTime;
				case rqqExtract:
					raradd(path, dr_info->Lfn, mode);
					goto checkTime;
				case zipExtract:
					zipadd(path, dr_info->Lfn, mode);
					goto checkTime;
				case zooExtract:
					zooadd(path, dr_info->Lfn, mode);
					goto checkTime;

// Fonts
				case gsfExtract:
				case pfaExtract:
				case pfbExtract:
					psfadd(path, dr_info->Lfn, mode);
					goto checkTime;
				case ttfExtract:
					ttfadd(path, dr_info->Lfn, mode);
					goto checkTime;

// Images
				case bmpExtract:
					bmpadd(path, dr_info->Lfn, mode);
					goto checkTime;
				case gifExtract:
					gifadd(path, dr_info->Lfn, mode);
					goto checkTime;
				case jpgExtract:
					jpgadd(path, dr_info->Lfn, mode);
					goto checkTime;
				case pcxExtract:
					pcxadd(path, dr_info->Lfn, mode);
					goto checkTime;
				case pngExtract:
					pngadd(path, dr_info->Lfn, mode);
					goto checkTime;
				case tgaExtract:
					tgaadd(path, dr_info->Lfn, mode);
					goto checkTime;
				case tifExtract:
					tifadd(path, dr_info->Lfn, mode);
					goto checkTime;

// Video
				case aviExtract:
					aviadd(path, dr_info->Lfn, mode);
					goto checkTime;
				case flcExtract:
				case fliExtract:
					fliadd(path, dr_info->Lfn, mode);
					goto checkTime;
				case flvExtract:
				case mpgExtract:
				case mkvExtract:
				case mp4Extract:
				case ogmExtract:
				case qtExtract:
				case swfExtract:
				case wmvExtract:
					mediaInfoVideoAdd(path, dr_info->Lfn, mode);
					goto checkTime;

// Audio
				case auExtract:
					auadd(path, dr_info->Lfn, mode);
					goto checkTime;
				case flacExtract:
				case m4aExtract:
					mediaInfoSoundAdd(path, dr_info->Lfn, mode);
					goto checkTime;
				case modExtract:
					modadd(path, dr_info->Lfn, mode);
					goto checkTime;
				case mp3Extract:
					mp3add(path, dr_info->Lfn, mode);
					goto checkTime;
				case vocExtract:
					vocadd(path, dr_info->Lfn, mode);
					goto checkTime;
				case wavExtract:
					wavadd(path, dr_info->Lfn, mode);
					goto checkTime;

// Following are the potential self-extracting files
// Since these all match *.exe, we can't break unless the extraction succeeded
				case arcExtractExe:
					if(EXTRACTED_OK == arcadd(path, dr_info->Lfn, mode))
						goto checkTime;
					continue;
				case arjExtractExe:
					if(EXTRACTED_OK == arjadd(path, dr_info->Lfn, mode))
						goto checkTime;
					continue;
				case lhaExtractExe:
					if(EXTRACTED_OK == lhaadd(path, dr_info->Lfn, mode))
						goto checkTime;
					continue;
				case zipExtractExe:
					if(EXTRACTED_OK == zipadd(path, dr_info->Lfn, mode))
						goto checkTime;
					continue;
// To quiet compiler warning, should never be triggered
				case endlistExtract:
					break;
				}
			break;
			}
		}
	return;

checkTime:
	long long extract_ms = timer.elapsed();
// ***** Is this reasonable???
	if(extract_ms > 100)
		qApp->processEvents();
	}
