// This file is part of the disk cataloger Wssindex.  Wssindex is copyright
// 1985-2021 by Robert W. Babcock and WSS Division of DDC.  Effective 2010,
// Wssindex is distributed under the terms of the GNU General Public
// License.  Contact the author at wssddc@wssddc.com or see the project web
// page at <https://www.wssddc.com/wssindex/>.
//
// Wssindex uses the open source Qt toolkit from the Qt Project (qt-project.org)
// and The Qt Company (www.qt.io).
//
// Wssindex is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the
// Free Software Foundation, either version 3 of the License, or (at your
// option) any later version.
//
// Wssindex is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with Wssindex.  If not, see <https://www.gnu.org/licenses/>.

#include "OtherWssi/wssindex.h"
#include <windows.h>

#include <QDateTime>

// timextract(k) - extract time variables from MS-DOS packed format hms
QTime timextract(unsigned short hms)
	{
	return QTime( (hms >> 11) & 63, (hms >> 5) & 63, (hms & 31) << 1);
	}

// datextract(k) - extract date variables from MS-DOS packed format dmy
QDate datextract(unsigned short dmy)
	{
	return QDate( ((dmy >> 9) & 127)+1980, (dmy >> 5) & 15, dmy & 31 );
	}

short msdosDate(QDateTime timestamp)
	{
	int year = timestamp.date().year();
	int month = timestamp.date().month();
	int day = timestamp.date().day();
	if(year < 1980)
		{
		year = 1980; month = 1; day = 1;
		}
	if(year > 2107)
		{
		year = 2107; month = 12; day = 31;
		}
	return static_cast<short>(((year-1980)<<9) + (month<<5) + day);
	}

// Compare two times ignoring milliseconds and possibly allowing for a
// one hour difference due to daylight savings time.
bool timeEqual(QDate d1, QTime t1, QDate d2, QTime t2, bool ignoreOneHour)
	{
	if(t1.minute() != t2.minute())
		return false;
	if(t1.second() != t2.second())
		return false;
	if(ignoreOneHour)
		{
		qint64 julianHour1 = d1.toJulianDay()*24 + t1.hour();
		qint64 julianHour2 = d2.toJulianDay()*24 + t2.hour();
		if(julianHour1 == julianHour2)
			return true;
		if(1 == abs(static_cast<int>(julianHour1-julianHour2)))
			return true;
		return false;
		}

	if(d1 != d2)
		return false;
	if(t1.hour() != t2.hour())
		return false;
	return true;
	}

// Convert FILETIME (100-nanosecond intervals since January 1, 1601 (UTC)) to
// local time and set database date/time values for one file.

void time1600(struct filerec *dr, unsigned int lowtime, unsigned int hightime)
	{
	FILETIME ft;
	SYSTEMTIME ust,st;
	ft.dwLowDateTime = lowtime;
	ft.dwHighDateTime = hightime;
	FileTimeToSystemTime(&ft, &ust);
	SystemTimeToTzSpecificLocalTime(nullptr, &ust, &st);

	dr->Dtime = QTime(st.wHour, st.wMinute, st.wSecond, st.wMilliseconds);
	dr->Ddate = QDate(st.wYear, st.wMonth, st.wDay);
	}
