// This file is part of the disk cataloger Wssindex.  Wssindex is copyright
// 1985-2021 by Robert W. Babcock and WSS Division of DDC.  Effective 2010,
// Wssindex is distributed under the terms of the GNU General Public
// License.  Contact the author at wssddc@wssddc.com or see the project web
// page at <https://www.wssddc.com/wssindex/>.
//
// Wssindex uses the open source Qt toolkit from the Qt Project (qt-project.org)
// and The Qt Company (www.qt.io).
//
// Wssindex is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the
// Free Software Foundation, either version 3 of the License, or (at your
// option) any later version.
//
// Wssindex is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with Wssindex.  If not, see <https://www.gnu.org/licenses/>.

#include "OtherWssi/pinmenu.h"
#include "3rdparty/focuswatcher.h"
#include <QMenu>
#include <QPalette>
#include <QtDebug>
#include <QKeyEvent>

PinMenuItem::PinMenuItem(QMenu *parent, int id, bool pinstate, const QString& title) : QWidgetAction(nullptr)
	{
	containingMenu = parent;
	pinmap = QPixmap(":/Images/pin.png");
	unpinmap = QPixmap(":/Images/unpin.png");
	pinWidget = new QWidget(containingMenu);
	pinWidget->setAutoFillBackground(true);
	pinLayout = new QHBoxLayout();
	pinLayout->setContentsMargins(18, 3, 5, 3);
	pinTitle = new ClickableLabel(title);
	pinWidget->setFocusPolicy(Qt::StrongFocus);
	pinWidget->installEventFilter(this);

	pinLayout->addWidget(pinTitle);
	pinImage = new ClickableLabel(nullptr);
	pinImage->setAlignment(Qt::AlignRight);
// Limit the width of pinImage so clicking on the blank space between pinTitle
// and pinImage will be interpreted as clicking on pinTitle, not pinImage.
	pinImage->setMaximumWidth(qMax(pinmap.width(), unpinmap.width()));
	pinLayout->addWidget(pinImage);
	pinWidget->setLayout(pinLayout);
	setPinState(pinstate);
	setDefaultWidget(pinWidget);
	idno = id;

	connect(pinTitle, SIGNAL(clicked()), this, SLOT(slot_textClicked()));
	connect(pinImage, SIGNAL(clicked()), this, SLOT(slot_pinClicked()));
	connect(pinTitle, SIGNAL(labelHovered()), this, SLOT(slot_hovered()));
	connect(pinTitle, SIGNAL(labelUnhovered()), this, SLOT(slot_unhovered()));
	FocusWatcher *fw = new FocusWatcher(pinWidget);
	connect(fw, SIGNAL(focusChanged(bool)), this, SLOT(slot_focusChanged(bool)));

	QPalette pal(pinWidget->palette());
	defaultBackground = pal.color(QPalette::Window);
	highlightBackground = pal.color(QPalette::Highlight);
	defaultText = pal.color(QPalette::Text).rgb();
	highlightText = pal.color(QPalette::HighlightedText).rgb();
	}

PinMenuItem::~PinMenuItem()
	{
	containingMenu->setFocus();
	delete pinImage;
	delete pinTitle;
	delete pinLayout;
	delete pinWidget;
	}

bool PinMenuItem::isPinned()
	{
	return pinned;
	}

bool PinMenuItem::togglePin()
	{
	setPinState(!pinned);
	return pinned;
	}

void PinMenuItem::setPinState(bool pin)
	{
	pinned = pin;
	pinImage->setPixmap(pinned ? pinmap : unpinmap);
	pinImage->setToolTip(pinned ? tr("<html><head/><body><p>Locked to menu<br><br>Space toggles pin, Enter opens selection</p></body></html>")
										 : tr("<html><head/><body><p>Will be removed if menu overflows<br><br>Space toggles pin, Enter opens selection</p></body></html>"));
	}

void PinMenuItem::setText(const QString& newtitle)
	{
	pinTitle->setText(newtitle);
	}

void PinMenuItem::slot_textClicked()
	{
	containingMenu->hide();
	Q_EMIT signal_textClicked(idno);
	}

void PinMenuItem::slot_pinClicked()
	{
	togglePin();
	Q_EMIT signal_pinClicked(idno);
	}

void PinMenuItem::slot_hovered()
	{
	QPalette pal(pinWidget->palette());
	pal.setColor(QPalette::Window, highlightBackground);
	this->pinTitle->setStyleSheet(tr("QLabel {color: rgb(%1,%2,%3);}")
		.arg(qRed(highlightText)).arg(qGreen(highlightText)).arg(qBlue(highlightText)));
	pinWidget->setPalette(pal);
	}

void PinMenuItem::slot_unhovered()
	{
	QPalette pal(pinWidget->palette());
	pal.setColor(QPalette::Window, defaultBackground);
	this->pinTitle->setStyleSheet(tr("QLabel {color: rgb(%1,%2,%3);}")
		.arg(qRed(defaultText)).arg(qGreen(defaultText)).arg(qBlue(defaultText)));
	pinWidget->setPalette(pal);
	}

void PinMenuItem::slot_focusChanged(bool /* in */)
	{
	if(pinWidget->hasFocus())
		slot_hovered();
	else
		slot_unhovered();
	}

bool PinMenuItem::eventFilter(QObject *obj, QEvent *event)
	{
	if(QEvent::KeyPress == event->type())
		{
		QKeyEvent *kevent = reinterpret_cast<QKeyEvent *>(event);
		if(0 == ((Qt::ShiftModifier |  Qt::ControlModifier | Qt::AltModifier) & kevent->modifiers()))
			{
			switch(kevent->key())
				{
				case Qt::Key_Space:
					slot_pinClicked();
					return true;
				case Qt::Key_Return:
				case Qt::Key_Enter:
					slot_textClicked();
					return true;
				default:
					break;
				}
			}
		}
	return QObject::eventFilter(obj, event);
	}
